"""Things that are shared among files. 
"""

__id__ = "$Id: common.py 549 2007-07-30 22:14:52Z theobald $"

from node import *
import os
import sys
import traceback
import ConfigParser

TYPE_UNCERTAIN = 0
TYPE_CERTAIN = 1
TYPE_CONFIDENCE = 2
TYPE_PROBABILISTIC = 3

ROOT = os.path.dirname(os.path.abspath(__file__))

cp = ConfigParser.ConfigParser()
cp.read(ROOT + "/trio.cfg")

def print_exception(msg=''):
    e = sys.exc_info()
    print
    if int(cp.get("global", "debug_level")) > 0:
        print str(e[0]) + ": " + str(e[1])
        traceback.print_tb(e[2])
    else:
        print str(e[1])
    print

def tableType(tables, cursor):
    """
    Return the table type of the newly created table based on the 
    types of base tables. 
    Types:
    0 -> uncertain without confidence
    1 -> certain
    2 -> uncertain with confidence
    """
    types = []
    for table in tables:
        sql = "select type from trio_tables where tablename = '%s'" % table
        cursor.execute(sql)
        tup = cursor.fetchall()
        if len(tup) == 0:
            raise Exception("Table '%s' does not exist in Trio schema!" % table)
        elif len(tup) > 1:
            raise Exception("Trio schema inconsistent for table '%s'")
        types.append(tup[0][0])
    if types.count(TYPE_UNCERTAIN) > 0:
        return TYPE_UNCERTAIN
    elif types.count(TYPE_CONFIDENCE) > 0:
        return TYPE_CONFIDENCE
    elif types.count(TYPE_PROBABILISTIC) > 0:
        return TYPE_PROBABILISTIC
    else:
        return TYPE_CERTAIN

def checkAmbiguity(columnname, columns):
    matched = 0
    for c in columns:
        if (isinstance(c, Dot) and columnname == c.children[1].value) or \
           (isinstance(c, As) and isinstance(c.children[0], Dot) and columnname == c.children[0].children[1].value) or \
           (isinstance(c, As) and isinstance(c.children[0], Node) and columnname == c.children[0].value) or \
           (isinstance(c, As) and isinstance(c.children[0], Node) and columnname == c.children[1].value) or \
           (isinstance(c, Node) and columnname == c.value):
               matched += 1
    return matched
    
def toString(val):
    if isinstance(val, basestring):
        return "'" + val.replace("\\'", "'").replace("'", "\\'") + "'"
    elif val is None:
        return "null"
    else:
        return str(val)

def compare_values(tup1, tup2, datavals):
    for i in range(datavals):
        if tup1[i] < tup2[i]:
            return -1
        elif tup1[i] > tup2[i]:
            return 1
    return 0
    
    
def partition(list, p, r, sortByConf):
    i = p - 1
    j = r + 1
    x = list[(p + r)/2]
    while i < j:
        i = i + 1
        while (sortByConf and (list[i].getConfidence() > x.getConfidence()\
                               or (list[i].getConfidence() == x.getConfidence() and str(list[i]) < str(x))))\
               or\
               ((not sortByConf) and (str(list[i]) < str(x)\
                                         or (str(list[i]) == str(x) and list[i].getConfidence() > x.getConfidence()))):
            i = i + 1
        j = j - 1
        while (sortByConf and (list[j].getConfidence() < x.getConfidence()\
                               or (list[j].getConfidence() == x.getConfidence() and str(list[j]) > str(x))))\
               or\
               ((not sortByConf) and (str(list[j]) > str(x)\
                                         or (str(list[j]) == str(x) and list[j].getConfidence() < x.getConfidence()))):
            j = j - 1
        if i < j:
            temp = list[i]
            list[i] = list[j]
            list[j] = temp
    return j


def quicksort(list, p, r, sortByConf):
    if p < r:
        q = partition(list, p, r, sortByConf)
        quicksort(list, p, q, sortByConf)
        quicksort(list, q+1, r, sortByConf)

