__id__ = "$Id: createtable.py 574 2007-10-29 17:15:39Z theobald $"

import common
from node import *

def execute(node, trio_cursor, isConfBase=True, isTemporary=False):
    name = node.children[0].value
    description = [[x.children[0].value, x.children[1].value] for x in node.children[1].children]
    u_columnnames = [x.value for x in node.children[2].children]
    type = node.children[3].value
    if type == common.TYPE_CONFIDENCE and not u_columnnames: # and int(common.cp.get("global", "enable_prob")) == 1:
        type = common.TYPE_PROBABILISTIC
    
    # Basic integrity check for uncertain attributes
    if not set(u_columnnames).issubset(set([x[0] for x in description])):
        raise Exception("Uncertainty set must be subset of columnlist")
    
    # Create data tables and full range of system indexes when there is no subquery
    create_trio_table(name, type, description, u_columnnames, trio_cursor, isConfBase, isTemporary)
    create_trio_indexes(name, type, u_columnnames, trio_cursor, isTemporary)

def get_columns(type, description, u_columnnames):
    v_cols = ",".join([col[0] for col in description])
    c_cols = ",".join(x[0] + " " + x[1] for x in description if not x[0] in u_columnnames)
    u_cols = ",".join(x[0] + " " + x[1] for x in description if x[0] in u_columnnames)
    if type == common.TYPE_CERTAIN:
        v_cols += ",trio_xid,trio_xid as trio_aid,1 as trio_alts,1 as trio_enc,1.0 as trio_conf"
        c_cols += ",trio_xid int"
    elif type == common.TYPE_PROBABILISTIC:
        v_cols += ",trio_xid,trio_xid as trio_aid,1 as trio_alts,2 as trio_enc,trio_conf"
        c_cols += ",trio_xid int,trio_conf decimal"
    elif type == common.TYPE_UNCERTAIN:
        v_cols += ",c.trio_xid,trio_aid,trio_alts,trio_enc,null::decimal as trio_conf" 
        c_cols += (len(c_cols) > 0 and "," or "") + "trio_xid int,trio_alts int,trio_enc int"
        u_cols += (len(u_cols) > 0 and "," or "") + "trio_xid int,trio_aid int"
    elif type == common.TYPE_CONFIDENCE:
        v_cols += ",c.trio_xid,trio_aid,trio_alts,trio_enc,trio_conf" 
        c_cols += (len(c_cols) > 0 and "," or "") + "trio_xid int,trio_alts int,trio_enc int"
        u_cols += (len(u_cols) > 0 and "," or "") + "trio_xid int,trio_aid int,trio_conf decimal"
    else:
        raise Exception("Unknown table type")
    return [v_cols, c_cols, u_cols]

def create_rule(c, name, type):
    query = ""
    if type == common.TYPE_CONFIDENCE:
        query = """CREATE RULE "trio_conf_%s" AS ON UPDATE TO %s
                   DO INSTEAD
                   UPDATE trio_u_%s SET trio_conf = new.trio_conf
                   WHERE trio_aid = old.trio_aid""" % (name, name, name)
    else:
        query = """CREATE RULE "trio_conf_%s" AS ON UPDATE TO %s
                   DO INSTEAD
                   UPDATE trio_c_%s SET trio_conf = new.trio_conf
                   WHERE trio_xid = old.trio_xid""" % (name, name, name)
    c.execute(query)

def create_trio_table(name, type, description, u_columnnames, trio_cursor, isConfBase=False, isTemporary=False):
    tmpstr = isTemporary and "TEMPORARY" or ""
    c_table = "trio_c_" + name
    u_table = "trio_u_" + name
    
    i = 0
    while i < len(description):
        j = i + 1
        while j < len(description):
            if description[i][0] == description[j][0]:
                raise Exception("Duplicate or ambiguous column reference '%s'" % description[j][0])
            j += 1
        i += 1
    v_cols, c_cols, u_cols = get_columns(type, description, u_columnnames)
    c = trio_cursor.tdb.pdb.cursor()
    
    try:
        
        # DDL commands
        if not isTemporary:
            c.execute("CREATE TABLE %s (%s)" % (c_table, c_cols))
            if type != common.TYPE_CERTAIN and type != common.TYPE_PROBABILISTIC:
                c.execute("CREATE TABLE %s (%s)" % (u_table, u_cols))
                c.execute("""CREATE VIEW %s AS 
                     SELECT %s FROM %s u,%s c 
                     WHERE u.trio_xid=c.trio_xid 
                     ORDER BY c.trio_xid""" % (name, v_cols, u_table, c_table))
            else:
                c.execute("""CREATE VIEW %s AS 
                             SELECT %s FROM %s 
                             ORDER BY trio_xid""" % (name, v_cols, c_table))
        else:
            # CAUTION: temporary tables must use full range of Trio columns to work with 'DISTINCT'
            c.execute("CREATE " + tmpstr + " TABLE " + name + " (" + ", ".join(str(col[0]) + " " + str(col[1]) for col in description) \
                                      + ",trio_xid int,trio_aid int,trio_alts int,trio_conf decimal,trio_enc int)")
        
        # Lineage table
        c.execute("CREATE %s TABLE trio_lin_%s (aid int,disjlin int,srclin int,srctable name)" % (tmpstr, name))
        
        # Metadata for attributes/columns
        query = "INSERT INTO trio_attribs (tablename,attrname,colnum,iscertain) VALUES (%s,%s,%s,%s)"
        for i in range(len(description)):
            c.execute(query, (name, description[i][0], i, not description[i][0] in u_columnnames))
        
        # Metadata for tables
        query = "INSERT INTO trio_tables (tablename,type,confbase,istemporary) VALUES ('%s',%s,%s,%s)"
        c.execute(query % (name, type, isConfBase, isTemporary))
        
        # Identifier sequences
        c.execute("CREATE %s SEQUENCE trio_xid_%s" % (tmpstr, name))
        if type == common.TYPE_UNCERTAIN or type == common.TYPE_CONFIDENCE or isTemporary:
            c.execute("CREATE %s SEQUENCE trio_aid_%s" % (tmpstr, name))
            
        # Update rule for confidence memoization
        if (type == common.TYPE_CONFIDENCE or type == common.TYPE_PROBABILISTIC) and not isTemporary:
            create_rule(c, name, type)
            
    except Exception, e:
        if 'already exists' in str(e): 
            raise Exception("Table '%s' already exists in Trio schema!" % name)
        else:
            raise e
        
def create_trio_indexes(name, type, u_columnnames, trio_cursor, isTemporary=False):
    c = trio_cursor.tdb.pdb.cursor()
    if not isTemporary:
        # Data table indexes for certain attributes
        if type == common.TYPE_PROBABILISTIC:
            c.execute("CREATE INDEX trio_c_%s_xid_idx ON trio_c_%s(%s)" % (name, name, "trio_xid,trio_conf"))
        else:
            c.execute("CREATE INDEX trio_c_%s_xid_idx ON trio_c_%s(%s)" % (name, name, "trio_xid"))
            
        # Uncertain attributes index
        if type != common.TYPE_CERTAIN and type != common.TYPE_PROBABILISTIC:
            if type == common.TYPE_CONFIDENCE:
                c.execute("CREATE INDEX trio_u_%s_xid_idx ON trio_u_%s(%s)" % (name, name, "trio_xid,trio_conf"))
                c.execute("CREATE INDEX trio_u_%s_aid_idx ON trio_u_%s(%s)" % (name, name, "trio_aid,trio_conf"))
            else:
                c.execute("CREATE INDEX trio_u_%s_xid_idx ON trio_u_%s(%s)" % (name, name, "trio_xid"))
                c.execute("CREATE INDEX trio_u_%s_aid_idx ON trio_u_%s(%s)" % (name, name, "trio_aid"))
                
    else:
        c.execute("CREATE INDEX trio_%s_xid_idx ON %s(%s)" % (name, name, "trio_xid,trio_conf"))
        c.execute("CREATE INDEX trio_%s_aid_idx ON %s(%s)" % (name, name, "trio_aid,trio_conf"))
        
    # Lineage index
    c.execute("CREATE INDEX trio_lin_%s_idx ON trio_lin_%s(%s)" % (name, name, "aid,srctable,srclin,disjlin"))
    
def update_trio_sequences(name, u_columnnames, trio_cursor, isTemporary=False, create=False):
    tmpstr = isTemporary and "TEMPORARY" or ""
    max_xid = "1"
    max_aid = "1"
    c = trio_cursor.tdb.pdb.cursor()
    if not create:
        c.execute("DROP SEQUENCE trio_xid_%s" % name)
        if trio_cursor.tableType != common.TYPE_CERTAIN or isTemporary:
            c.execute("DROP SEQUENCE trio_aid_%s" % name)
        c.execute("SELECT max(trio_xid)+1, max(trio_aid)+1 FROM %s" % name)
        [max_xid, max_aid] = c.fetchall()[0]
        if max_xid is None:
            max_xid = "1"
        if max_aid is None:
            max_aid = "1"
            
    # Identifier sequences
    c.execute("CREATE %s SEQUENCE trio_xid_%s START %s" % (tmpstr, name, max_xid))
    if trio_cursor.tableType == common.TYPE_UNCERTAIN or trio_cursor.tableType == common.TYPE_CONFIDENCE or isTemporary:
        c.execute("CREATE %s SEQUENCE trio_aid_%s START %s" % (tmpstr, name, max_aid))

