__id__ = "$Id: executor.py 468 2007-05-25 05:27:00Z theobald $"

import createindex
import createtable
import createtableas
import dropindex
import droptable
import insert
import select
import command
import setoperator
from node import *
from visitor import triql_printer

class Executor:
    def __init__(self, trio_cursor):
        self.tcursor = trio_cursor

    def execute(self, node):
        nodetype = node.__class__.__name__
        executefn = getattr(self, 'execute'+nodetype, self.executeDefault)
        executefn(node)

    def executeDefault(self, node):
        raise Exception("Unsupported command: %s" % node.__class__.__name__)

    def executeCreateIndex(self, node):
        createindex.execute(node, self.tcursor)

    def executeCreateTable(self, node): 
        createtable.execute(node, self.tcursor)

    def executeCreateTempTable(self, node):
        createtable.execute(node, self.tcursor, isConfBase=False, isTemporary=True)
        
    def executeCreateTableAs(self, node): 
        createtableas.execute(node, self.tcursor)

    def executeCreateTempTableAs(self, node): 
        createtableas.execute(node, self.tcursor, isTemporary=True)

    def executeDropIndex(self, node):
        dropindex.execute(node, self.tcursor)

    def executeDropTable(self, node): 
        droptable.execute(node, self.tcursor)

    def executeViewTable(self, node): 
        viewtable.execute(node, self.tcursor)

    def executeInsert(self, node):
        if not isinstance(node.children[1], Empty): 
            columnList = [node.children[1].children[i].value for i in range(len(node.children[1].children))]
            insert.InsertExecutor(node.children[0].value, self.tcursor, columnList).execute(node, self.tcursor)
        else:
            insert.InsertExecutor(node.children[0].value, self.tcursor).execute(node, self.tcursor)

    def executeSelect(self, node):
        select.execute(node, self.tcursor)

    def executeSetOperator(self, node):
        setoperator.execute(node, self.tcursor)
        
    def executeCommand(self, node):
        command.CommandExecutor().execute(node, self.tcursor)
