__id__ = "$Id: select.py 572 2007-10-22 23:53:25Z theobald $"

from node import *
from visitor import rewriter
from visitor import printer
from visitor import dumper
import common
import selectdistinct

def execute(node, trio_cursor, createtable=False):
    # Rewrite TriQL to SQL
    r = rewriter.Rewriter(trio_cursor)
    new_node = r.visit(node)
    if isinstance(node, SetOperator):
        new_node = new_node.children[0]
        
    # Set conf flags, e.g., a derived green table can become a confBase but not have asConf
    trio_cursor.isConfBase = new_node.isConfBase
    trio_cursor.gotAsConf = new_node.gotAsConf
    trio_cursor.gotUniform = new_node.gotUniform
    trio_cursor.u_columns = new_node.u_columns
    
    # Initialize the Trio cursor and determine various fetch options
    trio_cursor.root_node = new_node
    trio_cursor.num_joins = len(new_node.children[2].children)
    
    # The result of an approximate aggregate query is a blue table
    if new_node.gotAgg == 1:
        trio_cursor.tableType = common.TYPE_CERTAIN
        trio_cursor.num_joins = 1
    elif new_node.gotUniformT == 1:
        trio_cursor.tableType = common.TYPE_CONFIDENCE
    elif trio_cursor.gotAsConf:
        trio_cursor.tableType = common.TYPE_CONFIDENCE
    else:
        trio_cursor.tableType = common.tableType(new_node.names, trio_cursor.tdb.pdb.cursor())
    trio_cursor.num_datavalues = len(new_node.children[1].children) - (2*trio_cursor.num_joins) - 3
    
    for i in range(len(new_node.children[0].children)):
        
        # Apply distinct modifier
        if new_node.children[0].children[i].value == 'distinct':
            if trio_cursor.groupAlts < 0:
                if trio_cursor.distinct:
                    raise Exception("Only one 'Distinct' clause per select statement allowed!")
                else:
                    trio_cursor.distinct = True
            else:
                trio_cursor.merged = False
        
        # Apply merged modifier
        elif new_node.children[0].children[i].value == 'merged':
            if not trio_cursor.distinct:
                if trio_cursor.merged:
                    raise Exception("Only one 'Merged' clause per select statement allowed!")
                else:
                    trio_cursor.merged = True
        
        # Apply flatten modifier
        elif new_node.children[0].children[i].value == 'flatten':
            if trio_cursor.flatten:
                raise Exception("Only one 'Flatten' clause per select statement allowed!")
            elif trio_cursor.groupAlts != -1:
                raise Exception("'GroupAlts' not allowed in conjunction with 'Flatten'!")
            else:
                trio_cursor.u_columns = []
                trio_cursor.flatten = True
                if trio_cursor.nomaybe:
                    trio_cursor.tableType = common.TYPE_CERTAIN
        
        # Apply nolineage modifier
        elif new_node.children[0].children[i].value == 'nolineage':
            if trio_cursor.nolineage:
                raise Exception("Only one 'NoLineage' clause per select statement allowed!")
            else:
                trio_cursor.nolineage = True
                trio_cursor.isConfBase = True
        
        # Apply nomaybe modifier
        elif new_node.children[0].children[i].value == 'nomaybe':
            if trio_cursor.nomaybe:
                raise Exception("Only one 'NoMaybe' clause per select statement allowed!")
            else:
                if trio_cursor.gotAsConf:
                    raise Exception("'NoMaybe' not allowed in conjunction with 'As Conf'!")
                trio_cursor.nomaybe = True
                if trio_cursor.tableType == common.TYPE_CONFIDENCE:
                    trio_cursor.tableType = common.TYPE_UNCERTAIN
                if trio_cursor.flatten:
                    trio_cursor.tableType = common.TYPE_CERTAIN
                    
        # Apply noconf modifier
        elif new_node.children[0].children[i].value == 'noconf':
            if trio_cursor.noconf:
                raise Exception("Only one 'NoConf' clause per select statement allowed!")
            else:
                if not trio_cursor.nolineage and trio_cursor.gotAsConf:
                    raise Exception("'NoConf' clause not allowed in conjunction with 'As Conf' and 'Compute Confidences'!")
                confNode = As(children=[Literal(None), Node("trio_conf")])
                new_children = []
                for child in new_node.children[1].children:
                    if isinstance(child, As) and child.children[1].value == "trio_conf":
                        new_children.append(confNode)
                    else:
                        new_children.append(child)
                new_node.children[1].children = new_children
                trio_cursor.gotAsConf = False
                trio_cursor.noconf = True
                if trio_cursor.tableType == common.TYPE_CONFIDENCE:
                    trio_cursor.tableType = common.TYPE_UNCERTAIN
        
        # Apply groupalts modifier
        elif isinstance(new_node.children[0].children[i], GroupAlts):
            if trio_cursor.groupAlts > -1:
                raise Exception("Only one 'GroupAlts' clause per select statement allowed!")
            elif trio_cursor.flatten:
                raise Exception("'Flatten' not allowed in conjunction with 'GroupAlts'!")
            else:
                # Update modifiers
                if trio_cursor.distinct:
                    trio_cursor.merged = True
                    trio_cursor.distinct = False
                if not trio_cursor.gotAsConf and trio_cursor.tableType == common.TYPE_CONFIDENCE:
                    trio_cursor.tableType = common.TYPE_UNCERTAIN
                trio_cursor.isConfBase = True
                
                groupAltsClause = ""
                p = printer.Printer()
                trio_cursor.u_columns = []
                if isinstance(new_node.children[0].children[i].children[0].children[0], Star):
                    j = 0
                    for select_column in new_node.children[1].children:
                        if isinstance(select_column, As):
                            groupAltsClause += '(' + p.visit(select_column.children[0]) + ')'
                        else:
                            groupAltsClause += '(' + p.visit(select_column) + ')'
                        j += 1
                        if j < trio_cursor.num_datavalues:
                            groupAltsClause += "||\'_\'||"
                        if j == trio_cursor.num_datavalues:
                            break
                else:
                    c_columns = []
                    j = 0
                    for groupAlts_column in new_node.children[0].children[i].children[0].children:
                        k = 0
                        matched = False
                        for select_column in new_node.children[1].children:
                            if (isinstance(select_column, As) and \
                                common.checkAmbiguity(groupAlts_column.value, new_node.children[1].children) == 1 and \
                                groupAlts_column.value == select_column.children[1].value) or \
                               (isinstance(select_column, As) and \
                                isinstance(select_column.children[0], Dot) and \
                                isinstance(groupAlts_column, Dot) and \
                                groupAlts_column.children[0].value == select_column.children[0].children[0].value and \
                                groupAlts_column.children[1].value == select_column.children[0].children[1].value) or \
                               (isinstance(select_column, As) and \
                                isinstance(select_column.children[0], Dot) and \
                                common.checkAmbiguity(groupAlts_column.value, new_node.children[1].children) == 1 and \
                                groupAlts_column.value == select_column.children[0].children[1].value) or \
                               (isinstance(select_column, As) and \
                                common.checkAmbiguity(groupAlts_column.value, new_node.children[1].children) == 1 and \
                                groupAlts_column.value == select_column.children[0].value):
                                    if k in c_columns:
                                        raise Exception("Repeated column reference in GroupAlts clause!")
                                    groupAltsClause += '(' + p.visit(select_column.children[0]) + ')'
                                    c_columns.append(k)
                                    matched = True
                                    break
                                        
                            if (isinstance(select_column, Dot) and \
                                isinstance(groupAlts_column, Dot) and \
                                groupAlts_column.children[0].value == select_column.children[0].value and \
                                groupAlts_column.children[1].value == select_column.children[1].value) or \
                               (isinstance(select_column, Dot) and \
                                common.checkAmbiguity(groupAlts_column.value, new_node.children[1].children) == 1 and \
                                groupAlts_column.value == select_column.children[1].value) or \
                               ((not isinstance(select_column, As)) and \
                                (not isinstance(select_column, Dot)) and \
                                (not isinstance(groupAlts_column, As)) and \
                                (not isinstance(groupAlts_column, Dot)) and \
                                groupAlts_column.value is not None and \
                                select_column.value is not None and \
                                groupAlts_column.value == select_column.value):
                                    if k in c_columns:
                                        raise Exception("Repeated column reference in GroupAlts clause!")
                                    groupAltsClause += '(' + p.visit(select_column) + ')'
                                    c_columns.append(k)
                                    matched = True
                                    break
                            k += 1
                            
                        if not matched:
                            raise Exception('Unresolvable column reference in GroupAlts!')
                        if j < len(new_node.children[0].children[i].children[0].children) - 1:
                            groupAltsClause += "||\'_\'||"
                        j += 1
                        
                    # Conservatively make all remaining columns as uncertain
                    for j in range(trio_cursor.num_datavalues):
                        if j not in c_columns:
                            trio_cursor.u_columns.append(j)
                        
                trio_cursor.groupAlts = len(new_node.children[1].children)
                new_node.children[1].children.append(As(children=[Node(groupAltsClause), Node("trio_groupalts")]))

    # Invoke conf(*) for orange tables if isConfBase
    if trio_cursor.isConfBase and not trio_cursor.noconf and not trio_cursor.gotAsConf and trio_cursor.groupAlts == -1 and \
       new_node.gotUniformT != 1 and trio_cursor.tableType == common.TYPE_CONFIDENCE:
        r.context.append(new_node)
        confNode = As(children=[r.visit(Conf(children=[Star()])), Node("trio_conf")])
        r.context.pop()
        new_children = []
        
        # Replace former trio_conf
        for child in new_node.children[1].children:
            if isinstance(child, As) and child.children[1].value == "trio_conf":
                new_children.append(confNode)
            else:
                new_children.append(child)
        new_node.children[1].children = new_children
        trio_cursor.gotAsConf = True
        
    sql = printer.Printer(trio_cursor).visit(new_node)
    if int(common.cp.get("global", "debug_level")) > 0:
        if int(common.cp.get("global", "debug_level")) > 2:
            dumper.Dumper().visit(new_node)
        print sql
    
    c = trio_cursor.tdb.pdb.cursor()
    if (createtable and int(common.cp.get("global", "use_SPI")) != 0) or trio_cursor.distinct:
        # For SPI: run a dummy query in Python first to get the cursor metadata
        whereclause = new_node.children[3]
        new_node.children[3] = WhereClause(children=[BinaryExpression("=", [Literal(1), Literal(0)])])
        sql2 = printer.Printer().visit(new_node)
        new_node.children[3] = whereclause
        c.execute(sql2)
    else:
        c.execute(sql)
    try:
        trio_cursor.isViewTable = new_node.isViewTable
        trio_cursor.hasExplainLineage = new_node.hasExplainLineage
    except:
        trio_cursor.isViewTable = False
        trio_cursor.hasExplainLineage = False
        
    trio_cursor.cursor = c
    trio_cursor.description = c.description[:trio_cursor.num_datavalues]
    if new_node.gotAgg == 1:
        trio_cursor.nolineage = True
    if trio_cursor.distinct:
        sql, trio_cursor = selectdistinct.execute(new_node, trio_cursor, sql)
    trio_cursor.isXCursor = True
    trio_cursor.isFetchable = True
    
    return sql

