from node import *
import common

# SELECT DISTINCT
# Caution: - this algorithm subsumes horizontal merging
#          - merely assumes that at most all alternatives of an xtuple fit in memory at a time
#          - supports on-the-fly confidence computation
#          - for persistent queries, a second scan is required to get the ? right
def execute(node, trio_cursor, sql):
    gotUniform = trio_cursor.gotUniform
    tableType = trio_cursor.tableType
    isConfBase = trio_cursor.isConfBase
    gotAsConf = trio_cursor.gotAsConf
    
    flatten = trio_cursor.flatten
    noconf = trio_cursor.noconf
    nomaybe = trio_cursor.nomaybe
    nolineage = trio_cursor.nolineage
    u_columns = trio_cursor.u_columns
    
    cursor1 = trio_cursor.tdb.ncursor()
    cursor2 = trio_cursor.tdb.ncursor()
    
    columnnames = []
    for i in range(len(trio_cursor.description)):
        columnnames.append(trio_cursor.description[i][0])
        
    # Get a unique temporary table name
    cursor1.execute("SELECT nextval('trio_tmp')")
    tmp_table = "tmp_distinct_" + str(cursor1.fetchone()[0])
    
    # Initialize a temporary result table
    #try:
    #    trio_cursor.execute("DROP TABLE " + tmp_table);
    #    trio_cursor.tdb.commit()
    #except:
    #   pass
    
    # Create the temporary table
    trio_cursor.execute("CREATE TEMPORARY TABLE " + tmp_table \
                        + " (" + ",".join("%s %s" % (t[0], t[1]) for t in trio_cursor.description) + ")")
    cursor1.execute("UPDATE trio_tables SET confbase=FALSE WHERE tablename = '%s'" % tmp_table)
    
    # SPI fetch&deduplication
    if int(common.cp.get("global", "use_SPI")) == 1:

        tables = "|".join("%s" % name for name in node.names) + "|"
        cursor = trio_cursor.tdb.ncursor()
        columns = "|".join(x[0] for x in trio_cursor.description) + "|"
        cursor.execute("SELECT * FROM trio_selectinto_distinct($$%s$$, $$%s$$, $$%s$$, $$%s$$, %d)"\
            % (tmp_table, sql, columns, tables, isinstance(node.children[4], ComputeConfidences)))
        if cursor.fetchone()[0] == 0:
            raise Exception("SPI call failed!")
    
    # Python fetch
    else:
        
        # Initialize data & lineage statements
        gotConfStar = False
        if isinstance(node.children[4], ComputeConfidences):
            gotConfStar = True
            insert_sql = "INSERT INTO " + tmp_table + " (" + (",".join("%s" % t[0] for t in trio_cursor.description))\
                                    + ",trio_aid,trio_xid,trio_alts,trio_enc,trio_conf) VALUES ("\
                                    + bindparamstr(trio_cursor.num_datavalues)\
                                    + ",%d,%d,%d,%d,trio_get_conf(array['" + tmp_table + "'], array[%d], array[1], NULL, 0, 1))"
        else:
            insert_sql = "INSERT INTO " + tmp_table + " (" + (",".join("%s" % t[0] for t in trio_cursor.description))\
                                    + ",trio_aid,trio_xid,trio_alts,trio_enc,trio_conf) VALUES ("\
                                    + bindparamstr(trio_cursor.num_datavalues)\
                                    + ",%d,%d,%d,%d,%s)"
        insert_lin_sql = "INSERT INTO trio_lin_" + tmp_table + " (aid, disjlin, srclin, srctable) VALUES (%d,%d,%d,'%s')"
        
        # Initialize xid's of input tables
        aid = 1
        xid = 0
        max_xids = []
        i = 0
        for table in node.names:
            cursor1.execute("SELECT nextval('trio_xid_%s')" % table)
            max_xids.append(cursor1.fetchone()[0])
            xid += pow(max_xids[i], i+1)
            i += 1
            
        # Run the query sorted by data columns & xid's
        cursor1.execute(sql)
        tup = cursor1.fetchone()
        
        # Read alts sequentially, first buffer alts with same data and xid
        # - when xid changes, flush lineage into new xtuple & stop buffering, else keep xid
        # - when data changes, flush distinct data with new aid each
        disj = 0
        do_buffer = True
        alts_buffer = []
        while tup != None:
            tmp = tup
            tmp_xid = getXid(tmp[trio_cursor.num_datavalues+trio_cursor.num_joins:trio_cursor.num_datavalues+(2*trio_cursor.num_joins)], max_xids)
            tup = cursor1.fetchone()
            if tup is None or common.compare_values(tmp, tup, trio_cursor.num_datavalues) != 0:
                # flush lineage buffer
                alts_buffer += [tmp]
                for alt in alts_buffer:
                    for lin in getLin(alt, node.names, trio_cursor):
                        cursor2.execute(insert_lin_sql % (aid, disj, lin[0], lin[1]))
                    disj += 1
                alts_buffer = []
                
                # insert data
                cols = tmp[:trio_cursor.num_datavalues]
                cols += [aid, int(do_buffer and tmp_xid or xid),\
                         tmp[trio_cursor.num_datavalues+(2*trio_cursor.num_joins)+2],\
                         tmp[trio_cursor.num_datavalues+(2*trio_cursor.num_joins)]]
                if gotConfStar:
                    cols += [aid]
                else:
                    cols += [tmp[trio_cursor.num_datavalues+(2*trio_cursor.num_joins)+1]]
                cursor2.execute(insert_sql, cols)
                do_buffer = True
                disj = 0
                
                # create new aid
                aid += 1
                
            else:
                
                tup_xid = getXid(tup[trio_cursor.num_datavalues+trio_cursor.num_joins:trio_cursor.num_datavalues+(2*trio_cursor.num_joins)], max_xids)
                if do_buffer and tmp_xid == tup_xid:
                    # buffer lineage
                    alts_buffer += [tmp]
                else:
                    if do_buffer:
                        # flush lineage buffer
                        alts_buffer += [tmp]
                        for alt in alts_buffer:
                            for lin in getLin(alt, node.names, trio_cursor):
                                cursor2.execute(insert_lin_sql % (aid, disj, lin[0], lin[1]))
                            disj += 1
                        alts_buffer = []
                        do_buffer = False
                        # create new xid
                        xid += 1
                    else:
                        # flush lineage w/out buffering
                        for lin in getLin(tmp, node.names, trio_cursor):
                            cursor2.execute(insert_lin_sql % (aid, disj, lin[0], lin[1]))
                        disj += 1
                        
    trio_cursor.tdb.commit()
    cursor1.close()
    cursor2.close()
    
    # Initialize SQL cursor on temporary table; this pass also fixes ? and reassigns contiguous XID's
    sql = "SELECT " + ",".join(x for x in columnnames) \
                               + ", trio_aid, trio_xid, trio_alts, trio_conf, trio_enc FROM " \
                               + tmp_table + " ORDER BY trio_xid"
    trio_cursor.cursor.execute(sql)
    
    trio_cursor.gotUniform = gotUniform
    trio_cursor.tableType = tableType
    trio_cursor.isConfBase = isConfBase
    trio_cursor.gotAsConf = gotAsConf
    
    trio_cursor.flatten = flatten
    trio_cursor.merged = False
    trio_cursor.noconf = noconf
    trio_cursor.nomaybe = nomaybe
    trio_cursor.nolineage = nolineage
    
    trio_cursor.distinct = True
    trio_cursor.distinctTable = tmp_table
    trio_cursor.num_joins = 1
    trio_cursor.u_columns = u_columns
    
    return sql, trio_cursor

def getLin(tup, tables, trio_cursor):
    lin = []
    for i in range(len(tables)):
        lin.append([tup[trio_cursor.num_datavalues+i], tables[i]])
    return lin
        
def getXid(xids, max_xids):
    xid = 0
    for i in range(len(xids)):
        xid += xids[i] * pow(max_xids[i],i)
    return xid

def bindparamstr(numparams):
    return ",".join(["%s" for x in range(numparams)])
