__id__ = "$Id: setoperator.py 322 2007-03-02 00:34:45Z theobald $"

from node import *
import createtable
from visitor import rewriter
from visitor import printer
import common

def execute(node, trio_cursor, tablename=None):
    gotUniform = trio_cursor.gotUniform
    tableType = trio_cursor.tableType
    isConfBase = trio_cursor.isConfBase
    gotAsConf = trio_cursor.gotAsConf
    
    isIntersect = False
    isExcept = False
    isUnion = False
    modifiedConfBase = False
    
    # Get a unique temporary table name
    cursor3 = trio_cursor.tdb.ncursor()
    cursor3.execute("SELECT nextval('trio_tmp')")
    tmp_c = str(cursor3.fetchone()[0])
    
    # Determine lineage encoding
    delta_enc = 0
    if node.value == "intersect":
        isIntersect = True
        delta_enc = 2  # positive lineage branch
    elif node.value == "except":
        isExcept = True
        delta_enc = 3 # negative lineage branch
    else:
        isUnion = True
        delta_enc = 1 # default delta for DNF lineage
        
    # Determine result table
    persistent = True
    if tablename is None:
        persistent = False
        node = rewriter.Rewriter(trio_cursor).visit(node)
        tablename = "tmp_diff_" + tmp_c
        #try:
        #    trio_cursor.execute("DROP TABLE " + tablename);
        #    trio_cursor.tdb.commit()
        #except:
        #    pass
            
    # Check select statements and enforce aliases
    select_cols1 = []
    select_cols2 = []
    for child in node.children[0].children[1].children:
        if isinstance(child, As) and child.children[1].value.startswith("trio_"):
            break
        if isinstance(child, As) or isinstance(child, Dot):
            select_cols1 += [child.children[1]]
        elif len(child.children) > 0:
            raise Exception("Alias required for expressions in UNION/EXCEPT/INTERSECT subquery!")
        else:
            select_cols1 += [child]
    for i in range(len(select_cols1)):
        j = i+1
        while j < len(select_cols1):
            if select_cols1[i] == select_cols1[j]:
                raise Exception("Ambiguous column reference '" + select_cols1[i] + "' in UNION/EXCEPT/INTERSECT subquery!")
            j += 1
    for child in node.children[1].children[1].children:
        if isinstance(child, As) and child.children[1].value.startswith("trio_"):
            break
        if isinstance(child, As) or isinstance(child, Dot):
            select_cols2 += [child.children[1]]
        elif len(child.children) > 0:
            raise Exception("Alias required for expressions in UNION/EXCEPT/INTERSECT subquery!")
        else:
            select_cols2 += [child]
    for i in range(len(select_cols2)):
        j = i+1
        while j < len(select_cols2):
            if select_cols2[i] == select_cols2[j]:
                raise Exception("Ambiguous column reference '" + select_cols2[i] + "' in UNION/EXCEPT/INTERSECT subquery!")
            j += 1
            
    # Basic integrity check
    if len(select_cols1) != len(select_cols2):
        raise Exception("Each UNION/EXCEPT/INTERSECT subquery must have the same number of columns!")
    datavals = len(select_cols1)
    trio_cursor.setoperator = True
    
    tableTypes = []
    if node.value == "union all":
        # Union-all
        if not persistent:
            trio_cursor.execute("CREATE TEMPORARY TABLE " + tablename + " AS " + node.leftquery)
            tableTypes.append(trio_cursor.tableType)
            isConfBase = trio_cursor.isConfBase
        else:
            trio_cursor.execute("INSERT INTO " + tablename + " " + node.leftquery)
            tableTypes.append(trio_cursor.tableType)
            isConfBase = trio_cursor.isConfBase
        u_columns = trio_cursor.u_columns
        
        trio_cursor.execute("INSERT INTO " + tablename + " " + node.rightquery)
        tableTypes.append(trio_cursor.tableType)
        
        if not set(trio_cursor.u_columns).issubset(set(u_columns)):
            raise Exception("Uncertain attributes in right subquery must be a subset of "\
                            "left subquery's uncertain attributes for UNION/UNION ALL!")
        
        # Advanced integrity check
        isConfBase = isConfBase or trio_cursor.isConfBase
        if tableTypes.count(common.TYPE_UNCERTAIN) > 0:
            tableType = common.TYPE_UNCERTAIN
        elif tableTypes.count(common.TYPE_CONFIDENCE) > 0:
            tableType = common.TYPE_CONFIDENCE
        elif tableTypes.count(common.TYPE_PROBABILISTIC) > 0:
            tableType = common.TYPE_PROBABILISTIC
        else:
            tableType = common.TYPE_CERTAIN
        trio_cursor.setoperator = False
         
    else:
        
        # Union/Intersect/Except
        #try:
        #    trio_cursor.execute("DROP TABLE tmp_diff_1_" + tmp_c);
        #    trio_cursor.tdb.commit()
        #except:
        #    pass
        #try:
        #    trio_cursor.execute("DROP TABLE tmp_diff_2_" + tmp_c);
        #    trio_cursor.tdb.commit()
        #except:
        #    pass
        
        # Create temporary tables for left and right subqueries, use left query for metadata
        trio_cursor.u_columns = []
        trio_cursor.execute("CREATE TEMPORARY TABLE tmp_diff_1_" + tmp_c + " AS " + node.leftquery)
        nolineage1 = trio_cursor.nolineage
        tableTypes.append(trio_cursor.tableType)
        isConfBase = trio_cursor.isConfBase
        description = trio_cursor.description
        u_columns = trio_cursor.u_columns
        
        trio_cursor.execute("CREATE TEMPORARY TABLE tmp_diff_2_" + tmp_c + " AS " + node.rightquery)
        tableType2 = trio_cursor.tableType
        nolineage2 = trio_cursor.nolineage
        tableTypes.append(trio_cursor.tableType)
        
        # Check uncertain sets
        if isUnion and not set(trio_cursor.u_columns).issubset(set(u_columns)):
            raise Exception("Uncertain attributes in right subquery must be a subset of "\
                            "left subquery's uncertain attributes for UNION/UNION ALL!")
            
        # Advanced integrity check
        isConfBase = isConfBase or trio_cursor.isConfBase
        if tableTypes.count(common.TYPE_UNCERTAIN) > 0:
            tableType = common.TYPE_UNCERTAIN
        elif tableTypes.count(common.TYPE_CONFIDENCE) > 0:
            tableType = common.TYPE_CONFIDENCE
        elif tableTypes.count(common.TYPE_PROBABILISTIC) > 0:
            tableType = common.TYPE_PROBABILISTIC
        else:
            tableType = common.TYPE_CERTAIN
            
        if u_columns != trio_cursor.u_columns:
            raise Exception("Each UNION/EXCEPT/INTERSECT subquery must have the same uncertainty set!")
          
        gotConfStar = isinstance(node.children[2], ComputeConfidences)
        if tableType != common.TYPE_PROBABILISTIC and \
           tableType != common.TYPE_CONFIDENCE and \
           gotConfStar:
            raise Exception("Conf() called on a table with no confidences")
        if gotConfStar and nolineage1 and nolineage2:
            raise Exception("Conf() not allowed on top of a table with no lineage!")
            
        # Create a new temporary table if not anyway persistent
        if not persistent:
            createtable.create_trio_table(tablename, tableType, description, 
                                          [description[i][0] for i in u_columns], 
                                          trio_cursor, isConfBase, True)
                                          
        # Restor cursor settings after running the temp queries
        trio_cursor.distinct = False
        trio_cursor.tableType = tableType
        trio_cursor.num_datavalues = datavals
        trio_cursor.u_columns = u_columns
        trio_cursor.gotUniform = gotUniform
        trio_cursor.tableType = tableType
        trio_cursor.isConfBase = isConfBase
        trio_cursor.gotAsConf = gotAsConf
        trio_cursor.setoperator = False
        trio_cursor.isTemporary = not persistent
        if not persistent:
            trio_cursor.distinctTable = tablename
            
        trio_cursor.flatten = False
        trio_cursor.noconf = False
        trio_cursor.nomaybe = False
        nolineage = trio_cursor.False
        
        cursor1 = trio_cursor.tdb.ncursor()
        cursor2 = trio_cursor.tdb.ncursor()
        
        if isConfBase:
            cursor1.execute("UPDATE trio_tables SET confbase=FALSE WHERE tablename = '" + tablename + "'")
            modifiedConfBase = True
            
        # Merge-join two SQL result sets (each itself a join of a data and lineage table)
        select_sql1 = ",".join(printer.Printer().visit(s) for s in select_cols1)
        select_sql2 = ",".join(printer.Printer().visit(s) for s in select_cols2)
        
        query1 = "SELECT " + select_sql1 + ",trio_aid,trio_xid,trio_alts,trio_enc,trio_conf,disjlin,srclin,srctable FROM tmp_diff_1_"\
                                  + tmp_c + ", trio_lin_tmp_diff_1_" + tmp_c + " WHERE trio_aid = aid ORDER BY "\
                                  + select_sql1 + ",trio_aid asc, disjlin asc, srclin asc, srctable"
        query2 = "SELECT " + select_sql2 + ",trio_aid,trio_xid,trio_alts,trio_enc,trio_conf,disjlin,srclin,srctable FROM tmp_diff_2_"\
                                  + tmp_c + ",trio_lin_tmp_diff_2_" + tmp_c + " WHERE trio_aid = aid ORDER BY "\
                                  + select_sql2 + ",trio_aid asc, disjlin asc, srclin asc, srctable"
                                  
        if int(common.cp.get("global", "use_SPI")) == 1:

            # SPI fetch&merge
            conf_option = 0
            if trio_cursor.noconf:
                conf_option = -1
            if trio_cursor.gotAsConf:
                conf_option = 1
            if trio_cursor.gotUniform:
                conf_option = 2
            setop_option = 0
            if isIntersect:
                setop_option = 1
            elif isExcept:
                setop_option = 2
            columns = "|".join(x[0] for x in trio_cursor.description) + "|"
            coltypes = "|".join(i in trio_cursor.u_columns and "u" or "c" for i in range(trio_cursor.num_datavalues)) + "|"
            type = tableType
            if not persistent:
                type = -1
            cursor = trio_cursor.tdb.ncursor()
            cursor.execute("SELECT * FROM trio_mergeinto_distinct($$%s$$, $$%s$$, $$%s$$, $$%s$$, $$%s$$, %d, %d, %d, %d, %d, %d)"\
                 %(tablename, query1, query2, columns, coltypes, setop_option, type, nolineage1, nolineage2, tableType2, gotConfStar))
            if cursor.fetchone()[0] == 0:
                raise Exception("SPI call failed!")
         
        else:
                                  
            # Initialize system id's
            [xid, aid] = [1, 1]
            if isUnion:
                cursor1.execute("SELECT nextval('trio_xid_tmp_diff_1_" + tmp_c + "'), nextval('trio_aid_tmp_diff_1_" + tmp_c + "')")
                tup = cursor1.fetchone()
                if tup is not None:
                    xid = int(tup[0])
                    aid = int(tup[1])
                    
            if not persistent:
                insert_c_sql = None
                if not isinstance(node.children[2], ComputeConfidences):
                    insert_u_sql = "INSERT INTO " + tablename + " (" + ",".join(x[0] for x in description)\
                                                  + ",trio_xid,trio_aid,trio_alts,trio_enc) VALUES ("\
                                                  + bindparamstr(datavals) + ",%d,%d,%d,%d)"
                else:
                    gotConfStar = True
                    insert_u_sql = "INSERT INTO " + tablename + " (" + ",".join(x[0] for x in description)\
                                                  + ",trio_xid,trio_aid,trio_alts,trio_enc,trio_conf) VALUES ("\
                                                  + bindparamstr(datavals) + ",%d,%d,%d,%d,trio_get_conf(array['"\
                                                  + tablename + "'],array[%d],array[1],NULL,0,1))"
            else:
                insert_c_sql = "INSERT INTO trio_c_" + tablename + " ("
                insert_u_sql = "INSERT INTO trio_u_" + tablename + " ("
                c = len(description) - len(u_columns)
                u = len(u_columns)
                for i in range(len(description)):
                    if i in u_columns:
                        insert_u_sql += description[i][0] + ","
                    else:
                        insert_c_sql += description[i][0] + ","
                if tableType == common.TYPE_CERTAIN or (tableType == common.TYPE_PROBABILISTIC and not gotConfStar):
                        insert_c_sql += "trio_xid) VALUES (" + bindparamstr(c) + ",%d)"
                elif tableType == common.TYPE_PROBABILISTIC:
                        insert_c_sql += "trio_xid,trio_conf) VALUES (" + bindparamstr(c) + ",%d,%d)"
                else:
                    insert_c_sql += "trio_xid,trio_alts,trio_enc) VALUES (" + bindparamstr(c) + (c>0 and "," or "") + "%d,%d,%d)"
                    if tableType == common.TYPE_UNCERTAIN or\
                      (tableType == common.TYPE_CONFIDENCE and not gotConfStar):
                        insert_u_sql += "trio_xid,trio_aid) VALUES (" + bindparamstr(u) + (u>0 and "," or "") + "%d,%d)"
                    else:
                        insert_u_sql += "trio_xid,trio_aid,trio_conf) VALUES ("\
                                     + bindparamstr(u) + (u>0 and "," or "")\
                                     + "%d,%d,trio_get_conf(array['" + tablename + "'],array[%d],array[1],NULL,0,1))"
            insert_lin_sql = "INSERT INTO trio_lin_" + tablename + " (aid, disjlin, srclin, srctable) VALUES (%d,%d,%d,'%s')"
                                
            # Merge algorithm
            cursor1.execute(query1)
            cursor2.execute(query2)
            tup1 = cursor1.fetchone()
            tup2 = cursor2.fetchone()
            while tup1 is not None or (isUnion and tup2 is not None):
                skipCertain = False
                
                # Move cursor sequentially through left-hand result set, insert data & lineage
                while tup1 is not None:
                    if tup2 is not None and common.compare_values(tup1, tup2, datavals) >= 0:
                        break
                    tmp = tup1
                    tup1 = cursor1.fetchone()
                    if isUnion or isExcept:
                        if not nolineage1:
                            cursor3.execute(insert_lin_sql % (tmp[datavals], tmp[datavals+5], tmp[datavals+6], tmp[datavals+7]))
                        if tup1 is None or tmp[datavals] != tup1[datavals]: # assumes vertical data dups eliminated
                            insertData(cursor3, insert_c_sql, insert_u_sql, 
                                       tmp[:datavals], tmp[datavals], tmp[datavals+1], tmp[datavals+2], tmp[datavals+3],
                                       trio_cursor, gotConfStar)
                                       
                # Move cursor sequentially through right-hand result set, insert data & lineage
                while tup2 is not None:
                    if tup1 is not None and common.compare_values(tup1, tup2, datavals) <= 0:
                        break
                    tmp = tup2
                    tup2 = cursor2.fetchone()
                    if isExcept and (\
                      (tableType2 != common.TYPE_CONFIDENCE and tmp[datavals+2] == 1) or\
                      (tableType2 == common.TYPE_CONFIDENCE and tmp[datavals+4] is not None and tmp[datavals+4] == 1.0)):
                        skipCertain = True
                    if isUnion:
                        if not nolineage2:
                            cursor3.execute(insert_lin_sql % (aid + tmp[datavals], tmp[datavals+5], tmp[datavals+6], tmp[datavals+7]))
                        if tup2 is None or tmp[datavals] != tup2[datavals]: # assumes vertical data dups eliminated
                            insertData(cursor3, insert_c_sql, insert_u_sql, 
                                       tmp[:datavals], aid + tmp[datavals], xid + tmp[datavals+1], tmp[datavals+2], tmp[datavals+3],
                                       trio_cursor, gotConfStar)
                                       
                # Insert lineage of matching alternative from left-hand query
                tmp1 = None
                while tup1 is not None and tup2 is not None and common.compare_values(tup1, tup2, datavals) == 0:
                    tmp1 = tup1
                    tup1 = cursor1.fetchone()
                    if not skipCertain:
                        if isExcept or isIntersect:
                            tmp1[datavals+3] = tmp1[datavals+3] + 1 # conservatively add a ? for EXCEPT/INTERSECT
                        if not nolineage1:
                            cursor3.execute(insert_lin_sql % (tmp1[datavals], tmp1[datavals+5], tmp1[datavals+6], tmp1[datavals+7]))
                        
                # Insert data and lineage of matching alternative from right-hand query to left-hand alternative
                if  tup2 is not None:
                    disj = tup2[datavals+5]
                while tmp1 is not None and tup2 is not None and common.compare_values(tmp1, tup2, datavals) == 0:
                    tmp = tup2
                    tup2 = cursor2.fetchone()
                    if not skipCertain:
                        if not nolineage2:
                            cursor3.execute(insert_lin_sql % (tmp1[datavals], tmp1[datavals+5] + delta_enc + tmp[datavals+5] - disj, tmp[datavals+6], tmp[datavals+7]))
                        if tup2 is None or tmp[datavals] != tup2[datavals]: # assumes vertical data dups eliminated
                            insertData(cursor3, insert_c_sql, insert_u_sql, 
                                       tmp1[:datavals], tmp1[datavals], tmp1[datavals+1], tmp1[datavals+2], tmp1[datavals+3], 
                                       trio_cursor, gotConfStar)
                                       
            cursor1.close()
            cursor2.close()
                       
    if modifiedConfBase:
        cursor3.execute("UPDATE trio_tables SET confbase=TRUE WHERE tablename = '" + tablename + "'")
    trio_cursor.tdb.commit()
    
    if not persistent:
        if not trio_cursor.nolineage:
            # Update metadata for temporary schema-level lineage DAG
            lineage_tables = node.children[0].names + node.children[1].names
            for lin_table in lineage_tables:
               cursor3.execute("INSERT INTO trio_tables_lin VALUES ('%s', '%s')" % (tablename, lin_table))
            trio_cursor.tdb.commit()
        trio_cursor.execute("VIEW TABLE " + tablename)
    cursor3.close()
    return trio_cursor
    
def insertData(cursor, insert_c_sql, insert_u_sql, tuple, aid, xid, alts, enc, trio_cursor, gotConfStar):
    # Insert data wrt. encoding type
    u_cols = []
    c_cols = []
    if trio_cursor.isTemporary:
        u_cols = tuple
        u_cols += [xid, aid, alts, enc]
        if gotConfStar:
            u_cols += [aid]
    else:
        for i in range(trio_cursor.num_datavalues):
            if i in trio_cursor.u_columns:
                u_cols += [tuple[i]]
            else:
                c_cols += [tuple[i]]
        c_cols += [xid]
        u_cols += [xid]
        if trio_cursor.tableType == common.TYPE_PROBABILISTIC and gotConfStar:
            c_cols += [aid]
        if trio_cursor.tableType == common.TYPE_UNCERTAIN or trio_cursor.tableType == common.TYPE_CONFIDENCE:
            c_cols += [alts, enc]
            u_cols += [aid]
            if gotConfStar:
                u_cols += [aid]
    
    # Execute insert statments
    if not trio_cursor.isTemporary:
        cursor.execute(insert_c_sql, c_cols)
    if trio_cursor.isTemporary or trio_cursor.tableType == common.TYPE_UNCERTAIN or trio_cursor.tableType == common.TYPE_CONFIDENCE:
        cursor.execute(insert_u_sql, u_cols)
        
def bindparamstr(numparams):
    return ",".join(["%s" for x in range(numparams)])

