from explorer.lib.base import *
import pylons
import pgdb
import triodb
import common
import misc
from xtable import XTable
from pagedresults import PagedResults
import pydot
import os
import re
import traceback
import sys
from time import strftime
from paste.deploy import CONFIG
config = CONFIG

__id__ = "$Id: trioexplorer.py 583 2007-12-05 22:48:26Z theobald $"

def needsLogin(fn):
    '''Login decorator: if the user is not logged in, send up a login
    rather than the page content.  Point the login form back to the
    same page to avoid a redirect when the user logs in correctly.
    '''
    def _wrapper(*args, **kwargs):
        if 'user' in session and cnx.has_key(session['user'].name):
            # User is logged in; allow access
            #return fn(*args, **kwargs)
            return fn(*args)
        else:
            if request.environ.get('PATH_INFO') == '/index':
                return redirect_to('login?'+request.environ.get('QUERY_STRING'))
            else:
                return redirect_to('login')
    return _wrapper

def addToHistory(query):
    if session['user'].history.count(misc.escapeSpecialChars(query)) != 0:
        session['user'].history.remove(misc.escapeSpecialChars(query))
    session['user'].history.append(misc.escapeSpecialChars(query))
    if len(session['user'].history) == session['user'].history_size + 1:
        del session['user'].history[0]
    session.save()


def calc_conf(table, aid, disj='0', doLog=True):
    tcnx = cnx[session['user'].name]
    tnc = tcnx.ncursor()
    tables = table.split(",")
    aids = [int(x) for x in aid.split(",")]
    disjs = [int(x) for x in disj.split(",")]
    curr = max(aids)
    newdisj = []
    
    if doLog and int(common.cp.get("global", "log_level")) > 0:
            h.log("\nSTART-TIME: " + strftime("%d.%m.%y / %H:%M:%S"))
            h.log("USER:       " + session['user'].name)
            h.log("CONF:       " + str(tables) + "|" + str(aids))
            
    for i in disjs:
        if i == 1:
            curr += 1
        newdisj.append(curr)
        
    # Actual call to stored procedure
    # use this for no memoization
    #query = "select trio_get_conf_spi($$" + "|".join("%s" % x for x in tables) + "|$$, " + \
    #                                  "$$" + "|".join("%d" % x for x in aids) + "|$$, " + \
    #                                  "$$" + "|".join("%d" % x for x in newdisj) + "|$$, 0)"
    # use this for memoization
    query = "select trio_get_conf(array[" + ",".join("'%s'" % x for x in tables) + "]," + \
                                 "array[" + ",".join("%d" % x for x in aids) + "]," + \
                                 "array[" + ",".join("%d" % x for x in newdisj) + "], NULL, 0, 1)"
    conf = None
    
    # Memoize result
    try:
        tnc.execute(query)
        conf = tnc.fetchone()[0]
        if len(tables) == 1:
            tnc.execute("update " + tables[0] + " set trio_conf= " + str(conf) + " where trio_aid=" + str(aids[0]))
    except Exception, ex:
        tcnx.rollback()
        conf = "error"
        h.log(ex)
        
    if doLog and int(common.cp.get("global", "log_level")) > 0:
            h.log("END-TIME:   " + strftime("%d.%m.%y / %H:%M:%S"))
            
    return conf


def executeQuery(query, limit=-1, num_results=-1):
    res_lxtps = []
    error = None
    desc = None
    tableType = None
    truncated = False
    
    try:
        tcnx = cnx[session['user'].name]
        tc = tcnx.cursor()
        if ';' in query:
            for q in query.split(";"):
                if len(q.strip()) > 0:
                    tc.execute(q)
        else:
            tc.execute(query)
            
        conf_computed = True
        if tc.isFetchable:
            while True:
                try:
                    xtuple = tc.xfetchone()
                    if not xtuple or (limit != -1 and len(res_lxtps) == limit):
                        break
                    conf_computed = conf_computed and xtuple.getConfidence() is not None
                    common.quicksort(xtuple.alternatives, 0, len(xtuple.alternatives) -1, xtuple.getConfidence() is not None)
                    res_lxtps.append(xtuple)
                except Exception, ex:
                    break
            desc = tc.description
            tableType = tc.tableType
            common.quicksort(res_lxtps, 0, len(res_lxtps) -1, tableType == common.TYPE_CONFIDENCE or tableType == common.TYPE_PROBABILISTIC)
            if not conf_computed and not tc.gotAsConf:
                conf_computed = False
            elif not conf_computed and (tc.tableType == common.TYPE_CONFIDENCE or tc.tableType == common.TYPE_PROBABILISTIC) and len(res_lxtps) > 0:
                lin = res_lxtps[0][0].getLineage()
                if len(lin.srctable) == 1:
                    tnc = cnx[session['user'].name].ncursor()
                    q = "select confbase from trio_tables where tablename = '%s'" % lin.srctable[0]
                    try:
                        tnc.execute(q)
                        conf_computed = tnc.fetchone()[0]
                    except Exception, ex:
                        conf_computed = False
                    
    except Exception, error:
        h.log(error)
        h.log(query)
        try:
            tcnx.rollback()
        except Exception, ex:
            h.log(ex)
        if int(common.cp.get("global", "debug_level")) > 0:
            e = sys.exc_info()
            traceback.print_tb(e[2])
        return PagedResults(query, res_lxtps, error, desc, tableType)
        
    return PagedResults(query, res_lxtps, error, desc, tableType, tc.nolineage, truncated, conf_computed, 
        tc.isViewTable, tc.distinct, tc.u_columns, num_results)
        
def generateGraph():
    """Generate relation graph."""
    tcnx = cnx[session['user'].name]
    tc = tcnx.cursor()
    tc.execute("\d")
    graph = pydot.Dot(graph_name="schema")
    for xr in tc.fetchall(): 
        thisRelation = xr[0].upper()
        tname =  xr[0]
        xt = XTable(tname, cnx[session['user'].name])
        for attr in xt.getAttributes(): 
            thisRelation += "|" + str(attr[0]).strip("[()]*\'\ ").lower()
            if not attr[1]:
                thisRelation += "*"
        node = pydot.Node(tname, label="{" + thisRelation + "}")
        
        type = xt.getType()
        if type == common.TYPE_CONFIDENCE:
            node.set_color("orange")
        elif type == common.TYPE_PROBABILISTIC:
            node.set_color("purple")
        elif type == common.TYPE_UNCERTAIN:
            node.set_color("green")
        else:
            node.set_color("navy")
        
        node.set_fontname("Garuda")
        node.set_fontsize("10")
        node.set_shape("Mrecord")
        node.set_URL("index?query=view table %s;" % tname)
        
        l = xt.getLineage()
        for parent in l:
            edge = pydot.Edge(tname, parent)
            edge.set_color("skyblue2")
            graph.add_edge(edge)
        if l == []:
            node.set_shape("record")
        graph.add_node(node)
        
    graph.write_png(path=common.ROOT + "/explorer/explorer/public/img/graph.%s.png" % session['user'].name)
    graph.write_cmapx(path=common.ROOT + "/explorer/explorer/public/img/graph.%s.cmapx" % session['user'].name)


class TrioUserData:
    def __init__(self, username):
        self.name = username

        # Hold query history
        self.history = []

        # Number of queries saved in self.history.
        self.history_size = 10

        # Hold queries loaded from file.
        self.queries = []

        # Hold result of a query.
        self.result = None

        # Number of xtuples displayed in one page.
        self.num_tuples = 10

        # Used to assign unique id to alternatives. 
        self.sequence = [0]


cnx = {} # Holds trio connection(s).
class TrioexplorerController(BaseController):
    def __init__(self):
        pass

    #
    # Pages before login
    #
    def createDB(self):
        return render_response('/createdb.myt')

    def doCreateDB(self):
        user = request.params['user']
        password = request.params['password']
        password2 = request.params['password2']

        if password != password2:
            session['error']  = 'Your passwords did not match.'
            session.save()
            h.redirect_to('createDB')

        if user == '' or password ==''or password2 =='':
            session['error'] = 'Required field cannot be left blank'
            session.save()
            return h.redirect_to('createDB')
            
        try:
            
            # CONNECT AS DBA
            db = None
            conf = config['app_conf']
            if "username" in conf and "password" in conf:
                db = pgdb.connect(database=conf["username"], user=conf["username"], password=conf["password"])
            elif "default_user" in conf:
                db = pgdb.connect(database=conf["default_user"], user=conf["default_user"])
            else:
                db = pgdb.connect()
                
            # CREATE NEW DB
            c = db.cursor()
            c.execute("commit;")
            c.execute("create database %s ;" % user)
            c.execute("create role %s login password  \'%s\' ;" % (user, password))
            c.execute("alter database %s owner to %s ;" %  (user, user))
            c.close()
            db.close()
            
            #os.system("createlang plpgsql %s" % user)
            #os.system("createlang plpython %s" % user)
            
            # CONNECT AS NEW USER & SETUP SPI FUNCTIONS
            if "username" in conf and "password" in conf:
                db = pgdb.connect(database=user, user=conf["username"], password=conf["password"])
            elif "default_user" in conf:
                db = pgdb.connect(database=user, user=conf["default_user"])
            else:
                db = pgdb.connect(database=user)
            c = db.cursor()
            file = open(conf["cache_dir"] + '/../../setup/setup_triospi.sql',"r")
            q = file.read()
            c.execute(q)
            
            # SETUP CONF FUNCTION
            file = open(conf["cache_dir"] + '/../../setup/trio_get_conf.sql',"r")
            q = file.read()
            c.execute(q)
            c.execute("commit;")
            c.close()
            db.close()
            
            # RUN SETUP SCRIPT FOR NEW USER
            db = pgdb.connect(user=user, database=user, password=password)
            c = db.cursor()
            file = open(conf["cache_dir"] + '/../../setup/setup.sql',"r")
            q = file.read()
            c.execute(q)
            c.execute("commit;")
            c.close()
            db.close()
            
        except Exception, ex:

            if ('already exists' in str(ex)):
                session['error'] = 'User already exists.'
            else:
                session['error'] = str(ex)
            session.save()
            return h.redirect_to('createDB')
            
        return self.doLogin(user, password)

    def deleteDB(self):
        try:
            tcnx = cnx[session['user'].name]
            tcnx.commit()
            tcnx.close()
            db = None
            conf = config['app_conf']
            if "username" in conf and "password" in conf:
                db = pgdb.connect(user=conf["username"], password=conf["password"])
            elif "default_user" in conf:
                db = pgdb.connect(database=conf["default_user"], user=conf["default_user"])
            else:
                db = pgdb.connect()
            c = db.cursor()
            c.execute("commit")
            c.execute("drop database %s" % session['user'].name)
            c.execute("drop role %s" % session['user'].name)
            db.close()
        except Exception, ex:
            session['error'] = str(ex)
            session.save()
        return h.redirect_to('login')

    def doLogin(self, user='', password=''):
        if user == "" and password == "":
            user = request.params['user']
            password = request.params['password']

        if user == "" or password == "":
            session['error'] = 'Required field cannot be left blank'
            session.save()
            return h.redirect_to('login')

        try:
            tdb = triodb.connect(user=user, database=user, password=password)
        except Exception, ex:
            if ('password authentication failed' in str(ex)):
                session['error'] = "Username and password didn't match"
            elif ('does not exist' in str(ex)):
                session['error'] = "Database '" + request.params['user'] + "' does not exist"
            else:
                session['error'] = str(ex)
            session.save()
            return h.redirect_to('login')
        session['user'] = TrioUserData(user)
        cnx[session['user'].name] = tdb
        session.save()
        if request.params.has_key('referer'):
            referer = request.params['referer'] or 'index'
        else:
            referer = 'index'
        h.redirect_to(referer)

    def login(self):
        if request.params.has_key('query'):
            c.referer = 'index?query='+request.params['query']
        return render_response("/login.myt")

    def logout(self):
        try:
            del cnx[session['user'].name]
        except:
            pass
        try:
            del session['user']
        except:
            pass
        session.save()
        h.redirect_to('login')

    def clearHistory(self):
        session['user'].history= []
        session.save()
        return h.redirect_to('index')

    #
    # Pages after login
    #
    @needsLogin
    def index(self, compconf=False):
        c.tcnx = cnx[session['user'].name]
        query = ''
        page_num = -1
        execute  = 'True'
        compconf = 'False'
        xindex = -1
        
        if request.params.has_key('query'):
            query = request.params['query']
        if request.params.has_key('page_num'):
            page_num = request.params['page_num']
        if request.params.has_key('execute'):
            execute = request.params['execute']
        if request.params.has_key('compconf'):
            compconf = request.params['compconf']
        if request.params.has_key('xindex'):
            xindex = request.params['xindex']
            
        c.query=query
        c.results=None
        num_results = 0
        
        if int(common.cp.get("global", "log_level")) > 0:
            h.log("\nSTART-TIME: " + strftime("%d.%m.%y / %H:%M:%S"))
            h.log("USER:       " + session['user'].name)
            h.log("QUERY:      " + query)
            h.log("CONF:       " + compconf)
            h.log("EXECUTE:    "  + execute)
            
        if query[0:4] == 'view':
            try:
                tcnx = cnx[session['user'].name]
                tnc = tcnx.ncursor()
                tnc.execute('select count(distinct trio_xid) from ' + query.split(" ")[-1])
                num_results = tnc.fetchone()[0]
            except Exception, error:
                e = sys.exc_info()
                h.log(error)
                try:
                    tcnx.rollback()
                except Exception, ex:
                    h.log(ex)
        if num_results > 0:
            if page_num == -1:
                addToHistory(query)
                c.results = executeQuery(query,10,num_results)
                session['user'].results = c.results
            if page_num >= 0:
                c.results = executeQuery(query,(int(page_num)+1)*10,num_results)
                session['user'].results = c.results
                c.results.setPage(int(page_num))
        else:
            if query != '' and execute != 'False' and page_num == -1:
                addToHistory(query)
                c.results = executeQuery(query)
                session['user'].results = c.results
                if c.results.error is None:
                    c.tcnx.commit()
            elif page_num > 0:
                c.results = session['user'].results
                session['user'].results.setPage(int(page_num))
            

        if compconf == 'True':
            if xindex >= 0:
                xtup = session['user'].results.results[int(xindex)]
                aindex = 0
                for alt in xtup:
                    table = ",".join(alt.lineage.tables())
                    aid = ",".join([str(x) for x in alt.lineage.aids()])
                    disj = ",".join([str(x) for x in alt.lineage.disj()])
                    conf = calc_conf(table, aid, disj, doLog=False)
                    session['user'].results.results[int(xindex)][int(aindex)].confidence = conf
                    aindex += 1

            else:
                xindex = 0
                session['user'].results.conf_computed = True
                for xtup in session['user'].results.results:
                    aindex = 0
                    for alt in xtup:
                        table = ",".join(alt.lineage.tables())
                        aid = ",".join([str(x) for x in alt.lineage.aids()])
                        disj = ",".join([str(x) for x in alt.lineage.disj()])
                        conf = calc_conf(table, aid, disj, doLog=False)
                        session['user'].results.results[int(xindex)][int(aindex)].confidence = conf
                        aindex += 1
                    xindex += 1
                    
        if int(common.cp.get("global", "log_level")) > 0:
            h.log("END-TIME:   " + strftime("%d.%m.%y / %H:%M:%S"))
        
        session.save()
        return render_response("/index.myt")

    @needsLogin
    def abort(self):
        cnx[session['user'].name].pdb._pgdbCnx__cnx.cancel()
        return  Response("")

    @needsLogin
    def xconfidence(self):
        table = request.params['table']
        xid = request.params['xid']
        xindex = request.params['xindex']

        tcnx = cnx[session['user'].name]
        tnc = tcnx.ncursor()
        query = "select trio_aid from " + table + " where trio_xid = " + xid
        tnc.execute(query)
        conf = 0.0
        aindex = 0
        try:
            for aid in tnc.fetchall():
                conf += float(calc_conf(table, str(aid[0])))
                aindex += 1
        except:
            conf = "error"
        return Response(str(conf))

    def confidence(self):
        table = request.params['table']
        aid = request.params['aid']
        try:
            disj = request.params['disj']
        except:
            disj = "0"
        xindex, aindex = -1, -1
        if request.params.has_key('xindex'):
            xindex = request.params['xindex']
        if request.params.has_key('aindex'):
            aindex = request.params['aindex']

        conf = calc_conf(table, aid, disj)
        if xindex >= 0 and aindex >= 0:
            session['user'].results.results[int(xindex)][int(aindex)].confidence = conf
            session.save()
        return Response(str(conf))

    @needsLogin
    def lineage(self):
        c.tcnx = cnx[session['user'].name]
        c.top = False
        if request.params.has_key('top'):
            c.top = True
        if request.params.has_key('tables'):
            c.tables = request.params['tables'].split(',')
        if request.params.has_key('aids'):
            c.aids = request.params['aids'].split(',')
        if request.params.has_key('disj'):
            c.disj = request.params['disj'].split(',')
            c.disj = [int(x) for x in c.disj]
            for i in range(len(c.disj)):
                if i > 0 and c.disj[i] == 0:
                    c.disj[i] = c.disj[i - 1]
                elif i > 0:
                    c.disj[i] = c.disj[i - 1] + 1
        return render_response("/lineage.myt")
        
    @needsLogin
    def loadData(self):
        tcnx = cnx[session['user'].name]
        tc = tcnx.cursor()
        tc.execute("\d")
        for xr in tc.fetchall():
            results = executeQuery("drop table %s cascade" % xr[0])
            if results.error is None:
                tcnx.commit()
            else:
                tcnx.rollback()

        if request.params.has_key('dataset'):
            dataset = request.params['dataset'] # movie data
        else:
            dataset = 'crime' # default

        if dataset == 'crime':
            fname = common.ROOT + "/explorer/explorer/public/sql/crime_data.triql"
        else:
            fname = common.ROOT + "/explorer/explorer/public/sql/movie_data.triql"

        f = open(fname)
        for line in f:
            line = re.sub(';','',line)
            results = executeQuery(line)
            if results.error is None:
                tcnx.commit()
            else:
                tcnx.rollback()
        
        session['loaded'] = True
        session.save()
        return h.redirect_to('samples')

    @needsLogin
    def upload(self):
        session['user'].queries = []
        query = ""
        filedata=request.params['myFile']
        if filedata == '':
            return h.redirect_to('/scripts')
        for line in filedata.file.readlines():
  	    if line[:-1].strip == "":
  		continue
  	    if line.startswith("--"):
  		session['user'].queries.append(line)
  	    else:
  		if query == "":
  		    query = line
  		else:
  		    query += " " + line
 		if query[:-1].strip().endswith(";"):
  		    session['user'].queries.append(query.strip())
  		    query = ""
        session.save()
        return h.redirect_to('scripts')

    @needsLogin
    def schema(self):
        generateGraph()
        return render_response("/schema.myt")

    @needsLogin
    def samples(self):
        c.tcnx = cnx[session['user'].name]
        return render_response("/samples.myt")

    @needsLogin
    def scripts(self):
        return render_response("/scripts.myt")

    @needsLogin
    def help(self):
        return render_response("/help.myt")
