#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "triospi.h"
#include "executor/spi.h"

#ifdef PG_MODULE_MAGIC
PG_MODULE_MAGIC;
#endif

int init_fetch(char* select_stmt1, char* select_stmt2, char* lineage_tables, char* column_names, char* column_mappings, int table_type)
{
	int i, j;
	char* tok, *tmp;

	int* c_cols;

	// Stores system column indexes
	aid_idx = palloc(MAX_JOINS * sizeof(int));
	xid_idx = palloc(MAX_JOINS * sizeof(int));
	data_idx = palloc(MAX_COLUMNS * sizeof(int));
	c_cols = palloc(MAX_COLUMNS * sizeof(int));

	// Stores system column mappings for certain and uncertain attributes
	c_datacolumns = palloc(MAX_COLUMNS * sizeof(int));
	u_datacolumns = palloc(MAX_COLUMNS * sizeof(int));

	column_c_list = palloc(MAX_COLUMNNAME * MAX_COLUMNS * sizeof(char));
	column_u_list = palloc(MAX_COLUMNNAME * MAX_COLUMNS * sizeof(char));

	column_c_list[0] = '\0';
	column_u_list[0] = '\0';

	num_c_datacolumns = 0;
	num_u_datacolumns = 0;
	num_joins = 0;

	// Stores the lineage table names
	tablenames = palloc(MAX_COLUMNS * sizeof(char*));
	if (lineage_tables != NULL)
	{
		tmp = palloc(MAX_COLUMNNAME * MAX_COLUMNS * sizeof(char));
		strcpy(tmp, lineage_tables);
		tok = strtok(tmp, "|");
		i=0;
		while (tok != NULL)
		{	
			tablenames[i] = palloc(MAX_COLUMNNAME * sizeof(char));
			strcpy(tablenames[i], tok);
			tok = strtok(NULL, "|");
			i++;
		}
		pfree(tmp);
		num_joins = i;
	}

	// Extract column mappings for certain and uncertain attributes
	if (column_mappings != NULL && table_type > -1)
	{
		tmp = palloc(MAX_COLUMNNAME * MAX_COLUMNS * sizeof(char));
		strcpy(tmp, column_mappings);
		tok = strtok(tmp, "|");
		i=0;
		while (tok != NULL)
		{	
			if (strcmp(tok, "c") == 0) 
			{
				c_cols[i] = 1;
				c_datacolumns[num_c_datacolumns] = i;
				num_c_datacolumns++;
			}
			else 
			{
				c_cols[i] = 0;
				u_datacolumns[num_u_datacolumns] = i;
				num_u_datacolumns++;
			}
			data_idx[i] = i;
			tok = strtok(NULL, "|");
			i++;
		}
		pfree(tmp);

		// Stores the column lists
		tmp = palloc(MAX_COLUMNNAME * MAX_COLUMNS * sizeof(char));
		strcpy(tmp, column_names);
		tok = strtok(tmp, "|");
		i=0;
		while (tok != NULL)
		{	
			if (c_cols[i] == 1) 
			{	
				if (strlen(column_c_list)>0)
					strcat(column_c_list, ",");
				strcat(column_c_list, tok);
			}
			else 
			{
				if (strlen(column_u_list)>0)
					strcat(column_u_list, ",");
				strcat(column_u_list, tok);
			}
			tok = strtok(NULL, "|");
			i++;
		}
		pfree(tmp);
	} 
	else 
	{
		// Stores the column lists
		tmp = palloc(MAX_COLUMNNAME * MAX_COLUMNS * sizeof(char));
		strcpy(tmp, column_names);
		tok = strtok(tmp, "|");
		i=0;
		while (tok != NULL)
		{	
			if (strlen(column_u_list) > 0)
				strcat(column_u_list, ",");
			strcat(column_u_list, tok);
			u_datacolumns[i] = i;
			data_idx[i] = i;
			tok = strtok(NULL, "|");
			i++;
		}
		pfree(tmp);
		num_c_datacolumns = 0;
		num_u_datacolumns = i;
	}
	i = num_c_datacolumns + num_u_datacolumns;

	// Trio column indexes
	for(j = 0; j < num_joins; j++)
	{
		aid_idx[j] = i+j;
	}
	i+=j;
	for(j = 0; j < num_joins; j++) 
	{
		xid_idx[j] = i+j;
	}
	enc_idx = i+j;
	conf_idx = i+j+1;
	alts_idx = i+j+2;

	// Initialize actual cursor
	fetch_plan = SPI_prepare(select_stmt1, 0, NULL);
	if (fetch_plan == NULL) 
	{
		elog(ERROR, "Trio: create fetch plan returned %d",	SPI_result);
		return 0;
	}

	// Allocate a stack for intermediate SPI tuple-tables
	fetch_stack = palloc(sizeof(SPITupleTable*) * MAX_ALTERNATIVES);
	fetch_stack_size = 0;

	// Create portal containing the SPI cursor
	fetch_portal = SPI_cursor_open(NULL, fetch_plan, NULL, NULL, true);

	// Do an initial fetch to get some metadata
	SPI_cursor_fetch(fetch_portal, true, PREFETCH_BUFFER);
	fetch_tuptable = SPI_tuptable;
	fetch_processed = SPI_processed;
	if (fetch_tuptable == NULL) 
	{
		elog(ERROR, "Trio: open cursor returned %d", SPI_result);
		return 0;
	}

	if (select_stmt2 != NULL)
	{
		// Initialize actual cursor
		fetch_plan_setops = SPI_prepare(select_stmt2, 0, NULL);
		if (fetch_plan_setops == NULL) 
		{
			elog(ERROR, "Trio: create fetch plan returned %d",	SPI_result);
			return 0;
		}

		// Create portal containing the SPI cursor
		fetch_portal_setops = SPI_cursor_open(NULL, fetch_plan_setops, NULL, NULL, true);

		// Do an initial fetch to get some metadata
		SPI_cursor_fetch(fetch_portal_setops, true, PREFETCH_BUFFER);
		fetch_tuptable_setops = SPI_tuptable;
		fetch_processed_setops = SPI_processed;
		if (fetch_tuptable_setops == NULL) 
		{
			elog(ERROR, "Trio: open cursor returned %d", SPI_result);
			return 0;
		}
	}

	return 1;
}

int fetchone(XTUPLE* xtuple, int* aid, int* xid)
{
	char* val, *temp;
	VALUE* temp1;
	LINEAGE* temp2;

	HeapTuple tuple;
	bool isnull;

	int i, j = 0, alts = 0;
	int c_vals, c_aids, c_xids;

	// Clean-up dispensible stack elements
	for (i=0; i < fetch_stack_size; i++)
	{
		SPI_freetuptable(fetch_stack[i]);
	}
	fetch_stack_size = 0;

	// Check for buffered alternative first
	current_alt = 0;
	if (buffered_alternative != NULL) 
	{
		// Integrate previously buffered alternative into current X-Tuple
		alternatives[0].aid = buffered_alternative->aid;
		alternatives[0].encoding = buffered_alternative->encoding;
		alternatives[0].confidence = buffered_alternative->confidence;
		alternatives[0].num_values = buffered_alternative->num_values;
		alternatives[0].num_lineage = buffered_alternative->num_lineage;

		// Apply question mark adjustments for groupAlts
		questionMark = buffered_questionMark;

		// Move pointers for buffered alternative
		temp1 = alternatives[0].values;
		temp2 = alternatives[0].lineage;

		alternatives[0].values = buffered_alternative->values;
		alternatives[0].lineage = buffered_alternative->lineage;

		buffered_alternative->values = temp1;
		buffered_alternative->lineage = temp2;

		if (num_joins == 1)
		{
			xtuple->xid = buffered_xid;
		} 
		else
		{
			xtuple->xid = *xid;
		}

		//elog(INFO, "  BUFFERED %d", alternatives[0].aid);
		current_alt = 1;

		if (fetch_tuptable == NULL || fetch_processed == 0)
		{
			buffered_alternative = NULL;

			// Save pointers in X-Tuple
			xtuple->num_alts = 1;
			xtuple->alternatives = alternatives;

			return 1;
		}
	} 
	else
	{
		questionMark = true;
	}

	// The actual fetch loop
	while(fetch_tuptable != NULL && fetch_processed > 0) 
	{
		for (j = current_tuple; j < fetch_processed; j++)
		{
			// Fetch the next tuple from the current cursor
			tuple = fetch_tuptable->vals[j];

            // This should not happen, otherwise change parameter!
			if (current_alt >= MAX_ALTERNATIVES)
			{
				elog(ERROR, "Trio: current xtuple exceeds max. number of alternatives allowed, increase MAX_ALTERNATIVES!");
				return 0;
			}
            
			// Initialize next alternative
			alternatives[current_alt].aid = *aid;
			alternatives[current_alt].confidence = -1;
			alternatives[current_alt].num_lineage = 0;

			c_vals = 0;
			c_aids = 0;
			c_xids = 0;

			current_xid[0] = '\0';
			for (i = 0; i < fetch_tuptable->tupdesc->natts; i++) 
			{
				if (i == enc_idx) 
				{
					val = SPI_getvalue(tuple, fetch_tuptable->tupdesc, i+1);
					//elog(INFO, "%d ADD ENC: %s %d", i, val, strlen(val));
					alternatives[current_alt].encoding = atoi(val);
					pfree(val);
				} 
				else if (i == aid_idx[c_aids]) 
				{
					val = SPI_getvalue(tuple, fetch_tuptable->tupdesc, i+1);
					//elog(INFO, "%d ADD LIN: %s %d %s", i, val, strlen(val), tablenames[c_aids]);
					alternatives[current_alt].lineage[c_aids].disj_aid = 0;
					alternatives[current_alt].lineage[c_aids].src_aid = atoi(val);
					alternatives[current_alt].lineage[c_aids].src_table = tablenames[c_aids];
					pfree(val);
					if (num_joins == 1)
					{
						alternatives[current_alt].aid = alternatives[current_alt].lineage[c_aids].src_aid;
					} 
					// Update aid counter
					else if (c_aids == 0)
					{
						alternatives[current_alt].aid = *aid;
						*aid = alternatives[current_alt].aid+1;
					}	
					if (trio_nolineage == 0)
						alternatives[current_alt].num_lineage++;
					c_aids++;
				}
				else if (i == xid_idx[c_xids]) 
				{
					val = SPI_getvalue(tuple, fetch_tuptable->tupdesc, i+1);
					//elog(INFO, "%d ADD XID: %s %d", i, val, strlen(val));
					if (c_xids == 0)
					{
						if (num_joins == 1)
						{
							xtuple->xid = atoi(val);
						} 
						else
						{
							xtuple->xid = *xid;
						}
					}
					if (trio_groupAlts == -1)
					{
						if (c_xids != 0)
						{
							strcat(current_xid, "|");
						}
						strcat(current_xid, val);
					}
					pfree(val);
					c_xids++;
				}
				else if (i == conf_idx) 
				{
					SPI_getbinval(tuple, fetch_tuptable->tupdesc, i+1, &isnull);
					if (trio_noconf > -1 && isnull==0)
					{
						val = SPI_getvalue(tuple, fetch_tuptable->tupdesc, i+1);
						//elog(INFO, "%d ADD CONF: %s %d", i, val, strlen(val));
						alternatives[current_alt].confidence = atof(val);
						pfree(val);
					}
					else
					{
						//elog(INFO, "%d ADD NULL CONF", i);
						alternatives[current_alt].confidence = NOVAL;
					}
				}
				else if (i == alts_idx) 
				{
					val = SPI_getvalue(tuple, fetch_tuptable->tupdesc, i+1);
					//elog(INFO, "%d ADD ALTS: %s %d", i, val, strlen(val));
					alts = atoi(val);
					pfree(val);
				}
				else if (i == trio_groupAlts) 
				{
					val = SPI_getvalue(tuple, fetch_tuptable->tupdesc, i+1);
					strcpy(current_xid, val);
					pfree(val);
				}
				else 
				{
					alternatives[current_alt].values[c_vals].value_d = SPI_getbinval(tuple, fetch_tuptable->tupdesc, i+1, &isnull);
					alternatives[current_alt].values[c_vals].column_idx = i;
					alternatives[current_alt].values[c_vals].isnull = isnull;
					if (isnull==0)
					{
						val = SPI_getvalue(tuple, fetch_tuptable->tupdesc, i+1);
						//elog(INFO, "   %d %d %d ADD DATA: %s %d", current_alt, i, c_vals, val, strlen(val));
						strcpy(alternatives[current_alt].values[c_vals].value, val);
						alternatives[current_alt].values[c_vals].length = strlen(val);
						pfree(val);
					} 
					else 
					{
						//elog(INFO, "   %d %d ADD NULL DATA", current_alt, i);
						alternatives[current_alt].values[c_vals].length = 0;
					}
					c_vals++;
				}
			}
			current_tuple++;
			num_tuples++;

			// Finalize this alternative
			alternatives[current_alt].num_values = c_vals;

			// Break if XID changes & remember last alternative
			//elog(INFO, "%s <> %s", current_xid, previous_xid);
			if (num_tuples > 1 && (trio_flatten !=0 || strcasecmp(current_xid, previous_xid) != 0)) 
			{
				//elog(INFO, "  %d BREAK: %s->%s", num_tuples, previous_xid, current_xid);
				buffered_alternative = &alternatives[current_alt];
				buffered_questionMark = questionMark && (alts > 1 || alternatives[current_alt].encoding > alts);
				buffered_xid = xtuple->xid;
				temp = previous_xid;
				previous_xid = current_xid;
				current_xid = temp;

				// Save pointers in X-Tuple
				xtuple->num_alts = current_alt;
				xtuple->alternatives = alternatives;

				return 1;
			}
			questionMark = questionMark && (alts > 1 || alternatives[current_alt].encoding > alts);
			temp = previous_xid;
			previous_xid = current_xid;
			current_xid = temp;
			current_alt++;
		}
		buffered_alternative = NULL;
		current_tuple = 0;

		// Push current tuple-table onto stack (preserving pass-through Datum pointers)	
		fetch_stack[fetch_stack_size] = fetch_tuptable;
		fetch_stack_size++;

		// Pre-fetch next set of physical tuples
		SPI_cursor_fetch(fetch_portal, true, PREFETCH_BUFFER);
		fetch_tuptable = SPI_tuptable;
		fetch_processed = SPI_processed;
		if (fetch_tuptable == NULL || fetch_processed == 0) 
		{	
			// Save pointers in X-Tuple
			xtuple->num_alts = current_alt;
			xtuple->alternatives = alternatives;

			return current_alt > 0;
		}
		//else
		//{
		//	elog(INFO, "PREFETCH NEXT BLOCK %d", num_tuples);
		//}
	}
	return 0;
}

// Merges horizontal alternative duplicates by a nested-loop value check
int horizontal_merging(XTUPLE* xtuple)
{
	VALUE* temp1;
	LINEAGE* temp2;

	int i, j, l, d = 0, num_alts = 0;
	bool are_different;

	for (i=0; i < xtuple->num_alts; i++)
	{
		is_unique[i] = 0;
	}
	for (i=0; i < xtuple->num_alts; i++)
	{
		if (is_unique[i] == 0) 
		{
			for (j=i+1; j < xtuple->num_alts; j++)
			{
				are_different = false;
				for (l=0; l < num_c_datacolumns+num_u_datacolumns; l++)
				{
					// Compare string representations of corresponding column values
					if (xtuple->alternatives[i].values[l].isnull == 1 || xtuple->alternatives[j].values[l].isnull == 1 ||
						strcmp(xtuple->alternatives[i].values[l].value, xtuple->alternatives[j].values[l].value) != 0) 
					{
						are_different = true;
						break;
					} 
				}
				if (!are_different)
				{
					// Aggregate confidences
					if (xtuple->alternatives[i].confidence != NOVAL && xtuple->alternatives[j].confidence != NOVAL) 
					{
						xtuple->alternatives[i].confidence += xtuple->alternatives[j].confidence;
					}
					else
					{
						xtuple->alternatives[i].confidence = NOVAL;
					}
                    
					// Track disjunctive lineage
					if (is_unique[i] == 0) 
					{
						for (l=0; l < xtuple->alternatives[i].num_lineage; l++)
						{
							xtuple->alternatives[i].lineage[l].disj_aid = d;
						}
						d++;
					}
					for (l=0; l < xtuple->alternatives[j].num_lineage; l++)
					{
						xtuple->alternatives[i].lineage[xtuple->alternatives[i].num_lineage].src_table = xtuple->alternatives[j].lineage[l].src_table;
						xtuple->alternatives[i].lineage[xtuple->alternatives[i].num_lineage].src_aid = xtuple->alternatives[j].lineage[l].src_aid;
						xtuple->alternatives[i].lineage[xtuple->alternatives[i].num_lineage].disj_aid = d;
						xtuple->alternatives[i].num_lineage++;
					}
					is_unique[i]++;
					is_unique[j]++;
					d++;
				}
			}

			// Integrate merged alternative into current X-Tuple
			merged_alternatives[num_alts].aid = alternatives[i].aid;
			merged_alternatives[num_alts].encoding = alternatives[i].encoding;
			merged_alternatives[num_alts].confidence = alternatives[i].confidence;
			merged_alternatives[num_alts].num_values = alternatives[i].num_values;
			merged_alternatives[num_alts].num_lineage = alternatives[i].num_lineage;

			// Move pointers for merged alternatives
			temp1 = merged_alternatives[num_alts].values;
			temp2 = merged_alternatives[num_alts].lineage;

			merged_alternatives[num_alts].values = xtuple->alternatives[i].values;
			merged_alternatives[num_alts].lineage = xtuple->alternatives[i].lineage;

			xtuple->alternatives[i].values = temp1;
			xtuple->alternatives[i].lineage = temp2;

			num_alts++;
		}
	}

	// Normalize encoding and update xtuple alternatives
	for (i=0; i < num_alts; i++)
	{
		if (xtuple->alternatives[0].encoding > xtuple->num_alts)
		{
			merged_alternatives[i].encoding = num_alts + 1;
		}
		else
		{
			merged_alternatives[i].encoding = num_alts;
		}
	}

	//elog(INFO, "HORIZONTAL MERGING: %d -> %d", xtuple->num_alts, num_alts);
	xtuple->num_alts = num_alts;
	xtuple->alternatives = merged_alternatives;

	return 1;
}

int
trio_selectinto(text* table, text* query, text* fromtables, text* columns, text* columnmappings, int table_type, 
				int merged, int flatten, int groupAlts, int nolineage, int noconf, int nomaybe, int distinct)
{
	XTUPLE xtuple;

	Oid* ctypes, *utypes, *ltypes;	/* column types */
	Datum* cvals, *uvals, *lvals;	/* column values */

	void* insert_c_plan = NULL;
	void* insert_u_plan = NULL;
	void* insert_l_plan;

	char* tablename;
	char* select_stmt;
	char* lineage_tables;
	char* column_list;
	char* column_mappings;

	char* val;
	char* cnulls;
	char* unulls;

	char stmt[8192];
	char insert_c_stmt[8192];
	char insert_u_stmt[8192];
	char insert_l_stmt[8192];
	char separ;

	int i, j, ret = 0, xid = 1, aid = 1, prev_xid = 1, c = 0, u = 0;
	
	// Create an SPI server connection first
	SPI_connect();

	stmt[0] = '\0';
	insert_c_stmt[0] = '\0';
	insert_u_stmt[0] = '\0';
	insert_l_stmt[0] = '\0';

	// Take over some parameters
	trio_merged = merged;
	trio_flatten = flatten;
	trio_groupAlts = groupAlts;
	trio_nolineage = nolineage;
	trio_nomaybe = nomaybe;
	trio_noconf = noconf;
	trio_distinct = distinct;

	// Translate parameters
	tablename = DatumGetCString(DirectFunctionCall1(textout,
		PointerGetDatum(table)));
	select_stmt = DatumGetCString(DirectFunctionCall1(textout,
		PointerGetDatum(query)));
	lineage_tables = DatumGetCString(DirectFunctionCall1(textout,
		PointerGetDatum(fromtables)));
	column_list = DatumGetCString(DirectFunctionCall1(textout,
		PointerGetDatum(columns)));
	column_mappings = DatumGetCString(DirectFunctionCall1(textout,
		PointerGetDatum(columnmappings)));

	if (init_fetch(select_stmt, NULL, lineage_tables, column_list, column_mappings, table_type))
	{
		// Update XID and AID counters
		if (table_type == -1 || table_type == 0 || table_type == 2) 
		{
			snprintf(stmt, sizeof(stmt), "SELECT nextval('trio_xid_%s'), nextval('trio_aid_%s')", tablename, tablename);
		} 
		else 
		{
			snprintf(stmt, sizeof(stmt), "SELECT nextval('trio_xid_%s')", tablename);
		}
		SPI_execute(stmt, true, 0);
		val = SPI_getvalue(SPI_tuptable->vals[0], SPI_tuptable->tupdesc, 1);
		if (val != NULL) 
		{
			prev_xid = atoi(val);
			xid = prev_xid;
			pfree(val);
		}
		if (table_type == -1 || table_type == 0 || table_type == 2) 
		{
			val = SPI_getvalue(SPI_tuptable->vals[0], SPI_tuptable->tupdesc, 2);
			if (val != NULL) 
			{
				aid = atoi(val);
				pfree(val);
			}
		}
		SPI_freetuptable(SPI_tuptable);

		// Some static allocation buffer
		alternatives = palloc(sizeof(ALTERNATIVE) * MAX_ALTERNATIVES);
		merged_alternatives = palloc(sizeof(ALTERNATIVE) * MAX_ALTERNATIVES);
		is_unique = palloc(sizeof(int) * MAX_ALTERNATIVES);
		for (i = 0; i < MAX_ALTERNATIVES; i++)
		{
			alternatives[i].values = palloc((num_c_datacolumns+num_u_datacolumns) * sizeof(VALUE));
			merged_alternatives[i].values = palloc((num_c_datacolumns+num_u_datacolumns) * sizeof(VALUE));
			for (j = 0; j < num_c_datacolumns+num_u_datacolumns; j++)
			{
				//elog(INFO, "PRESERVE %d %d %d", i, j, MAX_COLUMNLENGTH * sizeof(char));
				alternatives[i].values[j].value = palloc(MAX_COLUMNLENGTH * sizeof(char));
				merged_alternatives[i].values[j].value = palloc(MAX_COLUMNLENGTH * sizeof(char));
			}
			alternatives[i].lineage = palloc(num_joins * MAX_ALTERNATIVES * sizeof(LINEAGE));
			merged_alternatives[i].lineage = palloc(num_joins * MAX_ALTERNATIVES * sizeof(LINEAGE));
		}
		previous_xid = palloc(((MAX_COLUMNLENGTH * 2) + 1) * sizeof(char));
		current_xid = palloc(((MAX_COLUMNLENGTH * 2) + 1) * sizeof(char));

		// Allocate data & types of prepared statement for certain attributes
		cvals = (Datum*) palloc((num_c_datacolumns+num_u_datacolumns+5) * sizeof(Datum));
		ctypes = (Oid *) palloc((num_c_datacolumns+num_u_datacolumns+5) * sizeof(Oid));
		cnulls = (char*) palloc((num_c_datacolumns+num_u_datacolumns+5) * sizeof(char));

		// Allocate data & types for prepared statement for uncertain attributes
		uvals = (Datum*) palloc((num_c_datacolumns+num_u_datacolumns+5) * sizeof(Datum));
		utypes = (Oid *) palloc((num_c_datacolumns+num_u_datacolumns+5) * sizeof(Oid));
		unulls = (char*) palloc((num_c_datacolumns+num_u_datacolumns+5) * sizeof(char));

		// Allocate data & types for prepared statement for lineage
		lvals = (Datum*) palloc(4 * sizeof(Datum));
		ltypes = (Oid *) palloc(4 * sizeof(Oid));

		// Construct data query for certain attributes: INSERT INTO table VALUES ($1, ...)
		separ = '(';
		snprintf(insert_c_stmt, sizeof(insert_c_stmt), "INSERT INTO trio_c_%s (%s%s", tablename, column_c_list, num_c_datacolumns > 0 ? "," : "");
		snprintf(insert_u_stmt, sizeof(insert_u_stmt), "INSERT INTO %s%s (%s%s", table_type > -1 ? "trio_u_" : "", tablename, column_u_list, num_u_datacolumns > 0 ? "," : "");
		if (table_type == -1) // temporary table
		{
			snprintf(insert_u_stmt + strlen(insert_u_stmt), sizeof(insert_u_stmt) - strlen(insert_u_stmt), "trio_xid,trio_aid,trio_alts,trio_conf,trio_enc) VALUES ");
			for (i = 0; i < num_u_datacolumns; i++)
			{
				utypes[i] = SPI_gettypeid(fetch_tuptable->tupdesc, u_datacolumns[i]+1);
				snprintf(insert_u_stmt + strlen(insert_u_stmt), sizeof(insert_u_stmt) - strlen(insert_u_stmt), "%c$%d", separ, i+1);
				separ = ',';
			}
			utypes[i] = INT4OID; // xid
			utypes[i+1] = INT4OID; // aid
			utypes[i+2] = INT4OID; // alts
			utypes[i+3] = FLOAT8OID; // conf
			utypes[i+4] = INT4OID; // enc
			snprintf(insert_u_stmt + strlen(insert_u_stmt), sizeof(insert_u_stmt) - strlen(insert_u_stmt), ",$%d,$%d,$%d,$%d,$%d)", i+1, i+2, i+3, i+4, i+5);
			c = 0;
			u = 5;
		} 
		else if (table_type == 1)
		{
			snprintf(insert_c_stmt + strlen(insert_c_stmt), sizeof(insert_c_stmt) - strlen(insert_c_stmt), "trio_xid) VALUES ");
			for (i = 0; i < num_c_datacolumns; i++)
			{
				ctypes[i] = SPI_gettypeid(fetch_tuptable->tupdesc, c_datacolumns[i]+1);
				snprintf(insert_c_stmt + strlen(insert_c_stmt), sizeof(insert_c_stmt) - strlen(insert_c_stmt), "%c$%d", separ, i+1);
				separ = ',';
			}
			ctypes[i] = INT4OID; // xid
			snprintf(insert_c_stmt + strlen(insert_c_stmt), sizeof(insert_c_stmt) - strlen(insert_c_stmt), ",$%d)", i+1);
			c = 1;
			u = 0;
		}
		else if (table_type == 3)
		{
			snprintf(insert_c_stmt + strlen(insert_c_stmt), sizeof(insert_c_stmt) - strlen(insert_c_stmt), "trio_xid,trio_conf) VALUES ");
			for (i = 0; i < num_c_datacolumns; i++)
			{
				ctypes[i] = SPI_gettypeid(fetch_tuptable->tupdesc, c_datacolumns[i]+1);
				snprintf(insert_c_stmt + strlen(insert_c_stmt), sizeof(insert_c_stmt) - strlen(insert_c_stmt), "%c$%d", separ, i+1);
				separ = ',';
			}
			ctypes[i] = INT4OID; // xid
			ctypes[i+1] = FLOAT8OID; // conf
			snprintf(insert_c_stmt + strlen(insert_c_stmt), sizeof(insert_c_stmt) - strlen(insert_c_stmt), ",$%d,$%d)", i+1, i+2);
			c = 2;
			u = 0;
		} 
		else
		{
			snprintf(insert_c_stmt + strlen(insert_c_stmt), sizeof(insert_c_stmt) - strlen(insert_c_stmt), "trio_xid,trio_alts,trio_enc) VALUES ");
			for (i = 0; i < num_c_datacolumns; i++)
			{
				ctypes[i] = SPI_gettypeid(fetch_tuptable->tupdesc, c_datacolumns[i]+1);
				snprintf(insert_c_stmt + strlen(insert_c_stmt), sizeof(insert_c_stmt) - strlen(insert_c_stmt), "%c$%d", separ, i+1);
				separ = ',';
			}
			ctypes[i] = INT4OID; // xid
			ctypes[i+1] = INT4OID; // alts
			ctypes[i+2] = INT4OID; // enc
			snprintf(insert_c_stmt + strlen(insert_c_stmt), sizeof(insert_c_stmt) - strlen(insert_c_stmt), "%s$%d,$%d,$%d)", num_c_datacolumns > 0 ? "," : "(", i+1, i+2, i+3);
			c = 3;
			separ = '(';
			if (table_type == 0)
			{
				snprintf(insert_u_stmt + strlen(insert_u_stmt), sizeof(insert_u_stmt) - strlen(insert_u_stmt), "trio_xid,trio_aid) VALUES ");
				for (i = 0; i < num_u_datacolumns; i++)
				{
					utypes[i] = SPI_gettypeid(fetch_tuptable->tupdesc, u_datacolumns[i]+1);
					snprintf(insert_u_stmt + strlen(insert_u_stmt), sizeof(insert_u_stmt) - strlen(insert_u_stmt), "%c$%d", separ, i+1);
					separ = ',';
				}
				utypes[i] = INT4OID; // xid
				utypes[i+1] = INT4OID; // aid
				snprintf(insert_u_stmt + strlen(insert_u_stmt), sizeof(insert_u_stmt) - strlen(insert_u_stmt), "%s$%d,$%d)", num_u_datacolumns > 0 ? "," : "(", i+1, i+2);
				u = 2;
			} 
			else if (table_type == 2) 
			{
				snprintf(insert_u_stmt + strlen(insert_u_stmt), sizeof(insert_u_stmt) - strlen(insert_u_stmt), "trio_xid,trio_aid,trio_conf) VALUES ");
				for (i = 0; i < num_u_datacolumns; i++)
				{
					utypes[i] = SPI_gettypeid(fetch_tuptable->tupdesc, u_datacolumns[i]+1);
					snprintf(insert_u_stmt + strlen(insert_u_stmt), sizeof(insert_u_stmt) - strlen(insert_u_stmt), "%c$%d", separ, i+1);
					separ = ',';
				}
				utypes[i] = INT4OID; // xid
				utypes[i+1] = INT4OID; // aid
				utypes[i+2] = FLOAT8OID; // conf
				snprintf(insert_u_stmt + strlen(insert_u_stmt), sizeof(insert_u_stmt) - strlen(insert_u_stmt), "%s$%d,$%d,$%d)", num_u_datacolumns > 0 ? "," : "(", i+1, i+2, i+3);
				u = 3;
			}
		}
		
		// Construct lineage query: INSERT INTO trio_lin_table VALUES ($1, ...)
		snprintf(insert_l_stmt, sizeof(insert_l_stmt), "INSERT INTO trio_lin_%s (aid, disjlin, srclin, srctable) VALUES ($1,$2,$3,lower($4))", tablename);
		ltypes[0] = INT4OID;
		ltypes[1] = INT4OID;
		ltypes[2] = INT4OID;
		ltypes[3] = VARCHAROID;
		
		// Prepare plan for certain data insert statement
		if (table_type > -1) 
		{
			insert_c_plan = SPI_prepare(insert_c_stmt, num_c_datacolumns+c, ctypes);
			if (insert_c_plan == NULL)
				elog(ERROR, "Trio: create table %s returned %d", tablename, SPI_result);
		}

		// Prepare plan for uncertain data insert statement
		if (table_type == -1 || table_type == 0 || table_type == 2) {
			insert_u_plan = SPI_prepare(insert_u_stmt, num_u_datacolumns+u, utypes);
			if (insert_u_plan == NULL)
				elog(ERROR, "Trio: create table %s returned %d", tablename, SPI_result);
		}

		// Prepare plan for lineage insert statement
		insert_l_plan = SPI_prepare(insert_l_stmt, 4, ltypes);
		if (insert_l_plan == NULL)
			elog(ERROR, "Trio: create table %s returned %d", tablename, SPI_result);

		// Begin of transaction
		SPI_execute("begin", true, 0);
		
		buffered_alternative = NULL;
		current_alt = 0;
		num_tuples = 0;

		while(fetchone(&xtuple, &aid, &xid) == 1) 
		{
			// Forward aid from temp table (only for 'DISTINCT' in 'CREATE TABLE AS' statements)
			if (trio_distinct != 0) 
			{
				for (i=0; i < xtuple.num_alts; i++)
				{
					xtuple.alternatives[i].aid = xtuple.alternatives[i].lineage[0].src_aid;
				}
			}

			// Take over aid as xid for certain and probabilistic tables w/o alternatives
			if (table_type == 1 || table_type == 3)
			{
				xtuple.xid = xtuple.alternatives[0].aid;
			}
			else
			{
				xtuple.xid = xid;
			}

			// Questionmark (GroupAlts) adjustments before horizontal merging
			for(i = 0; i < xtuple.num_alts; i++) 
			{
				// Adjust encoding
				if (trio_nomaybe != 0)
				{	
					xtuple.alternatives[i].encoding = xtuple.num_alts;
					xtuple.alternatives[i].confidence = NOVAL;
				}
				else 
				{
					if (((trio_groupAlts > -1 || trio_distinct != 0) && questionMark) || xtuple.alternatives[i].encoding > xtuple.num_alts)
					{
						xtuple.alternatives[i].encoding = xtuple.num_alts + 1;
					} 
					else
					{
						xtuple.alternatives[i].encoding = xtuple.num_alts;
					}
				}
				if (trio_groupAlts > -1 && trio_noconf < 1)
				{
					xtuple.alternatives[i].confidence = NOVAL;
				} 
				else if (trio_noconf == 2)
				{
					// Uniform as conf
					if (xtuple.alternatives[i].encoding > xtuple.num_alts)
					{
						xtuple.alternatives[i].confidence = 1.0 / (1 + xtuple.num_alts);
					} 
					else
					{
						xtuple.alternatives[i].confidence = 1.0 / xtuple.num_alts;
					}
				}
			}
            
			// Merge duplicate alternatives "horizontically"
			if (trio_merged == 1 && trio_flatten == 0)
			{
				horizontal_merging(&xtuple);
			}
			//elog(INFO, "XTUP-%d ALTS %d", xtuple.xid, xtuple.num_alts);


			for(i = 0; i < xtuple.num_alts; i++) 
			{
				// Set certain data columns for new xtuple
				for(j = 0; j < num_c_datacolumns; j++)
				{
					if (xtuple.alternatives[i].values[c_datacolumns[j]].isnull == 0)
					{
						cvals[j] = xtuple.alternatives[i].values[c_datacolumns[j]].value_d;
							//DirectFunctionCall1(textin, 
							//CStringGetDatum(xtuple.alternatives[0].values[c_datacolumns[j]].value));
						cnulls[j] = ' ';
					} 
					else 
					{
						cnulls[j] = 'n';
					}
				}

				// Set uncertain data columns for new alternative
				for(j = 0; j < num_u_datacolumns; j++)
				{
					if (xtuple.alternatives[i].values[u_datacolumns[j]].isnull == 0)
					{
						uvals[j] = xtuple.alternatives[i].values[u_datacolumns[j]].value_d;
							//DirectFunctionCall1(textin, 
							//CStringGetDatum(xtuple.alternatives[i].values[u_datacolumns[j]].value));
						unulls[j] = ' ';
					} 
					else 
					{
						unulls[j] = 'n';
					}
				}

				// Add Trio system columns for new xtuple
				cvals[num_c_datacolumns] = Int32GetDatum(xtuple.xid);
				uvals[num_u_datacolumns] = Int32GetDatum(xtuple.xid);
				uvals[num_u_datacolumns+1] = Int32GetDatum(xtuple.alternatives[i].aid);
				if (table_type == -1)
				{
					uvals[num_u_datacolumns+2] = Int32GetDatum(xtuple.num_alts);
					if (xtuple.alternatives[i].confidence == NOVAL) 
					{
						unulls[num_u_datacolumns+3] = 'n';
					}
					else
					{
						uvals[num_u_datacolumns+3] = Float8GetDatum(xtuple.alternatives[i].confidence);
						unulls[num_u_datacolumns+3] = ' ';
					}
					uvals[num_u_datacolumns+4] = Int32GetDatum(xtuple.alternatives[i].encoding > xtuple.num_alts ? 
						xtuple.num_alts + 1 : xtuple.alternatives[i].encoding);
				} 
				else if (table_type == 3) 
				{
					if (xtuple.alternatives[i].confidence == NOVAL) 
					{
						cnulls[num_c_datacolumns+1] = 'n';
					}
					else
					{
						cvals[num_c_datacolumns+1] = Float8GetDatum(xtuple.alternatives[i].confidence);
						cnulls[num_c_datacolumns+1] = ' ';
					}
				}
				else if (table_type != 1)
				{
					cvals[num_c_datacolumns+1] = Int32GetDatum(xtuple.num_alts);
					cvals[num_c_datacolumns+2] = Int32GetDatum(xtuple.alternatives[i].encoding > xtuple.num_alts ? 
						xtuple.num_alts + 1 : xtuple.alternatives[i].encoding);
					if (table_type == 2)
					{
						if (xtuple.alternatives[i].confidence == NOVAL) 
						{
							unulls[num_u_datacolumns+2] = 'n';
						}
						else
						{
							uvals[num_u_datacolumns+2] = Float8GetDatum(xtuple.alternatives[i].confidence);
							unulls[num_u_datacolumns+2] = ' ';
						}
					}
				}

				// Execute data insert statement for certain attributes
				if (i == 0 && table_type > -1)
				{
					ret = SPI_execp(insert_c_plan, cvals, cnulls, 0);
					SPI_freetuptable(SPI_tuptable);
					if (ret < 0) 
					{
						elog(ERROR, "Trio: SPI_execp returned %d", ret);
						return 0;
					}
				}
				
				// Execute data insert statement for uncertain attributes
				if (table_type != 1 && table_type != 3)
				{
					ret = SPI_execp(insert_u_plan, uvals, unulls, 0);
					SPI_freetuptable(SPI_tuptable);
					if (ret < 0) 
					{
						elog(ERROR, "Trio: SPI_execp returned %d", ret);
						return 0;
					}
				}

				// Set lineage entries for new alternative
				if (trio_distinct == 0)
				{
					for(j = 0; j < xtuple.alternatives[i].num_lineage; j++)
					{
						// Enumerate lineage entries
						lvals[0] = Int32GetDatum(xtuple.alternatives[i].aid);
						lvals[1] = Int32GetDatum(xtuple.alternatives[i].lineage[j].disj_aid);
						lvals[2] = Int32GetDatum(xtuple.alternatives[i].lineage[j].src_aid);
						lvals[3] = DirectFunctionCall1(textin, CStringGetDatum(xtuple.alternatives[i].lineage[j].src_table));

						// Execute lineage insert statement
						ret = SPI_execp(insert_l_plan, lvals, NULL, 0);
						//pfree(DatumGetPointer(lvals[3]));
						SPI_freetuptable(SPI_tuptable);
						if (ret < 0) 
						{
							elog(ERROR, "Trio: SPI_execp returned %d", ret);
							return 0;
						}
					}
				}

				// Periodically commit
				if (aid % 5000) 
				{
					SPI_execute("commit", true, 0);
					SPI_execute("begin", true, 0);
				}
			}

			// new XID
			xid++;

		}
		SPI_execute("commit", true, 0);
		//elog(INFO, "SPI: %d X-TUPLE(S) INSERTED.", xid - prev_xid);

		pfree(cvals);
		pfree(ctypes);
		pfree(uvals);
		pfree(utypes);
		pfree(lvals);
		pfree(ltypes);
	}
	
	// Clean up
	SPI_finish();

	return 1;
}

int
trio_selectinto_distinct(text* table, text* query, text* columns, text* fromtables, int compute_conf)
{
	Oid* ctypes, *ltypes;	/* column types */
	Datum *cvals, *lvals;	/* column values */

	HeapTuple tuple, tmp_tuple;
	
	Datum d1, d2;
	
	int** lineage_buffer;

	void* insert_plan;
	void* insert_lin_plan;

	char* tmp_table;
	char* column_list;
	char* select_stmt;
	char* lineage_tables;

	char* val, *val1, *val2;
	char* cnulls;

	char stmt[8192];
	char insert_stmt[8192];
	char insert_lin_stmt[8192];
	char separ = '(';

	int b, i, j, l, xid = 1, tuple_xid, tmp_xid, aid = 1, disj = 0, ret, buffer;
	int* max_xids;

	bool do_buffer, are_different, isnull;

	stmt[0] = '\0';
	insert_stmt[0] = '\0';
	insert_lin_stmt[0] = '\0';

	// Create an SPI server connection first
	SPI_connect();

	// Translate parameters
	tmp_table = DatumGetCString(DirectFunctionCall1(textout,
		PointerGetDatum(table)));
	column_list = DatumGetCString(DirectFunctionCall1(textout,
		PointerGetDatum(columns)));
	select_stmt = DatumGetCString(DirectFunctionCall1(textout,
		PointerGetDatum(query)));
	lineage_tables = DatumGetCString(DirectFunctionCall1(textout,
		PointerGetDatum(fromtables)));

	// Run the query sorted by data columns & xid's
	if (init_fetch(select_stmt, NULL, lineage_tables, column_list, NULL, -1))
	{
		// Update XID's
		max_xids = (int*) palloc(num_joins * sizeof(int));
		for (i = 0; i < num_joins; i++)
		{
			snprintf(stmt, sizeof(stmt), "SELECT nextval('trio_xid_%s')", tablenames[i]);
			SPI_execute(stmt, true, 0);
			val = SPI_getvalue(SPI_tuptable->vals[0], SPI_tuptable->tupdesc, 1);
			if (val != NULL) 
			{
				max_xids[i] = atoi(val) +1;
				xid += pow(max_xids[i], i+1);
				pfree(val);
			}
		}
		SPI_freetuptable(SPI_tuptable);

		// Allocate some lineage buffer (this is going to be at most the size of an x-tuple)
		lineage_buffer = palloc(sizeof(int*) * MAX_ALTERNATIVES);
		for (i = 0; i < MAX_ALTERNATIVES; i++)
			lineage_buffer[i] = palloc(sizeof(int) * num_joins);

		// Allocate ctypes for prepared statement 
		ctypes = (Oid *) palloc((num_u_datacolumns+5) * sizeof(Oid));
		cnulls = (char*) palloc((num_u_datacolumns+5) * sizeof(char));

		// Construct data query: INSERT INTO table VALUES ($1, ...)
		snprintf(insert_stmt, sizeof(insert_stmt), "INSERT INTO %s (%s,trio_xid,trio_aid,trio_alts,trio_conf,trio_enc) VALUES ", tmp_table, column_u_list);
		for (i = 0; i < num_u_datacolumns+5; i++)
		{
			cnulls[i] = ' ';
			if (i == num_u_datacolumns)
				ctypes[i] = INT4OID;
			else if (i == num_u_datacolumns+1)
				ctypes[i] = INT4OID;
			else if (i == num_u_datacolumns+2) 
				ctypes[i] = INT4OID;
			else if (i == num_u_datacolumns+3) 
				if (compute_conf == 1) 
					ctypes[i] = INT4OID;
				else
					ctypes[i] = FLOAT8OID;
			else if (i == num_u_datacolumns+4)
				ctypes[i] = INT4OID;
			else
				ctypes[i] = SPI_gettypeid(fetch_tuptable->tupdesc, i+1);

			if (compute_conf == 1 && i == num_u_datacolumns+3)
				snprintf(insert_stmt + strlen(insert_stmt), sizeof(insert_stmt) - strlen(insert_stmt), 
				    "%ctrio_get_conf(array['%s'],array[$%d]::integer[],array[1],NULL::float[],0,1)", separ, tmp_table, i+1);
			else
				snprintf(insert_stmt + strlen(insert_stmt), sizeof(insert_stmt) - strlen(insert_stmt), "%c$%d", separ, i+1);
			separ = ',';
		}
		snprintf(insert_stmt + strlen(insert_stmt), sizeof(insert_stmt) - strlen(insert_stmt), ")");

		// Prepare plan for data insert statement
		insert_plan = SPI_prepare(insert_stmt, num_u_datacolumns+5, ctypes);
		if (insert_plan == NULL)
			elog(ERROR, "Trio: create table %s returned %d", tmp_table, SPI_result);

		// Allocate ctypes for prepared statement 
		ltypes = (Oid *) palloc(4 * sizeof(Oid));

		// Construct lineage query: INSERT INTO trio_lin_table VALUES ($1, ...)
		snprintf(insert_lin_stmt, sizeof(insert_lin_stmt), "INSERT INTO trio_lin_%s (aid, disjlin, srclin, srctable) VALUES ($1,$2,$3,$4)", tmp_table);
		ltypes[0] = INT4OID;
		ltypes[1] = INT4OID;
		ltypes[2] = INT4OID;
		ltypes[3] = VARCHAROID;

		// Prepare plan for lineage insert statement
		insert_lin_plan = SPI_prepare(insert_lin_stmt, 4, ltypes);
		if (insert_plan == NULL)
			elog(ERROR, "Trio: create table trio_lin_%s returned %d", tmp_table, SPI_result);

		// Container for new tuples
		cvals = (Datum*) palloc((num_u_datacolumns+5) * sizeof(Datum));
		lvals = (Datum*) palloc(4 * sizeof(Datum));

		SPITupleTable* prev_fetch_tuptable = NULL;

		// Begin of transaction
		SPI_execute("begin", true, 0);

		d1 = NOVAL;
		d2 = NOVAL;

		tmp_tuple = NULL;
		tmp_xid = NOVAL;

		// Read alts sequentially, first buffer alts with same data and xid
		// - when xid changes, flush lineage into new x-tuple & stop buffering, else keep xid
		// - when data changes, flush distinct data with new aid each
		disj = 0;
		buffer = 0;
		do_buffer = true;
		num_tuples = 0;

		// The actual fetch loop
		while(fetch_tuptable != NULL && fetch_processed > 0) 
		{
			for (j = 0; j < fetch_processed; j++)
			{
				// Fetch the next tuple from the current cursor
				tuple = fetch_tuptable->vals[j];
				tuple_xid = get_xid(tuple, max_xids);

				if (num_tuples>0)
				{
					are_different = false;
					for (l=0; l < num_u_datacolumns; l++)
					{
						if (!are_different)
						{
							d1 = SPI_getbinval(tuple, fetch_tuptable->tupdesc, l+1, &isnull);
							if (isnull!=0)
							{
								// we define: null != null
								are_different = true;
								break;
							}
						}
						d2 = SPI_getbinval(tmp_tuple, fetch_tuptable->tupdesc, l+1, &isnull);
						if (isnull!=0)
						{
							// we define: null != null
							are_different = true;
							cnulls[l] = 'n';
						}
						else
						{
							cnulls[l] = ' ';
						}
						// Compare string representations of corresponding column values
						//elog(INFO, "%s <> %s", 
						//	SPI_getvalue(tuple, fetch_tuptable->tupdesc, l+1),
						//	SPI_getvalue(tmp_tuple, fetch_tuptable->tupdesc, l+1));
						if (!are_different)
						{
							val1 = SPI_getvalue(tuple, fetch_tuptable->tupdesc, l+1);
							val2 = SPI_getvalue(tmp_tuple, fetch_tuptable->tupdesc, l+1);
							are_different = (strcmp(val1, val2) != 0);
							pfree(val1);
							pfree(val2);
						}
					}

					// Deduplicate data and process lineage
					if (are_different)
					{
						// flush lineage buffer
						for(l = 0; l < num_joins; l++)
						{
							val = SPI_getvalue(tmp_tuple, fetch_tuptable->tupdesc, num_u_datacolumns+l+1);
							lineage_buffer[buffer][l] = atoi(val);
							pfree(val);
						}
						buffer++;
						for (b = 0; b < buffer; b++)
						{
							for(l = 0; l < num_joins; l++)
							{
								// Enumerate lineage entries
								lvals[0] = Int32GetDatum(aid);
								lvals[1] = Int32GetDatum(disj);
								lvals[2] = Int32GetDatum(lineage_buffer[b][l]);
								lvals[3] = DirectFunctionCall1(textin, CStringGetDatum(tablenames[l]));

								// Execute lineage insert statement
								ret = SPI_execp(insert_lin_plan, lvals, NULL, 0);
								//pfree(DatumGetPointer(lvals[3]));
								SPI_freetuptable(SPI_tuptable);
								if (ret < 0) 
								{
									elog(ERROR, "Trio: SPI_execp returned %d", ret);
									return 0;
								}
							}
							disj += 1;
						}

						// Set data columns for new x-tuple
						for(l = 0; l < num_u_datacolumns; l++)
						{
							if (cnulls[l] == ' ') 
							{
								cvals[l] = SPI_getbinval(tmp_tuple, fetch_tuptable->tupdesc, l+1, &isnull);
								//val = SPI_getvalue(tmp_tuple, fetch_tuptable->tupdesc, l+1);
								//cvals[l] = DirectFunctionCall1(textin, CStringGetDatum(val)); 
								//pfree(val);
							}
						}

						// Add Trio system columns for the new x-tuple alternative
						// trio_aid
						cvals[l] = Int32GetDatum(do_buffer ? tmp_xid : xid);
						// trio_xid
						cvals[l + 1] = Int32GetDatum(aid);
						// trio_alts
						val = SPI_getvalue(tmp_tuple, fetch_tuptable->tupdesc, num_u_datacolumns+(2*num_joins)+3);
						cvals[l + 2] = Int32GetDatum(atoi(val));
						pfree(val);
						// trio_conf
						if (compute_conf==1)
						{
							cvals[l + 3] = Int32GetDatum(aid);
							cnulls[l + 3] = ' ';
						}
						else
						{
							SPI_getbinval(tmp_tuple, fetch_tuptable->tupdesc, num_u_datacolumns+(2*num_joins)+2, &isnull);
							if (isnull==0) 
							{
								val = SPI_getvalue(tmp_tuple, fetch_tuptable->tupdesc, num_u_datacolumns+(2*num_joins)+2);
								cvals[l + 3] = Float8GetDatum(atof(val));
								cnulls[l + 3] = ' ';
								pfree(val);
							}
							else
							{
								cnulls[l + 3] = 'n';
							}
						}
						// trio_enc
						val = SPI_getvalue(tmp_tuple, fetch_tuptable->tupdesc, num_u_datacolumns+(2*num_joins)+1);
						cvals[l + 4] = Int32GetDatum(atoi(val));
						pfree(val);

						// Execute insert plan
						ret = SPI_execp(insert_plan, cvals, cnulls, 0);
						SPI_freetuptable(SPI_tuptable);
						if (ret < 0) 
						{
							elog(ERROR, "Trio: SPI_execp returned %d", ret);
							return 0;
						}
						if (aid % 5000) 
						{
							SPI_execute("commit", true, 0);
							SPI_execute("begin", true, 0);
						}

						disj = 0;
						buffer = 0;
						do_buffer = true;
						// create new aid
						aid += 1;
					}
					else
					{
						if (do_buffer && tmp_xid == tuple_xid)
						{
							// keep buffering lineage
							for(l = 0; l < num_joins; l++)
							{
								val = SPI_getvalue(tmp_tuple, fetch_tuptable->tupdesc, num_u_datacolumns+l+1);
								lineage_buffer[buffer][l] = atoi(val);
								pfree(val);
							}
							buffer++;
						}
						else
						{
							if (do_buffer)
							{
								// flush lineage buffer
								for(l = 0; l < num_joins; l++)
								{
									val = SPI_getvalue(tmp_tuple, fetch_tuptable->tupdesc, num_u_datacolumns+l+1);
									lineage_buffer[buffer][l] = atoi(val);
									pfree(val);
								}
								buffer++;

								for (b = 0; b < buffer; b++)
								{
									for(l = 0; l < num_joins; l++)
									{
										// Enumerate lineage entries
										lvals[0] = Int32GetDatum(aid);
										lvals[1] = Int32GetDatum(disj);
										lvals[2] = Int32GetDatum(lineage_buffer[b][l]);
										lvals[3] = DirectFunctionCall1(textin, CStringGetDatum(tablenames[l]));

										// Execute lineage insert statement
										ret = SPI_execp(insert_lin_plan, lvals, NULL, 0);
										//pfree(DatumGetPointer(lvals[3]));
										SPI_freetuptable(SPI_tuptable);
										if (ret < 0) 
										{
											elog(ERROR, "Trio: SPI_execp returned %d", ret);
											return 0;
										}
									}
									disj += 1;
								}
								buffer = 0;
								do_buffer = false;
								// create new xid
								xid += 1;
							}
							else
							{
								// flush lineage w/out buffering
								for(l = 0; l < num_joins; l++)
								{
									// Enumerate lineage entries
									lvals[0] = Int32GetDatum(aid);
									lvals[1] = Int32GetDatum(disj);
									lvals[2] = SPI_getbinval(tmp_tuple, fetch_tuptable->tupdesc, num_u_datacolumns+l+1, &isnull);
									lvals[3] = DirectFunctionCall1(textin, CStringGetDatum(tablenames[l]));

									// Execute lineage insert statement
									ret = SPI_execp(insert_lin_plan, lvals, NULL, 0);
									//pfree(DatumGetPointer(lvals[3]));
									SPI_freetuptable(SPI_tuptable);
									if (ret < 0) 
									{
										elog(ERROR, "Trio: SPI_execp returned %d", ret);
										return 0;
									}
								}
								disj += 1;
							}
						}
					}
				}

				tmp_tuple = tuple;
				tmp_xid = tuple_xid;

				num_tuples++;
			}

			// Pre-fetch next block of physical tuples
			if (prev_fetch_tuptable != NULL)
				SPI_freetuptable(prev_fetch_tuptable);
			prev_fetch_tuptable = fetch_tuptable;
			SPI_cursor_fetch(fetch_portal, true, PREFETCH_BUFFER);
			fetch_tuptable = SPI_tuptable;
			fetch_processed = SPI_processed;
		}

		// Dump final lineage buffered for tmp_tuple
		if ((fetch_tuptable == NULL || fetch_processed == 0) && tmp_tuple != NULL) 
		{
			// flush buffer
			for(l = 0; l < num_joins; l++)
			{
				val = SPI_getvalue(tmp_tuple, fetch_tuptable->tupdesc, num_u_datacolumns+l+1);
				lineage_buffer[buffer][l] = atoi(val);
				pfree(val);
			}
			buffer++;
			for (b = 0; b < buffer; b++)
			{
				for(l = 0; l < num_joins; l++)
				{
					// Enumerate lineage entries
					lvals[0] = Int32GetDatum(aid);
					lvals[1] = Int32GetDatum(disj);
					lvals[2] = Int32GetDatum(lineage_buffer[b][l]);
					lvals[3] = DirectFunctionCall1(textin, CStringGetDatum(tablenames[l]));

					// Execute lineage insert statement
					ret = SPI_execp(insert_lin_plan, lvals, NULL, 0);
					//pfree(DatumGetPointer(lvals[3]));
					SPI_freetuptable(SPI_tuptable);
					if (ret < 0) 
					{
						elog(ERROR, "Trio: SPI_execp returned %d", ret);
						return 0;
					}
				}
				disj += 1;
			}

			// Set data columns for new x-tuple
			for(l = 0; l < num_u_datacolumns; l++)
			{
				cvals[l] = SPI_getbinval(tmp_tuple, fetch_tuptable->tupdesc, l+1, &isnull);
				if (isnull == 0)
				{
					//val = SPI_getvalue(tmp_tuple, fetch_tuptable->tupdesc, l+1);
					//cvals[l] = DirectFunctionCall1(textin, CStringGetDatum(val)); 
					//pfree(val);
					cnulls[l] = ' ';
				}
				else
				{
					cnulls[l] = 'n';
				}
			}

			// Add Trio system columns for the new x-tuple alternative
			// trio_aid
			cvals[l] = Int32GetDatum(do_buffer ? tmp_xid : xid);
			// trio_xid
			cvals[l + 1] = Int32GetDatum(aid);
			// trio_alts
			val = SPI_getvalue(tmp_tuple, fetch_tuptable->tupdesc, num_u_datacolumns+(2*num_joins)+3);
			cvals[l + 2] = Int32GetDatum(atoi(val));
			pfree(val);
			// trio_conf
			if (compute_conf==1)
			{
				cvals[l + 3] = Int32GetDatum(aid);
				cnulls[l + 3] = ' ';
			}
			else
			{
				SPI_getbinval(tmp_tuple, fetch_tuptable->tupdesc, num_u_datacolumns+(2*num_joins)+2, &isnull);
				if (isnull==0) 
				{
					val = SPI_getvalue(tmp_tuple, fetch_tuptable->tupdesc, num_u_datacolumns+(2*num_joins)+2);
					cvals[l + 3] = Float8GetDatum(atof(val));
					cnulls[l +3] = ' ';
					pfree(val);
				}
				else
				{
					cnulls[l + 3] = 'n';
				}
			}
			// trio_enc
			val = SPI_getvalue(tmp_tuple, fetch_tuptable->tupdesc, num_u_datacolumns+(2*num_joins)+1);
			cvals[l + 4] = Int32GetDatum(atoi(val));
			pfree(val);

			// Execute insert plan
			ret = SPI_execp(insert_plan, cvals, cnulls, 0);
			SPI_freetuptable(SPI_tuptable);
			if (ret < 0) 
			{
				elog(ERROR, "Trio: SPI_execp returned %d", ret);
				return 0;
			}
		}
		SPI_execute("commit", true, 0);
	}

	// Clean-up
	SPI_finish();

	return 1;
}

int
trio_mergeinto_distinct(text* table, text* query1, text* query2, text* columns, text* columnmappings, 
						int setop_type, int table_type, int nolineage1, int nolineage2, int table_type2, int compute_conf)
{
	Oid* ctypes, *utypes, *ltypes;	/* column types */
	Datum* cvals, *uvals, *lvals;	/* column values */

	HeapTuple tuple1, tuple2, tmp1, tmp2;

	void* insert_c_plan = NULL;
	void* insert_u_plan = NULL;
	void* insert_l_plan;

	char* tablename;
	char* select_stmt1;
	char* select_stmt2;
	char* column_list;
	char* column_mappings;

	char* val;
	char* cnulls;
	char* unulls;

	char stmt[8192];
	char insert_c_stmt[8192];
	char insert_u_stmt[8192];
	char insert_l_stmt[8192];
	char separ;

	int i, ret = 0, xid = 1, aid = 1, prev_xid = 1, c = 0, u = 0;

	stmt[0] = '\0';
	insert_c_stmt[0] = '\0';
	insert_u_stmt[0] = '\0';
	insert_l_stmt[0] = '\0';

	// Create an SPI server connection first
	SPI_connect();

	// Translate parameters
	tablename = DatumGetCString(DirectFunctionCall1(textout,
		PointerGetDatum(table)));
	select_stmt1 = DatumGetCString(DirectFunctionCall1(textout,
		PointerGetDatum(query1)));
	select_stmt2 = DatumGetCString(DirectFunctionCall1(textout,
		PointerGetDatum(query2)));
	column_list = DatumGetCString(DirectFunctionCall1(textout,
		PointerGetDatum(columns)));
	column_mappings = DatumGetCString(DirectFunctionCall1(textout,
		PointerGetDatum(columnmappings)));

	if (init_fetch(select_stmt1, select_stmt2, NULL, column_list, column_mappings, table_type))
	{
		// Update XID and AID counters
		if (table_type == -1 || num_u_datacolumns > 0) 
		{
			snprintf(stmt, sizeof(stmt), "SELECT nextval('trio_xid_%s'), nextval('trio_aid_%s')", tablename, tablename);
		} 
		else 
		{
			snprintf(stmt, sizeof(stmt), "SELECT nextval('trio_xid_%s')", tablename);
		}
		SPI_execute(stmt, true, 0);
		val = SPI_getvalue(SPI_tuptable->vals[0], SPI_tuptable->tupdesc, 1);
		if (val != NULL) 
		{
			prev_xid = atoi(val);
			xid = prev_xid;
			pfree(val);
		}
		if (table_type == -1 || num_u_datacolumns > 0) 
		{
			val = SPI_getvalue(SPI_tuptable->vals[0], SPI_tuptable->tupdesc, 2);
			if (val != NULL) 
			{
				aid = atoi(val);
				pfree(val);
			}
		}
		SPI_freetuptable(SPI_tuptable);

		// Allocate data & types of prepared statement for certain attributes
		cvals = (Datum*) palloc((num_c_datacolumns+num_u_datacolumns+5) * sizeof(Datum));
		ctypes = (Oid *) palloc((num_c_datacolumns+num_u_datacolumns+5) * sizeof(Oid));
		cnulls = (char*) palloc((num_c_datacolumns+num_u_datacolumns+5) * sizeof(char));

		// Allocate data & types for prepared statement for uncertain attributes
		uvals = (Datum*) palloc((num_c_datacolumns+num_u_datacolumns+5) * sizeof(Datum));
		utypes = (Oid *) palloc((num_c_datacolumns+num_u_datacolumns+5) * sizeof(Oid));
		unulls = (char*) palloc((num_c_datacolumns+num_u_datacolumns+5) * sizeof(char));

		// Allocate data & types for prepared statement for lineage
		lvals = (Datum*) palloc(4 * sizeof(Datum));
		ltypes = (Oid *) palloc(4 * sizeof(Oid));


		// Construct data query for certain attributes: INSERT INTO table VALUES ($1, ...)
		separ = '(';
		snprintf(insert_c_stmt, sizeof(insert_c_stmt), "INSERT INTO trio_c_%s (%s%s", tablename, column_c_list, num_c_datacolumns > 0 ? "," : "");
		snprintf(insert_u_stmt, sizeof(insert_u_stmt), "INSERT INTO %s%s (%s%s", table_type > -1 ? "trio_u_" : "", tablename, column_u_list, num_u_datacolumns > 0 ? "," : "");
		if (table_type == -1) // temporary table
		{
			snprintf(insert_u_stmt + strlen(insert_u_stmt), sizeof(insert_u_stmt) - strlen(insert_u_stmt), "trio_xid,trio_aid,trio_alts,trio_conf,trio_enc) VALUES ");
			for (i = 0; i < num_u_datacolumns; i++)
			{
				utypes[i] = SPI_gettypeid(fetch_tuptable->tupdesc, u_datacolumns[i]+1);
				snprintf(insert_u_stmt + strlen(insert_u_stmt), sizeof(insert_u_stmt) - strlen(insert_u_stmt), "%c$%d", separ, i+1);
				separ = ',';
			}
			utypes[i] = INT4OID; // xid
			utypes[i+1] = INT4OID; // aid
			utypes[i+2] = INT4OID; // alts
			if (compute_conf == 1)
			{
				utypes[i+3] = INT4OID; // aid for conf() call
				snprintf(insert_u_stmt + strlen(insert_u_stmt), sizeof(insert_u_stmt) - strlen(insert_u_stmt), 
					",$%d,$%d,$%d,trio_get_conf(array['%s'],array[$%d]::integer[],array[1],NULL::float[],0,1),$%d)", i+1, i+2, i+3, tablename, i+4, i+5);
			}
			else
			{
				utypes[i+3] = FLOAT8OID; // conf
				snprintf(insert_u_stmt + strlen(insert_u_stmt), sizeof(insert_u_stmt) - strlen(insert_u_stmt), ",$%d,$%d,$%d,$%d,$%d)", i+1, i+2, i+3, i+4, i+5);
			}
			utypes[i+4] = INT4OID; // enc
			c = 0;
			u = 5;
		} 
		else if (table_type == 1)
		{
			snprintf(insert_c_stmt + strlen(insert_c_stmt), sizeof(insert_c_stmt) - strlen(insert_c_stmt), "trio_xid) VALUES ");
			for (i = 0; i < num_c_datacolumns; i++)
			{
				ctypes[i] = SPI_gettypeid(fetch_tuptable->tupdesc, c_datacolumns[i]+1);
				snprintf(insert_c_stmt + strlen(insert_c_stmt), sizeof(insert_c_stmt) - strlen(insert_c_stmt), "%c$%d", separ, i+1);
				separ = ',';
			}
			ctypes[i] = INT4OID; // xid
			snprintf(insert_c_stmt + strlen(insert_c_stmt), sizeof(insert_c_stmt) - strlen(insert_c_stmt), ",$%d)", i+1);
			c = 1;
			u = 0;
		}
		else if (table_type == 3)
		{
			snprintf(insert_c_stmt + strlen(insert_c_stmt), sizeof(insert_c_stmt) - strlen(insert_c_stmt), "trio_xid,trio_conf) VALUES ");
			for (i = 0; i < num_c_datacolumns; i++)
			{
				ctypes[i] = SPI_gettypeid(fetch_tuptable->tupdesc, c_datacolumns[i]+1);
				snprintf(insert_c_stmt + strlen(insert_c_stmt), sizeof(insert_c_stmt) - strlen(insert_c_stmt), "%c$%d", separ, i+1);
				separ = ',';
			}
			ctypes[i] = INT4OID; // xid
			if (compute_conf == 1)
			{
				ctypes[i+1] = INT4OID; // aid for conf() call
				snprintf(insert_c_stmt + strlen(insert_c_stmt), sizeof(insert_c_stmt) - strlen(insert_c_stmt), 
					",$%d,trio_get_conf(array['%s'],array[$%d]::integer[],array[1],NULL::float[],0,1))", i+1, tablename, i+2);
			}
			else
			{
				ctypes[i+1] = FLOAT8OID; // conf
				snprintf(insert_c_stmt + strlen(insert_c_stmt), sizeof(insert_c_stmt) - strlen(insert_u_stmt), ",$%d,$%d)", i+1, i+2);
			}
			c = 2;
			u = 0;
		} 
		else
		{
			snprintf(insert_c_stmt + strlen(insert_c_stmt), sizeof(insert_c_stmt) - strlen(insert_c_stmt), "trio_xid,trio_alts,trio_enc) VALUES ");
			for (i = 0; i < num_c_datacolumns; i++)
			{
				ctypes[i] = SPI_gettypeid(fetch_tuptable->tupdesc, c_datacolumns[i]+1);
				snprintf(insert_c_stmt + strlen(insert_c_stmt), sizeof(insert_c_stmt) - strlen(insert_c_stmt), "%c$%d", separ, i+1);
				separ = ',';
			}
			ctypes[i] = INT4OID; // xid
			ctypes[i+1] = INT4OID; // alts
			ctypes[i+2] = INT4OID; // enc
			snprintf(insert_c_stmt + strlen(insert_c_stmt), sizeof(insert_c_stmt) - strlen(insert_c_stmt), "%s$%d,$%d,$%d)", num_c_datacolumns > 0 ? "," : "(", i+1, i+2, i+3);
			c = 3;
			separ = '(';
			if (table_type == 0)
			{
				snprintf(insert_u_stmt + strlen(insert_u_stmt), sizeof(insert_u_stmt) - strlen(insert_u_stmt), "trio_xid,trio_aid) VALUES ");
				for (i = 0; i < num_u_datacolumns; i++)
				{
					utypes[i] = SPI_gettypeid(fetch_tuptable->tupdesc, u_datacolumns[i]+1);
					snprintf(insert_u_stmt + strlen(insert_u_stmt), sizeof(insert_u_stmt) - strlen(insert_u_stmt), "%c$%d", separ, i+1);
					separ = ',';
				}
				utypes[i] = INT4OID; // xid
				utypes[i+1] = INT4OID; // aid
				snprintf(insert_u_stmt + strlen(insert_u_stmt), sizeof(insert_u_stmt) - strlen(insert_u_stmt), "%s$%d,$%d)", num_u_datacolumns > 0 ? "," : "(", i+1, i+2);
				u = 2;
			} 
			else if (table_type == 2) 
			{
				snprintf(insert_u_stmt + strlen(insert_u_stmt), sizeof(insert_u_stmt) - strlen(insert_u_stmt), "trio_xid,trio_aid,trio_conf) VALUES ");
				for (i = 0; i < num_u_datacolumns; i++)
				{
					utypes[i] = SPI_gettypeid(fetch_tuptable->tupdesc, u_datacolumns[i]+1);
					snprintf(insert_u_stmt + strlen(insert_u_stmt), sizeof(insert_u_stmt) - strlen(insert_u_stmt), "%c$%d", separ, i+1);
					separ = ',';
				}
				utypes[i] = INT4OID; // xid
				utypes[i+1] = INT4OID; // aid
				if (compute_conf == 1)
				{
					utypes[i+2] = INT4OID; // aid for conf() call
					snprintf(insert_u_stmt + strlen(insert_u_stmt), sizeof(insert_u_stmt) - strlen(insert_u_stmt), 
						"%s$%d,$%d,trio_get_conf(array['%s'],array[$%d]::integer[],array[1],NULL::float[],0,1))", num_u_datacolumns > 0 ? "," : "(", i+1, i+2, tablename, i+3);
				}
				else
				{
					utypes[i+2] = FLOAT8OID; // conf
					snprintf(insert_u_stmt + strlen(insert_u_stmt), sizeof(insert_u_stmt) - strlen(insert_u_stmt), "%s$%d,$%d,$%d)", num_u_datacolumns > 0 ? "," : "(", i+1, i+2, i+3);
				}
				u = 3;
			}
		}

		// Construct lineage query: INSERT INTO trio_lin_table VALUES ($1, ...)
		snprintf(insert_l_stmt, sizeof(insert_l_stmt), "INSERT INTO trio_lin_%s (aid, disjlin, srclin, srctable) VALUES ($1,$2,$3,lower($4))", tablename);
		ltypes[0] = INT4OID;
		ltypes[1] = INT4OID;
		ltypes[2] = INT4OID;
		ltypes[3] = VARCHAROID;
		
		// Prepare plan for certain data insert statement
		if (table_type > -1) 
		{
			insert_c_plan = SPI_prepare(insert_c_stmt, num_c_datacolumns+c, ctypes);
			if (insert_c_plan == NULL)
				elog(ERROR, "Trio: create table %s returned %d", tablename, SPI_result);
		}

		// Prepare plan for uncertain data insert statement
		if (table_type == -1 || table_type == 0 || table_type == 2) {
			insert_u_plan = SPI_prepare(insert_u_stmt, num_u_datacolumns+u, utypes);
			if (insert_u_plan == NULL)
				elog(ERROR, "Trio: create table %s returned %d", tablename, SPI_result);
		}

		// Prepare plan for lineage insert statement
		insert_l_plan = SPI_prepare(insert_l_stmt, 4, ltypes);
		if (insert_l_plan == NULL)
			elog(ERROR, "Trio: create table %s returned %d", tablename, SPI_result);

		// Begin of transaction
		SPI_execute("begin", true, 0);
		
		int c1 = 0, c2 = 0, num_datacolumns = num_c_datacolumns+num_u_datacolumns;
		int aid1 = 0, aid2 = 0, disj1 = 0, disj2 = 0, disj = 0, tmp_xid1, tmp_xid2, tmp_aid1, tmp_aid2, delta = 1, enc;
		bool isnull, skip_certain;

		if (setop_type == 1) // isIntersect
			delta = 2;
		else if (setop_type == 2) // isExcept
			delta = 3;

		SPITupleTable* prev_fetch_tuptable = NULL, *prev_fetch_tuptable_setops = NULL;

		tuple1 = NULL;
		tuple2 = NULL;
		if (fetch_processed>0)
			tuple1 = fetch_tuptable->vals[0];
		if (fetch_processed_setops>0)
			tuple2 = fetch_tuptable_setops->vals[0];

		num_tuples = 0;
		while(tuple1 != NULL || (setop_type == 0 && tuple2 != NULL)) 
		{
			skip_certain = false;

			// Move cursor sequentially through left-hand result set, insert data & lineage
            while (tuple1 != NULL)
			{
                if (tuple2 != NULL && compare_values(tuple1, tuple2, num_datacolumns) >= 0)
                    break;
				//elog(INFO, "%s < %s", SPI_getvalue(tuple1, fetch_tuptable->tupdesc, 1), SPI_getvalue(tuple2, fetch_tuptable_setops->tupdesc, 1));
				tmp1 = tuple1;
				if (c1 == fetch_processed)
				{
					c1 = 1;
					if (prev_fetch_tuptable != NULL)
						SPI_freetuptable(prev_fetch_tuptable);
					prev_fetch_tuptable = fetch_tuptable;
					SPI_cursor_fetch(fetch_portal, true, PREFETCH_BUFFER);
					fetch_tuptable = SPI_tuptable;
					fetch_processed = SPI_processed;
					if (fetch_tuptable == NULL || fetch_processed == 0) 
						tuple1 = NULL;
					else
						tuple1 = fetch_tuptable->vals[0];
				} 
				else
				{
					tuple1 = fetch_tuptable->vals[c1];
					c1++;
				}
				if (setop_type != 1) //isUnion or isExcept
				{
					val = SPI_getvalue(tmp1, fetch_tuptable->tupdesc, num_datacolumns+1);
					tmp_aid1 = atoi(val);
					pfree(val);
                    if (!nolineage1)
					{
						// Enumerate lineage entries
						lvals[0] = Int32GetDatum(tmp_aid1);
						val = SPI_getvalue(tmp1, fetch_tuptable->tupdesc, num_datacolumns+6);
						disj1 = atoi(val);
						lvals[1] = Int32GetDatum(disj1);
						pfree(val);
						val = SPI_getvalue(tmp1, fetch_tuptable->tupdesc, num_datacolumns+7);
						lvals[2] = Int32GetDatum(atoi(val));
						pfree(val);
						val = SPI_getvalue(tmp1, fetch_tuptable->tupdesc, num_datacolumns+8);
						lvals[3] = DirectFunctionCall1(textin, CStringGetDatum(val));
						pfree(val);

						// Execute lineage insert statement
						ret = SPI_execp(insert_l_plan, lvals, NULL, 0);
						//pfree(DatumGetPointer(lvals[3]));
						SPI_freetuptable(SPI_tuptable);
						if (ret < 0) 
						{
							elog(ERROR, "Trio: SPI_execp returned %d", ret);
							return 0;
						}
					}
					if (tuple1 != NULL)
					{
						val = SPI_getvalue(tuple1, fetch_tuptable->tupdesc, num_datacolumns+1);
						aid1 = atoi(val);
						pfree(val);
					}
					if (tuple1 == NULL || aid1 != tmp_aid1) // assumes vertical data dups eliminated
					{
						// Set certain data columns for new xtuple
						for(i = 0; i < num_c_datacolumns; i++)
						{
							cvals[i] = SPI_getbinval(tmp1, fetch_tuptable->tupdesc, c_datacolumns[i]+1, &isnull);
							if (isnull == 0) {
								//val = SPI_getvalue(tmp1, fetch_tuptable->tupdesc, c_datacolumns[i]+1);
								//cvals[i] = DirectFunctionCall1(textin, CStringGetDatum(val));
								//pfree(val);
								cnulls[i] = ' ';
							}
							else 
							{
								cnulls[i] = 'n';
							}
						}

						// Set uncertain data columns for new alternative
						for(i = 0; i < num_u_datacolumns; i++)
						{
							uvals[i] = SPI_getbinval(tmp1, fetch_tuptable->tupdesc, u_datacolumns[i]+1, &isnull);
							if (isnull == 0)
							{
								//val = SPI_getvalue(tmp1, fetch_tuptable->tupdesc, u_datacolumns[i]+1);
								//uvals[i] = DirectFunctionCall1(textin, CStringGetDatum(val));
								//pfree(val);
								unulls[i] = ' ';
							}
							else 
							{
								unulls[i] = 'n';
							}
						}

						val = SPI_getvalue(tmp1, fetch_tuptable->tupdesc, num_datacolumns+2);
						tmp_xid1 = atoi(val);
						pfree(val);
						cvals[num_c_datacolumns] = Int32GetDatum(tmp_xid1); // xid
						uvals[num_u_datacolumns] = Int32GetDatum(tmp_xid1); // xid
						uvals[num_u_datacolumns+1] = Int32GetDatum(tmp_aid1); // aid
						if (table_type == -1)
						{
							val = SPI_getvalue(tmp1, fetch_tuptable->tupdesc, num_datacolumns+3);
							uvals[num_u_datacolumns+2] = Int32GetDatum(atoi(val)); // alts
							pfree(val);
							if (compute_conf == 1)
							{ 
								uvals[num_u_datacolumns+3] = Int32GetDatum(tmp_aid1); // aid for conf() call
								unulls[num_u_datacolumns+3] = ' '; 
							}
							else
							{
								unulls[num_u_datacolumns+3] = 'n';
							}
							val = SPI_getvalue(tmp1, fetch_tuptable->tupdesc, num_datacolumns+4);
							uvals[num_u_datacolumns+4] = Int32GetDatum(atoi(val)); // enc
							pfree(val);
						} 
						else if (table_type == 3) 
						{
							if (compute_conf == 1)
							{ 
								cvals[num_c_datacolumns+1] = Int32GetDatum(tmp_aid1); // aid for conf() call
								cnulls[num_c_datacolumns+1] = ' '; 
							}
							else
							{
								cnulls[num_c_datacolumns+1] = 'n';
							}
						}
						else if (table_type != 1)
						{
							val = SPI_getvalue(tmp1, fetch_tuptable->tupdesc, num_datacolumns+3);
							cvals[num_c_datacolumns+1] = Int32GetDatum(atoi(val)); // alts
							pfree(val);
							val = SPI_getvalue(tmp1, fetch_tuptable->tupdesc, num_datacolumns+4);
							cvals[num_c_datacolumns+2] = Int32GetDatum(atoi(val)); // enc
							pfree(val);
							if (compute_conf == 1)
							{ 
								uvals[num_u_datacolumns+2] = Int32GetDatum(tmp_aid1); // aid for conf() call
								unulls[num_u_datacolumns+2] = ' '; 
							}
							else
							{
								unulls[num_u_datacolumns+2] = 'n';
							}
						}

						// Execute data insert statement for certain attributes
						if (table_type > -1)
						{
							ret = SPI_execp(insert_c_plan, cvals, cnulls, 0);
							SPI_freetuptable(SPI_tuptable);
							if (ret < 0) 
							{
								elog(ERROR, "Trio: SPI_execp returned %d", ret);
								return 0;
							}
						}
						
						// Execute data insert statement for uncertain attributes
						if (table_type != 1 && table_type != 3)
						{
							ret = SPI_execp(insert_u_plan, uvals, unulls, 0);
							SPI_freetuptable(SPI_tuptable);
							if (ret < 0) 
							{
								elog(ERROR, "Trio: SPI_execp returned %d", ret);
								return 0;
							}
						} 
						
						// Periodically commit
						if (++num_tuples % 5000) 
						{
							SPI_execute("commit", true, 0);
							SPI_execute("begin", true, 0);
						}
					}
				}
			}
			
			// Move cursor sequentially through right-hand result set, insert data & lineage
			while (tuple2 != NULL)
			{
				if (tuple1 != NULL && compare_values(tuple1, tuple2, num_datacolumns) <= 0)
					break;
				//elog(INFO, "%s > %s", SPI_getvalue(tuple1, fetch_tuptable->tupdesc, 1), SPI_getvalue(tuple2, fetch_tuptable_setops->tupdesc, 1));
				tmp2 = tuple2;
				if (c2 == fetch_processed_setops)
				{
					c2 = 1;
					if (prev_fetch_tuptable_setops != NULL)
						SPI_freetuptable(prev_fetch_tuptable_setops);
					prev_fetch_tuptable_setops = fetch_tuptable_setops;
					SPI_cursor_fetch(fetch_portal_setops, true, PREFETCH_BUFFER);
					fetch_tuptable_setops = SPI_tuptable;
					fetch_processed_setops = SPI_processed;
					if (fetch_tuptable_setops == NULL || fetch_processed_setops == 0) 
						tuple2 = NULL;
					else
						tuple2 = fetch_tuptable_setops->vals[0];
				} 
				else
				{
					tuple2 = fetch_tuptable_setops->vals[c2];
					c2++;
				}
				if (setop_type == 2) // isExcept
				{
					if (table_type2 != 2)
					{
						val = SPI_getvalue(tmp2, fetch_tuptable_setops->tupdesc, num_datacolumns+3);
						if (atoi(val) == 1)
							skip_certain = true;
						pfree(val);
					}
					else
					{
						SPI_getbinval(tmp2, fetch_tuptable_setops->tupdesc, num_datacolumns+5, &isnull);
						if (isnull==0)
						{
							val = SPI_getvalue(tmp2, fetch_tuptable_setops->tupdesc, num_datacolumns+5);
							if (atof(val) == 1.0)
								skip_certain = true;
							pfree(val);
						}
					}
				}
                else if  (setop_type == 0) // isUnion
				{
					val = SPI_getvalue(tmp2, fetch_tuptable_setops->tupdesc, num_datacolumns+1);
					tmp_aid2 = atoi(val);
					pfree(val);
					if (!nolineage2)
					{
						// Enumerate lineage entries
						lvals[0] = Int32GetDatum(aid + tmp_aid2);
						val = SPI_getvalue(tmp2, fetch_tuptable_setops->tupdesc, num_datacolumns+6);
						disj2 = atoi(val);
						lvals[1] = Int32GetDatum(disj2);
						pfree(val);
						val = SPI_getvalue(tmp2, fetch_tuptable_setops->tupdesc, num_datacolumns+7);
						lvals[2] = Int32GetDatum(atoi(val));
						pfree(val);
						val = SPI_getvalue(tmp2, fetch_tuptable_setops->tupdesc, num_datacolumns+8);
						lvals[3] = DirectFunctionCall1(textin, CStringGetDatum(val));
						pfree(val);

						// Execute lineage insert statement
						ret = SPI_execp(insert_l_plan, lvals, NULL, 0);
						//pfree(DatumGetPointer(lvals[3]));
						SPI_freetuptable(SPI_tuptable);
						if (ret < 0) 
						{
							elog(ERROR, "Trio: SPI_execp returned %d", ret);
							return 0;
						}
					}
					if (tuple2 != NULL)
					{
						val = SPI_getvalue(tuple2, fetch_tuptable_setops->tupdesc, num_datacolumns+1);
						aid2 = atoi(val);
						pfree(val);
					}
					if (tuple2 == NULL || aid2 != tmp_aid2)
					{
						// Set certain data columns for new xtuple
						for(i = 0; i < num_c_datacolumns; i++)
						{
							cvals[i] = SPI_getbinval(tmp2, fetch_tuptable_setops->tupdesc, c_datacolumns[i]+1, &isnull);
							if (isnull == 0) {
								//val = SPI_getvalue(tmp2, fetch_tuptable_setops->tupdesc, c_datacolumns[i]+1);
								//cvals[i] = DirectFunctionCall1(textin, CStringGetDatum(val));
								//pfree(val);
								cnulls[i] = ' ';
							}
							else 
							{
								cnulls[i] = 'n';
							}
						}

						// Set uncertain data columns for new alternative
						for(i = 0; i < num_u_datacolumns; i++)
						{
							uvals[i] = SPI_getbinval(tmp2, fetch_tuptable_setops->tupdesc, u_datacolumns[i]+1, &isnull);
							if (isnull == 0)
							{
								//val = SPI_getvalue(tmp2, fetch_tuptable_setops->tupdesc, u_datacolumns[i]+1);
								//uvals[i] = DirectFunctionCall1(textin, CStringGetDatum(val));
								//pfree(val);
								unulls[i] = ' ';
							}
							else 
							{
								unulls[i] = 'n';
							}
						}

						val = SPI_getvalue(tmp2, fetch_tuptable_setops->tupdesc, num_datacolumns+2);
						tmp_xid2 = atoi(val);
						pfree(val);
						cvals[num_c_datacolumns] = Int32GetDatum(xid + tmp_xid2); // xid
						uvals[num_u_datacolumns] = Int32GetDatum(xid + tmp_xid2); // xid
						uvals[num_u_datacolumns+1] = Int32GetDatum(aid + tmp_aid2); // aid
						if (table_type == -1)
						{
							val = SPI_getvalue(tmp2, fetch_tuptable_setops->tupdesc, num_datacolumns+3);
							uvals[num_u_datacolumns+2] = Int32GetDatum(atoi(val)); // alts
							pfree(val);
							if (compute_conf == 1)
							{ 
								uvals[num_u_datacolumns+3] = Int32GetDatum(aid + tmp_aid2); // aid for conf() call
								unulls[num_u_datacolumns+3] = ' '; 
							}
							else
							{
								unulls[num_u_datacolumns+3] = 'n';
							}
							val = SPI_getvalue(tmp2, fetch_tuptable_setops->tupdesc, num_datacolumns+4);
							uvals[num_u_datacolumns+4] = Int32GetDatum(atoi(val)); // enc
							pfree(val);
						} 
						else if (table_type == 3) 
						{
							if (compute_conf == 1)
							{ 
								cvals[num_c_datacolumns+1] = Int32GetDatum(aid + tmp_aid2); // aid for conf() call
								cnulls[num_c_datacolumns+1] = ' '; 
							}
							else
							{
								cnulls[num_c_datacolumns+1] = 'n';
							}
						}
						else if (table_type != 1)
						{
							val = SPI_getvalue(tmp2, fetch_tuptable_setops->tupdesc, num_datacolumns+3);
							cvals[num_c_datacolumns+1] = Int32GetDatum(atoi(val)); // alts
							pfree(val);
							val = SPI_getvalue(tmp2, fetch_tuptable_setops->tupdesc, num_datacolumns+4);
							cvals[num_c_datacolumns+2] = Int32GetDatum(atoi(val)); // enc
							pfree(val);
							if (compute_conf == 1)
							{ 
								uvals[num_u_datacolumns+2] = Int32GetDatum(aid + tmp_aid2); // aid for conf() call
								unulls[num_u_datacolumns+2] = ' '; 
							}
							else
							{
								unulls[num_u_datacolumns+2] = 'n';
							}
						}

						// Execute data insert statement for certain attributes
						if (table_type > -1)
						{
							ret = SPI_execp(insert_c_plan, cvals, cnulls, 0);
							SPI_freetuptable(SPI_tuptable);
							if (ret < 0) 
							{
								elog(ERROR, "Trio: SPI_execp returned %d", ret);
								return 0;
							}
						}
						
						// Execute data insert statement for uncertain attributes
						if (table_type != 1 && table_type != 3)
						{
							ret = SPI_execp(insert_u_plan, uvals, unulls, 0);
							SPI_freetuptable(SPI_tuptable);
							if (ret < 0) 
							{
								elog(ERROR, "Trio: SPI_execp returned %d", ret);
								return 0;
							}
						}

						// Periodically commit
						if (++num_tuples % 5000) 
						{
							SPI_execute("commit", true, 0);
							SPI_execute("begin", true, 0);
						}
					}
				}
			}

			// Insert lineage of matching alternative from left-hand query
			tmp1 = NULL;
            while (tuple1 != NULL && tuple2 != NULL && compare_values(tuple1, tuple2, num_datacolumns) == 0)
			{
				//elog(INFO, "%s =1 %s", SPI_getvalue(tuple1, fetch_tuptable->tupdesc, 1), SPI_getvalue(tuple2, fetch_tuptable_setops->tupdesc, 1));
                tmp1 = tuple1;
				if (c1 == fetch_processed)
				{
					c1 = 1;
					if (prev_fetch_tuptable != NULL)
						SPI_freetuptable(prev_fetch_tuptable);
					prev_fetch_tuptable = fetch_tuptable;
					SPI_cursor_fetch(fetch_portal, true, PREFETCH_BUFFER);
					fetch_tuptable = SPI_tuptable;
					fetch_processed = SPI_processed;
					if (fetch_tuptable == NULL || fetch_processed == 0) 
						tuple1 = NULL;
					else
						tuple1 = fetch_tuptable->vals[0];
				} 
				else
				{
					tuple1 = fetch_tuptable->vals[c1];
					c1++;
				}
                if (!skip_certain && !nolineage1)
				{
					// Enumerate lineage entries
					val = SPI_getvalue(tmp1, fetch_tuptable->tupdesc, num_datacolumns+1);
					aid1 = atoi(val);
					lvals[0] = Int32GetDatum(aid1);
					pfree(val);
					val = SPI_getvalue(tmp1, fetch_tuptable->tupdesc, num_datacolumns+6);
					disj1 = atoi(val);
					lvals[1] = Int32GetDatum(disj1);
					pfree(val);
					val = SPI_getvalue(tmp1, fetch_tuptable->tupdesc, num_datacolumns+7);
					lvals[2] = Int32GetDatum(atoi(val));
					pfree(val);
					val = SPI_getvalue(tmp1, fetch_tuptable->tupdesc, num_datacolumns+8);
					lvals[3] = DirectFunctionCall1(textin, CStringGetDatum(val));
					pfree(val);

					// Execute lineage insert statement
					ret = SPI_execp(insert_l_plan, lvals, NULL, 0);
					//pfree(DatumGetPointer(lvals[3]));
					SPI_freetuptable(SPI_tuptable);
					if (ret < 0) 
					{
						elog(ERROR, "Trio: SPI_execp returned %d", ret);
						return 0;
					}
				}
			}

			// Insert data and lineage of matching alternative from right-hand query to left-hand alternative
			disj1 = 0;
            if (tuple2 != NULL)
			{
				val = SPI_getvalue(tuple2, fetch_tuptable_setops->tupdesc, num_datacolumns+6);
				disj = atoi(val);
				pfree(val);
			}
			tmp_aid1 = 0;
			if (tmp1 != NULL)
			{
				val = SPI_getvalue(tmp1, fetch_tuptable->tupdesc, num_datacolumns+1);
				tmp_aid1 = atoi(val);
				pfree(val);
				val = SPI_getvalue(tmp1, fetch_tuptable->tupdesc, num_datacolumns+6);
				disj1 = atoi(val);
				pfree(val);
			}
            while (tmp1 != NULL && tuple2 != NULL && compare_values(tmp1, tuple2, num_datacolumns) == 0)
			{
				//elog(INFO, "%s =2 %s", SPI_getvalue(tmp1, fetch_tuptable->tupdesc, 1), SPI_getvalue(tuple2, fetch_tuptable_setops->tupdesc, 1));
                tmp2 = tuple2;
				val = SPI_getvalue(tuple2, fetch_tuptable_setops->tupdesc, num_datacolumns+1);
				tmp_aid2 = atoi(val);
				pfree(val);
                if (c2 == fetch_processed_setops)
				{
					c2 = 1;
					if (prev_fetch_tuptable_setops != NULL)
						SPI_freetuptable(prev_fetch_tuptable_setops);
					prev_fetch_tuptable_setops = fetch_tuptable_setops;
					SPI_cursor_fetch(fetch_portal_setops, true, PREFETCH_BUFFER);
					fetch_tuptable_setops = SPI_tuptable;
					fetch_processed_setops = SPI_processed;
					if (fetch_tuptable_setops == NULL || fetch_processed_setops == 0) 
						tuple2 = NULL;
					else
						tuple2 = fetch_tuptable_setops->vals[0];
				}
				else
				{
					tuple2 = fetch_tuptable_setops->vals[c2];
					c2++;
				}
				
                if (!skip_certain)
				{
                    if (!nolineage2)
					{
						// Enumerate lineage entries
						lvals[0] = Int32GetDatum(aid1);
						val = SPI_getvalue(tmp2, fetch_tuptable_setops->tupdesc, num_datacolumns+6);
						disj2 = atoi(val);
						lvals[1] = Int32GetDatum(disj1 + delta + disj2 - disj);
						pfree(val);
						val = SPI_getvalue(tmp2, fetch_tuptable_setops->tupdesc, num_datacolumns+7);
						lvals[2] = Int32GetDatum(atoi(val));
						pfree(val);
						val = SPI_getvalue(tmp2, fetch_tuptable_setops->tupdesc, num_datacolumns+8);
						lvals[3] = DirectFunctionCall1(textin, CStringGetDatum(val));
						pfree(val);

						// Execute lineage insert statement
						ret = SPI_execp(insert_l_plan, lvals, NULL, 0);
						//pfree(DatumGetPointer(lvals[3]));
						SPI_freetuptable(SPI_tuptable);
						if (ret < 0) 
						{
							elog(ERROR, "Trio: SPI_execp returned %d", ret);
							return 0;
						}
					}
					if (tuple2 != NULL)
					{
						val = SPI_getvalue(tuple2, fetch_tuptable_setops->tupdesc, num_datacolumns+1);
						aid2 = atoi(val);
						pfree(val);
					}
					if (tuple2 == NULL || aid2 != tmp_aid2) // assumes vertical data dups eliminated
					{
						// Set certain data columns for new xtuple
						for(i = 0; i < num_c_datacolumns; i++)
						{
							cvals[i] = SPI_getbinval(tmp1, fetch_tuptable->tupdesc, c_datacolumns[i]+1, &isnull);
							if (isnull == 0) {

								//val = SPI_getvalue(tmp1, fetch_tuptable->tupdesc, c_datacolumns[i]+1);
								//cvals[i] = DirectFunctionCall1(textin, CStringGetDatum(val));
								//pfree(val);
								cnulls[i] = ' ';
							}
							else 
							{
								cnulls[i] = 'n';
							}
						}

						// Set uncertain data columns for new alternative
						for(i = 0; i < num_u_datacolumns; i++)
						{
							uvals[i] = SPI_getbinval(tmp1, fetch_tuptable->tupdesc, u_datacolumns[i]+1, &isnull);
							if (isnull == 0)
							{
								//val = SPI_getvalue(tmp1, fetch_tuptable->tupdesc, u_datacolumns[i]+1);
								//uvals[i] = DirectFunctionCall1(textin, CStringGetDatum(val));
								//pfree(val);
								unulls[i] = ' ';
							}
							else 
							{
								unulls[i] = 'n';
							}
						}

						val = SPI_getvalue(tmp1, fetch_tuptable->tupdesc, num_datacolumns+2);
						tmp_xid1 = atoi(val);
						pfree(val);
					
						val = SPI_getvalue(tmp1, fetch_tuptable->tupdesc, num_datacolumns+4);
						enc = atoi(val);
						pfree(val);
						// conservatively add a ? for EXCEPT/INTERSECT
						if (setop_type > 0) 
							enc++;

						cvals[num_c_datacolumns] = Int32GetDatum(tmp_xid1); // xid
						uvals[num_u_datacolumns] = Int32GetDatum(tmp_xid1); // xid
						uvals[num_u_datacolumns+1] = Int32GetDatum(tmp_aid1); // aid
						if (table_type == -1)
						{
							val = SPI_getvalue(tmp1, fetch_tuptable->tupdesc, num_datacolumns+3);
							uvals[num_u_datacolumns+2] = Int32GetDatum(atoi(val)); // alts
							pfree(val);
							if (compute_conf == 1)
							{ 
								uvals[num_u_datacolumns+3] = Int32GetDatum(tmp_aid1); // aid for conf() call
								unulls[num_u_datacolumns+3] = ' '; 
							}
							else
							{
								unulls[num_u_datacolumns+3] = 'n';
							}
							uvals[num_u_datacolumns+4] = Int32GetDatum(enc); // enc
						} 
						else if (table_type == 3) 
						{
							if (compute_conf == 1)
							{ 
								cvals[num_c_datacolumns+1] = Int32GetDatum(tmp_aid1); // aid for conf() call
								cnulls[num_c_datacolumns+1] = ' '; 
							}
							else
							{
								cnulls[num_c_datacolumns+1] = 'n';
							}
						}
						else if (table_type != 1)
						{
							val = SPI_getvalue(tmp1, fetch_tuptable->tupdesc, num_datacolumns+3);
							cvals[num_c_datacolumns+1] = Int32GetDatum(atoi(val)); // alts
							pfree(val);
							cvals[num_c_datacolumns+2] = Int32GetDatum(enc); // enc
							if (compute_conf == 1)
							{ 
								uvals[num_u_datacolumns+2] = Int32GetDatum(tmp_aid1); // aid for conf() call
								unulls[num_u_datacolumns+2] = ' '; 
							}
							else
							{
								unulls[num_u_datacolumns+2] = 'n';
							}
						}

						// Execute data insert statement for certain attributes
						if (table_type > -1)
						{
							ret = SPI_execp(insert_c_plan, cvals, cnulls, 0);
							SPI_freetuptable(SPI_tuptable);
							if (ret < 0) 
							{
								elog(ERROR, "Trio: SPI_execp returned %d", ret);
								return 0;
							}
						}

						// Execute data insert statement for uncertain attributes
						if (table_type != 1 && table_type != 3)
						{
							ret = SPI_execp(insert_u_plan, uvals, unulls, 0);
							SPI_freetuptable(SPI_tuptable);
							if (ret < 0) 
							{
								elog(ERROR, "Trio: SPI_execp returned %d", ret);
								return 0;
							}
						}

						// Periodically commit
						if (++num_tuples % 5000) 
						{
							SPI_execute("commit", true, 0);
							SPI_execute("begin", true, 0);
						}
					}
				}
			}
		}
		SPI_execute("commit", true, 0);
		//elog(INFO, "SPI: %d X-TUPLE(S) INSERTED.", num_tuples);

		pfree(cvals);
		pfree(ctypes);
		pfree(uvals);
		pfree(utypes);
		pfree(lvals);
		pfree(ltypes);
	}

	pfree(tablename);
	pfree(select_stmt1);
	pfree(select_stmt2);

	// Clean up
	SPI_finish();

	return 1;
}

int compare_values(HeapTuple tuple1, HeapTuple tuple2, int cols)
{
	bool isnull1, isnull2;
	char *val1, *val2;
	int i, c;
	for (i=0; i < cols; i++)
	{
		SPI_getbinval(tuple2, fetch_tuptable->tupdesc, i+1, &isnull1);
		SPI_getbinval(tuple2, fetch_tuptable_setops->tupdesc, i+1, &isnull2);
		if (isnull1)
			return -1;
		if (isnull2)
			return 1;
		val1 = SPI_getvalue(tuple1, fetch_tuptable->tupdesc, i+1);
		val2 = SPI_getvalue(tuple2, fetch_tuptable_setops->tupdesc, i+1);
		c = strcmp(val1, val2);
		pfree(val1);
		pfree(val2);
		if (c != 0)
			return c;
	}
	return 0;
}

int 
get_xid(HeapTuple tuple, int* max_xids)
{
    int i, xid = 0;
	char* val;
    for (i=0; i < num_joins; i++)
	{
		val = SPI_getvalue(tuple, fetch_tuptable->tupdesc, num_u_datacolumns+num_joins+i+1);
		xid += atoi(val) * pow(max_xids[i], i);
		pfree(val);
	}
    return xid;
}


node* 
child_tree(char* tname, int aid, int aids[MAXTUPS][MAXALTS], int xids[MAXTUPS], float confs[MAXTUPS][MAXALTS], char tables[MAXTUPS][TABLENAME], int* conjunctive, int stopatgreen){

	node* res;
	char linquery[8192];
	char metaquery[8192];
	int ret;
	int proc;
	char srctname[MAXDEG][TABLENAME];
	int srcaid[MAXDEG];
	int disj[MAXDEG];
	int j,i;
	int base=0;

	res=palloc(sizeof(node));
	snprintf(res->table,sizeof(res->table),"%s",tname);
	res->aid=aid;

	//add meta-data query to check whether base

	snprintf(metaquery,sizeof(metaquery),"select confbase, type from trio_tables where tablename='%s'",tname);
	//elog(INFO, "%s", metaquery);

	ret=SPI_execute(metaquery, true, 0);
	proc=SPI_processed;

	if (proc>1)
		elog(ERROR, "Table Metadata inconsistent!");
	else if (proc==0)
		base=0;
	else if(ret > 0 && SPI_tuptable != NULL)
	{

		TupleDesc tupdesc = SPI_tuptable->tupdesc;
		SPITupleTable *tuptable = SPI_tuptable;
		HeapTuple tuple = tuptable->vals[0];

		if(strcmp(SPI_getvalue(tuple, tupdesc, 1),"t")==0){
			base=1;
		}else{ 
			base=0;
		}
		//elog(INFO, "%s %i", tname, base);

		if(stopatgreen==1){
			if(atoi(SPI_getvalue(tuple, tupdesc , 2)) == 0)
				base=1;
		}

	}else{
		elog(ERROR, "Bad SPI result for metaquery");
	}

	if (base==1){

		int found=0;
		res->type=BASE;

		for (j=0;j<MAXTUPS;j++){
			for(i=0;i<MAXALTS;i++){
				if (aids[j][i]==NOVAL)
					break;
				if ((aids[j][i]==aid)&&(strcmp(tname,tables[j])==0)){
					res->tup=j;
					res->alt=i;
					found=1;
				}
			}
		}

		if (!found){

			snprintf(linquery,sizeof(linquery),"select trio_xid,trio_alts,trio_conf,trio_enc from %s where trio_aid=",tname); 
			snprintf(linquery+strlen(linquery),sizeof(linquery)-strlen(linquery),"%d",aid);
			//elog(INFO, "DATA-LIN %s\n", linquery);

			ret=SPI_execute(linquery, true, 0);
			proc=SPI_processed;

			if (proc!=1)
				elog(ERROR,"AID not unique or doesn't exist\n");

			if (ret > 0 && SPI_tuptable != NULL){
				float conf;
				int enc;
				int xid;
				int alts;
				TupleDesc tupdesc = SPI_tuptable->tupdesc;
				SPITupleTable *tuptable = SPI_tuptable;
				HeapTuple tuple = tuptable->vals[0];

				xid=atoi(SPI_getvalue(tuple, tupdesc, 1));
				alts=atoi(SPI_getvalue(tuple, tupdesc, 2));
				if(NULL==SPI_getvalue(tuple, tupdesc, 3))
					conf=GREENCONF;
				else 
					conf=atof(SPI_getvalue(tuple, tupdesc, 3));
				enc=atoi(SPI_getvalue(tuple, tupdesc, 4));

				if(conf==GREENCONF){
					if(alts>enc)
						conf=1.0/(enc+1);
					else
						conf=1.0/enc;
				}

				found=0;
				for (j=0;j<MAXTUPS;j++){
					if (xids[j]==NOVAL){
						res->tup=j;
						res->alt=0;
						found=1;
						break;
					}
					if ((xid==xids[j])&&(strcmp(tname,tables[j])==0)){
						for (i=0;i<MAXALTS;i++){
							if (aids[j][i]==NOVAL){
								res->tup=j;
								res->alt=i;
								found=1;
								break;
							}
						}
						break;
					}
				}
				if (!found) 
					elog(ERROR,"MAXTUPS or MAXALTS too small\n");

				xids[res->tup]=xid;
				aids[res->tup][res->alt]=aid;
				confs[res->tup][res->alt]=conf;
				//elog(INFO, " CONF %i %i = %f", res->tup, res->alt, conf);
				snprintf(tables[res->tup],sizeof(tables[res->tup]),"%s",tname);

			}else{
				elog(ERROR,"AID not unique or doesn't exist\n");
			}
		}


	}else{

		snprintf(linquery,sizeof(linquery), "select aid,disjlin,srclin,srctable from trio_lin_%s where aid=",tname);
		snprintf(linquery+strlen(linquery),sizeof(linquery)-strlen(linquery),"%d order by disjlin",aid);
		//elog(INFO, "LIN     %s\n", linquery);

		ret=SPI_execute(linquery, true, 0);
		proc=SPI_processed;

		if (proc>MAXDEG)
			elog(ERROR,"MAXDEG is too small\n");

		if (proc == 0)
			elog(ERROR," Table type not set properly in trio_tables. Derived table with no lineage found!");
		
		if (ret > 0 && SPI_tuptable != NULL){
		    TupleDesc tupdesc = SPI_tuptable->tupdesc;
		    SPITupleTable *tuptable = SPI_tuptable;
		    int lincase=0;
		    int spl=proc;
		    for (j = 0; j < proc; j++){
				HeapTuple tuple = tuptable->vals[j];
				snprintf(srctname[j],sizeof(srctname[j]),"%s",SPI_getvalue(tuple, tupdesc, 4));
				srcaid[j]=atoi(SPI_getvalue(tuple, tupdesc, 3));
				disj[j]=atoi(SPI_getvalue(tuple, tupdesc, 2));
								
				if (j>0 && (disj[j]-disj[j-1])>lincase){
				  lincase=disj[j]-disj[j-1];
				  if (lincase>1) spl=j;
				  *conjunctive=0;
				}
			}
			res->tup=NOVAL;
			res->alt=NOVAL;

			/////
			
			if (lincase>1){
			  
			  res->deg=2;
			  snprintf(res->table,sizeof(res->table),"%s",tname);
			  
			  node* res1;
			  node* res2;
			  res1=palloc(sizeof(node));
			  res2=palloc(sizeof(node));
			  res1->type=DIS;
			  res2->type=DIS;
			  res1->tup=NOVAL;
			  res2->tup=NOVAL;
			  res1->alt=NOVAL;
			  res2->alt=NOVAL;
			  node* dchild=NULL;

			  int lst;
			  
			  lst=-1;
			  res1->deg=0;
			  for (j = 0; j < spl; j++){
			    if(disj[j]!=lst){
			      dchild = palloc(sizeof(node));
			      res1->child[res1->deg]=dchild;
			      res1->deg++;
			      lst=disj[j];
			      dchild->aid=disj[j];
			      snprintf(res1->table,sizeof(res1->table),"%s",tname);
			      dchild->tup=NOVAL;
			      dchild->alt=NOVAL;
			      dchild->deg=0;
			      dchild->type=CON;
			    }
			    dchild->deg++;
			  }

			  lst=-1;
			  res2->deg=0;
			  for (j = spl; j < proc; j++){
			    if(disj[j]!=lst){
			      dchild = palloc(sizeof(node));
			      res2->child[res2->deg]=dchild;
			      res2->deg++;
			      lst=disj[j];
			      dchild->aid=disj[j];
			      snprintf(res2->table,sizeof(res2->table),"%s",tname);
			      dchild->tup=NOVAL;
			      dchild->alt=NOVAL;
			      dchild->deg=0;
			      dchild->type=CON;
			    }
			    dchild->deg++;
			  }

			  int l;
			  int k;
			  int ct=0;
			  
			  for(l=0;l<res1->deg;l++){
			    for(k=0;k<res1->child[l]->deg;k++){
			      res1->child[l]->child[k]=child_tree(srctname[ct],srcaid[ct],aids,xids,confs,tables,conjunctive,stopatgreen);
			      ct++;
			    }
			  } 

			  for(l=0;l<res2->deg;l++){
			    for(k=0;k<res2->child[l]->deg;k++){
			      res2->child[l]->child[k]=child_tree(srctname[ct],srcaid[ct],aids,xids,confs,tables,conjunctive,stopatgreen);
			      ct++;
			    }
			  } 
			  
			  res->child[0]=res1;
			  res->type=CON;
			  if(lincase==2){
			    res->child[1]=res2;
			  }else if(lincase==3){
			    node* negnode;
			    negnode=palloc(sizeof(node));
			    negnode->deg=1;
			    negnode->type=NEG;
			    negnode->tup=NOVAL;
			    negnode->alt=NOVAL;
			    negnode->child[0]=res2;
			    snprintf(negnode->table,sizeof(negnode->table),"%s",tname);
			    res->child[1]=negnode;
			  }

			}
			else if (lincase==1){
			  
			  int lst=-1;
			  res->type=DIS;
			  res->deg=0;
			  node* dchild=NULL;
			  for (j = 0; j < proc; j++){
			    if(disj[j]!=lst){
			      dchild = palloc(sizeof(node));
			      res->child[res->deg]=dchild;
			      res->deg++;
			      lst=disj[j];
			      dchild->aid=disj[j];
			      snprintf(res->table,sizeof(res->table),"%s",tname);
			      dchild->tup=NOVAL;
			      dchild->alt=NOVAL;
			      dchild->deg=0;
			      dchild->type=CON;
			    }
			    dchild->deg++;
			  }

				int l;
				int k;
				int ct=0;

				for(l=0;l<res->deg;l++){
				  for(k=0;k<res->child[l]->deg;k++){
				    res->child[l]->child[k]=child_tree(srctname[ct],srcaid[ct],aids,xids,confs,tables,conjunctive,stopatgreen);
				    ct++;
				  }
				} 

			}else if (lincase==0){

			  res->type=CON;
			  res->tup=NOVAL;
			  res->alt=NOVAL;
			  res->deg=proc;
			  for (j = 0; j < proc; j++){
			    res->child[j]=child_tree(srctname[j],srcaid[j],aids,xids,confs,tables,conjunctive,stopatgreen);
			  } 

			}
		  }else {
			elog(ERROR,"Unexpected SPI result");
		  }	
	}
	return res;
}

Datum trio_get_conf_spi(text* p_tables, text* p_aids, text* p_disj, int stopatgreen){
	float res;
	node* root;
	int aids[MAXTUPS][MAXALTS];
	int xids[MAXTUPS];
	int choice[MAXTUPS];
	float confs[MAXTUPS][MAXALTS];
	char tables[MAXTUPS][TABLENAME];
	int j,i;
	int* conjunctive=palloc(sizeof(int));

	int proc;
	char argtname[MAXDEG][TABLENAME];
	int argaid[MAXDEG];
	int disj[MAXDEG];
	int topcount;
    
	*conjunctive=1;

	char* t_tables = DatumGetCString(DirectFunctionCall1(textout,
		PointerGetDatum(p_tables)));
	char* t_aids = DatumGetCString(DirectFunctionCall1(textout,
		PointerGetDatum(p_aids)));
	char* t_disj = DatumGetCString(DirectFunctionCall1(textout,
		PointerGetDatum(p_disj)));

	char* tok = strtok(t_tables, "|");
	j=0;
	while (tok != NULL)
	{
		strcpy(argtname[j], tok);
		tok = strtok(NULL, "|");
		j++;
	}
	proc = j;
	tok = strtok(t_aids, "|");
	j=0;
	while (tok != NULL)
	{
		argaid[j]=atoi(tok);
		tok = strtok(NULL, "|");
		j++;
	}
	tok = strtok(t_disj, "|");
	j=0;
	topcount = 1;
	while (tok != NULL)
	{
		disj[j]=atoi(tok);
		if ((j>0) && (disj[j]!=disj[j-1]))
			topcount++;
		tok = strtok(NULL, "|");
		j++;
	}
	root=palloc(sizeof(node));
	for (j=0;j<MAXTUPS;j++){
		xids[j]=NOVAL;
		choice[j]=NOVAL;
		for(i=0;i<MAXALTS;i++){
			aids[j][i]=NOVAL;
			confs[j][i]=NOVAL;
		}
	}

	SPI_connect();
	if(topcount!=1){
		root->type=DIS;
		*conjunctive=0;
		int lst=-1;
		root->deg=0;
		node* dchild=NULL;
		for (j = 0; j < proc; j++){
			if(disj[j]!=lst){
				dchild = palloc(sizeof(node));
				root->child[root->deg]=dchild;
				root->deg++;
				lst=disj[j];
				dchild->aid=disj[j];
				snprintf(root->table,sizeof(root->table),"onthefly");
				dchild->tup=NOVAL;
				dchild->alt=NOVAL;
				dchild->deg=0;
				dchild->type=CON;
			}
			dchild->deg++;
		}

		int l;
		int k;
		int ct=0;
		for(l=0;l<root->deg;l++){
			for(k=0;k<root->child[l]->deg;k++){
				root->child[l]->child[k]=child_tree(argtname[ct],argaid[ct],aids,xids,confs,tables,conjunctive,stopatgreen);
				ct++;
			}
		}

	}else{

		root->type=CON;
		root->tup=NOVAL;
		root->alt=NOVAL;
		root->deg=proc;
		for (j = 0; j < proc; j++){
			root->child[j]=child_tree(argtname[j],argaid[j],aids,xids,confs,tables,conjunctive,stopatgreen);
		}
	}

	if(*conjunctive==0){

		for (j=0;j<MAXTUPS;j++){
			float confsum=0;
			int last=NOVAL;
			for (i=0;i<MAXALTS;i++){
				if (confs[j][i]==NOVAL){
					last=i;
					break;
				} else{
					confsum+=confs[j][i];
				}
			}
			if ((confsum<1)&&(last>0))
			{
				confs[j][last]=1-confsum;
			}
		}

		int maxLevel = 0;
		//if (j<12)
			res=reciter(0,confs,choice,root,&maxLevel);  
		//else {
			//elog(WARNING, "WARNING: maximum recursion level %d for conf() exceeded! Using Monte Carlo simulations instead.", maxLevel);
		//	res = montecarlo(0,confs,choice,root,&maxLevel);
		//}

	}else{

		res=1;
		for (i=0;i<MAXTUPS;i++){
			if (confs[i][0]==NOVAL) {break;}
			if (confs[i][1]!=NOVAL) {res=0; break;}
			res=res*confs[i][0];
		} 

	}

	// Clean up
	SPI_finish();

	PG_RETURN_FLOAT4(res);
}

float montecarlo(int level, float confs[MAXTUPS][MAXALTS], int choice[MAXTUPS], node* root, int* maxLevel){

	int i=0,j,alts,N=10000;
	float m=0;

	for(i=0; i<N; i++){
		for(j=0; j<MAXTUPS && confs[j][0]>NOVAL; j++){
			alts = 0;
			while(confs[j][alts]>NOVAL && alts<MAXALTS)
				alts++;
			choice[j]=rand()%alts;
		}
		if (eval(root,choice))
			m++;
	}

	return m/N;
}

float reciter(int level, float confs[MAXTUPS][MAXALTS], int choice[MAXTUPS], node* root, int* maxLevel){

	int i=0,j;
	float ret=0.0;

	while(confs[level][i]>NOVAL && i<MAXALTS){
		choice[level]=i;
		if (level>*maxLevel)
			(*maxLevel) = level;

		if(level+1<MAXTUPS && confs[level+1][0]>NOVAL && level<12){
			ret+=reciter(level+1,confs,choice,root,maxLevel);

		}else{

			/*
			elog(INFO,"\n");
			for (j=0;j<MAXTUPS;j++){
				if (choice[j]==NOVAL) break;
				elog(NOTICE,"%d %d\n",j,choice[j]);
			}
			*/

			if (eval(root,choice)){
				float res=1;
				for (j=0;j<MAXTUPS;j++){
					if (choice[j]!=NOVAL)
						res=res*confs[j][choice[j]];
				}
				//elog(INFO, "\n==> %f\n", res);
				ret+=res;
			} 
		}
		i++;
	}
	return ret;
}

int eval(node* tree, int choice[MAXTUPS]){
	int i;

	if (tree->type==BASE)
		return choice[tree->tup]==tree->alt;

	if (tree->type==DIS){
		for(i=0;i<tree->deg;i++){
			if (eval(tree->child[i],choice))
				return 1;
		}
		return 0;
	}

	if (tree->type==CON){
		for(i=0;i<tree->deg;i++){
			if (eval(tree->child[i],choice)==0)
				return 0;
		}
		return 1;
	}

	if(tree->type==NEG){
	  if (eval(tree->child[0],choice)==0)
	    return 1;
	  else return 0;
	}
	
	elog(ERROR,"Bad Type");
	return 0;
}
