#include "executor/spi.h"

#define PREFETCH_BUFFER 16384
#define MAX_ALTERNATIVES 1024
#define MAX_COLUMNNAME 1024
#define MAX_COLUMNLENGTH 1024
#define MAX_COLUMNS 128
#define MAX_JOINS 24
#define MAXDEG 512
#define TABLENAME 256
#define GREENCONF -1
#define MAXTUPS 256	
#define MAXALTS 256
#define NOVAL -1

enum linkType {BASE, CON, DIS, NEG};

struct noderec{
  int deg;
  enum linkType type;
  int aid;
  char table[TABLENAME];
  struct noderec* child[MAXDEG];
  int tup;
  int alt;
};

typedef struct noderec node;

typedef struct value
{ 
	int column_idx;
	int length;
	
	bool isnull;
	char* value;

	// Postgres native datatype
	Datum value_d;

} VALUE;

typedef struct lineage
{
	int src_aid;
	char* src_table;

	// the source alternative from the same table 
	// that this lineage is disjunctive to
	int disj_aid;

} LINEAGE;

typedef struct alternative
{ 
	int aid;
	int encoding;

	int num_values;
	int num_lineage;

	float confidence; 

	VALUE* values; // the actual data
	LINEAGE* lineage; // lineage information
	
} ALTERNATIVE;

typedef struct xtuple
{
	int xid;
	int num_alts;

	ALTERNATIVE* alternatives;

} XTUPLE;

// Reserve some static buffer for next fetch.
// Much faster than dynamic allocation.
ALTERNATIVE* alternatives;
ALTERNATIVE* merged_alternatives;
ALTERNATIVE* buffered_alternative;

bool buffered_questionMark, questionMark = true;
int buffered_xid;

void* fetch_plan;
int fetch_processed;

SPITupleTable** fetch_stack;
int fetch_stack_size;

SPITupleTable* fetch_tuptable;
Portal fetch_portal;

void* fetch_plan_setops;
int fetch_processed_setops;

SPITupleTable* fetch_tuptable_setops;
Portal fetch_portal_setops;

char** tablenames;

char* current_xid;
char* previous_xid;

char* column_c_list;
char* column_u_list;

int* aid_idx;
int* xid_idx;
int* data_idx;
int* is_unique;

int* c_datacolumns;
int* u_datacolumns;

int conf_idx;
int enc_idx;
int alts_idx;

int num_c_datacolumns;
int num_u_datacolumns;
int num_joins;
int num_tuples;

int trio_merged;
int trio_flatten;
int trio_groupAlts;
int trio_nolineage;
int trio_nomaybe;
int trio_noconf;
int trio_distinct;

int current_alt;
int current_tuple;

int init_fetch(char*, char*, char*, char*, char*, int);

int fetchone(XTUPLE*, int*, int*);

int trio_selectinto(text*, text*, text*, text*, text*, int, int, int, int, int, int, int, int);

int trio_selectinto_distinct(text*, text*, text*, text*, int);

int trio_mergeinto_distinct(text*, text*, text*, text*, text*, int, int, int, int, int, int);

int get_xid(HeapTuple, int*);

int compare_values(HeapTuple, HeapTuple, int);

int horizontal_merging(XTUPLE*);

Datum trio_get_conf_spi(text* tables, text* aids, text* disj, int stopatgreeen);

node* child_tree(char* tname, int aid, int aids[MAXTUPS][MAXALTS], int xids[MAXTUPS], float confs[MAXTUPS][MAXALTS], char tables[MAXTUPS][TABLENAME], int* conjunctive, int stopatgreen);

float reciter(int level, float confs[MAXTUPS][MAXALTS], int choice[MAXTUPS],node* root, int*);

float montecarlo(int level, float confs[MAXTUPS][MAXALTS], int choice[MAXTUPS],node* root, int*);

int eval(node* tree, int choice[MAXTUPS]);
