#!/usr/bin/env python

import optparse
import sys
import os
import triodb
from test.trio_test import query_file
from test.trio_test import initdb
from test.trio_test import loaddata
from test.trio_test import createtable
import ConfigParser
import common

test_conf = ConfigParser.ConfigParser()
test_conf.read(common.ROOT + "/test/test.cfg")

def main(argv=None):
    p = optparse.OptionParser()
    p.add_option("-u", "--user", dest="user", help="The database user account")
    p.add_option("-d", "--database", default=test_conf.get("global", "dbname"), 
        dest="database", help="The name of the database")
    p.add_option("-p", "--password", dest="password", action="store_true", help="Prompt password")
    (options, argv) = p.parse_args()
    options = options.__dict__
    if options["password"]:
        try:
            options["password"] = getpass('Enter password: ')
        except EOFError:
            return 0
    if len(argv) != 1:
        print >> sys.stderr, "Usage: generate_output.py filename"
        return 1
    if not argv[0].endswith(".in"):
        print >> sys.stderr, "Bad filename: %s: Filename must ends with '.in'" % argv[0]
        return 1
    outfile = argv[0][:-3] + ".out"
    if os.path.exists(outfile):
        print("File %s already exists." % outfile)
        answer = raw_input("Overwrite? (y/N): ")
        if answer != "y":
            return 0
    initdb()
    cnx = triodb.connect(**options)
    loaddata(cnx)
    createtable(cnx)
    query_file(cnx, argv[0], open(outfile, "w"))
    print "Output written to %s" % outfile

if __name__ == "__main__":
    sys.exit(main())
