#!/usr/bin/env python

__id__ = "$Id: trioplus.py 336 2007-04-17 17:39:54Z theobald $"

from getpass import getpass
import optparse
import signal
import sys
import triodb
import common
import ConfigParser
import os
import popen2
import traceback

test_conf = ConfigParser.ConfigParser()
test_conf.read(common.ROOT + "/test/test.cfg")

def print_action(msg):
    print "%s" % (msg + "...").ljust(60),

def print_success():
    print "SUCCESS!"

def print_failure():
    print "FAILURE!"

def print_exception(msg=''):
    e = sys.exc_info()
    print
    print str(e[0]) + ": " + str(e[1])
    traceback.print_tb(e[2])
    #if msg != '':
    #    raise Exception(msg)

def initdb():
    print_action("Initializing database %s" % test_conf.get("global", "dbname"))
    os.system("dropdb %s" % test_conf.get("global", "dbname"))
    if os.system("createdb %s" % test_conf.get("global", "dbname")) != 0:
        raise Exception("Couldn't create database %s" % test_conf.get("global", "dbname"))
    if os.system("%s/setup/setup.py %s" % (common.ROOT, test_conf.get("global", "dbname"))) != 0:
        raise Exception("Couldn't create database %s" % test_conf.get("global", "dbname"))
    print_success()

def xtuples2str(xtuples):
    s = [str(x) for x in xtuples]
    s.sort()
    val = ""
    for xtuple in s:
        val += xtuple
    return val

def query_file(cnx, file, outfile=None):
    c = cnx.cursor()
    commands = ""
    s = ""
    f = open(file)
    for line in f:
        if not line.strip().startswith("--"):
            commands += line
        parts = commands.split(";")
        commands = parts[-1:][0]
        comm_list = parts[:-1]
        for comm in comm_list:
            try:
                c.execute(comm)
                s += xtuples2str(c.xfetchall())
                s += "\n"
                cnx.commit()
            except Exception, ex:
                print_exception(ex)
                print_failure()
                print("\n################################################")
                print comm
                print("################################################\n")
                raise Exception(comm)
    if outfile:
        outfile.write(s)
    return s

def execute_file(cnx, file):
    c = cnx.cursor()
    commands = ""
    f = open(file)
    for line in f:
        if not line.strip().startswith("--"):
            commands += line
        parts = commands.split(";")
        commands = parts[-1:][0]
        comm_list = parts[:-1]
        for comm in comm_list:
            print comm
            try:
                c.execute(comm)
                cnx.commit()
            except Exception, ex:
                print_exception(ex)
                print_failure()
                print("\n################################################")
                print comm
                print("################################################\n")
                raise Exception(comm)

def test_file(cnx, file):
    c = cnx.cursor()
    commands = ""
    command_list = []
    s = ""
    f = open(file)
    f2 = open(file[:-3] + ".out")
    output = f2.read().strip().split("\n\n")
    for line in f:
        commands += line
        parts = commands.split(";")
        commands = parts[-1:][0]
        command_list += parts[:-1]
    cmd_out = map(None, command_list, output)
    for comm in cmd_out:
        c.execute(comm[0])
        s = xtuples2str(c.xfetchall())
        if s.strip() != comm[1].strip():
            print_failure()
            print "\nQUERY"
            print("################################################")
            print comm[0].strip()
            print "\nMY OUTPUT"
            print("################################################")
            print s.strip()
            print "\nOUTPUT FROM %s" % file[:-3] + ".out"
            print("################################################")
            print comm[1].strip()
            raise Exception()
        cnx.commit()

def test1(cnx):
    print_action("Running test 1 on %s" % test_conf.get("global", "dbname"))

def loaddata(cnx):
    print_action("Loading data to database %s" % test_conf.get("global", "dbname"))
    execute_file(cnx, common.ROOT + "/test/" + test_conf.get("global", "loaddata"))
    print_success()

def removedata(cnx):
    print_action("Removing data from database %s" % test_conf.get("global", "dbname"))
    execute_file(cnx, common.ROOT + "/test/" + test_conf.get("global", "removedata"))
    print_success()

def createtable(cnx):
    print_action("Creating tables in database %s" % test_conf.get("global", "dbname"))
    execute_file(cnx, common.ROOT + "/test/" + test_conf.get("global", "createtable"))
    print_success()

def runtest(cnx, file):
    print_action("Running %s" % file)
    test_file(cnx, file)
    print_success()

def runtests(cnx):
    infiles = [x for x in os.listdir(common.ROOT + "/test/queries") if x.endswith("in")]
    for file in infiles:
        runtest(cnx, common.ROOT + "/test/queries/" + file)
 
def main(argv=None):
    if argv is None:
        argv = sys.argv
    p = optparse.OptionParser()
    p.add_option("-u", "--user", dest="user", help="The database user account")
    p.add_option("-d", "--database", default=test_conf.get("global", "dbname"), 
        dest="database", help="The name of the database")
    p.add_option("-p", "--password", dest="password", action="store_true", help="Prompt password")
    (options, args) = p.parse_args()
    options = options.__dict__
    if options["password"]:
        try:
            options["password"] = getpass('Enter password: ')
        except EOFError:
            return 0
    try:
        initdb()
        cnx = triodb.connect(**options)
        loaddata(cnx)
        createtable(cnx)
        #runtests(cnx)
        #removedata(cnx)
        cnx.close()
        print("\n################################################\n")
        print "You passed the test. You are now officially awesome."
        print("\n################################################\n")
    except Exception, ex:
        print ex
        return 1

if __name__ == "__main__":
    signal.signal(signal.SIGINT, signal.SIG_IGN) # ignores ctl-c
    sys.exit(main())
