__id__ = "$Id: triodb.py 578 2007-11-21 01:08:06Z theobald $"

import pgdb
import common
import xtuple

from node import *
from visitor import dumper
from visitor import printer
from executor import executor
from visitor import restructurer
from trioparser import parser

def connect(**options):
    pdb = pgdb.connect(**options)
    try:
		pdb.cursor().execute("""DELETE FROM trio_tables_lin WHERE tablename IN
							   (SELECT distinct tablename FROM trio_tables WHERE istemporary=TRUE)""")
		pdb.cursor().execute("""DELETE FROM trio_attribs WHERE tablename IN
							   (SELECT distinct tablename FROM trio_tables WHERE istemporary=TRUE)""")
		pdb.cursor().execute("""DELETE FROM trio_tables WHERE istemporary=TRUE""")
    except Exception, e:
        print "\nWARNING: Trio metadata inconsistent or not initialized!"
        pdb.rollback()
    return TrioCnx(pdb)

class TrioCnx:
    def __init__(self, pdb):
        self.pdb = pdb

    def ncursor(self):
        return self.pdb.cursor()

    def cursor(self):
        return TrioCursor(self)

    def commit(self):
        self.pdb.commit()
      
    def rollback(self):
        self.pdb.rollback()

    def close(self):
        self.pdb.close()
   
class TrioCursor:
    def __init__(self, tdb):
        self.tdb = tdb
        self.num_joins = 0
        self.num_datavalues = 0
        self.root_node = None
        self.cursor = None
        self.sqlCommand = None
        self.setoperator = False
        
    def execute(self, sqlCommand):
        self.isXCursor = False
        self.isFetchable = False
        self.lastAlt = None
        self.lastXid = None
        
        self.isTemorary = False
        
        self.gotUniform = False
        self.isConfBase = False
        self.gotAsConf = False
        self.merged = False
        self.flatten = False
        self.noconf = False
        self.nomaybe = False
        self.nolineage = False
        self.groupAlts = -1
        self.distinct = False
        
        self.sqlCommand = sqlCommand
        s = parser.parse(sqlCommand)
        
        # SQL parse tree
        if int(common.cp.get("global", "debug_level")) > 0:
            print self.sqlCommand
            print
            dumper.Dumper().visit(s)
            print
        
        r = restructurer.Restructurer()
        t = r.visit(s)
        
        # Rewritten TriQL parse tree
        if int(common.cp.get("global", "debug_level")) > 1:
            print
            dumper.Dumper().visit(t)
            print
            
        e = executor.Executor(self)
        e.execute(s)
        
    def __getattr__(self, attrname):
        if self.cursor is None:
            return None
        if attrname == "rowcount":
            return self.cursor.rowcount
        elif attrname == "description":
            return self.cursor.description
        elif attrname == "tabletype":
            return self.cursor.tabletype
      
    def fetchall(self):
        return self.cursor.fetchall()
         
    def fetchone(self):
        if self.isXCursor:
            xtup = self.xfetchone()
            if xtup is not None:
                if xtup.enc > len(xtup) + 1:
                    xtup.enc = len(xtup) + 1
            return xtup
        else:
            return self.cursor.fetchone()

    def xfetchall(self):
        tuples = []
        while True:
            tuple = self.xfetchone()
            if tuple is None:
                break
            tuples.append(tuple)
        return tuples

    def xfetchone(self):
        questionMark = True
        xtup = xtuple.XTuple()
        if self.lastAlt is not None:
            xtup.xid = self.lastXid
            if self.nolineage:
                a = xtuple.Alternative(self.lastAlt[:self.num_datavalues])
            else:
                a = xtuple.Alternative(self.lastAlt[:self.num_datavalues], 
                                   self.getLin(self.lastAlt[self.num_datavalues : 
                                          self.num_datavalues+self.num_joins]))
            if not self.noconf and not self.gotUniform and self.lastAlt[self.num_datavalues+(2*self.num_joins)+1] is not None:
                a.confidence = float(self.lastAlt[self.num_datavalues+(2*self.num_joins)+1])
            xtup.alternatives.append(a)
            xtup.enc = int(self.lastAlt[self.num_datavalues+(2*self.num_joins)])
            alts = int(self.lastAlt[self.num_datavalues+(2*self.num_joins)+2])
            questionMark = (alts > 1 or xtup.enc > alts)
            self.lastAlt = None
            
        while True:
            self.lastAlt = self.cursor.fetchone()
            if self.lastAlt is None:
                
                if ((self.distinct or self.groupAlts > -1) and questionMark) or xtup.enc > len(xtup.alternatives):
                    xtup.enc = len(xtup.alternatives) + 1
                else:
                    xtup.enc = len(xtup.alternatives)
                            
                if self.gotUniform:
                    for alt in xtup.alternatives:
                        if xtup.enc > len(xtup.alternatives):
                            alt.confidence = 1.0 / (len(xtup.alternatives)+1)
                        else:
                            alt.confidence = 1.0 / len(xtup.alternatives)
                            
                if self.merged:
                    self.horizontal_merging(xtup)
                    
                if self.nomaybe:
                    xtup.enc = len(xtup.alternatives)
                if self.nomaybe:
                    for alt in xtup.alternatives:
                        alt.confidence = None
                
                if len(xtup) == 0:
                    return None
                else:
                    return xtup
                    
            if self.groupAlts == -1:
                xid = self.getXid(self.lastAlt[self.num_datavalues+self.num_joins : 
                                       self.num_datavalues+(2*self.num_joins)])
            else:
                xid = self.lastAlt[self.groupAlts]
            
            if self.lastXid is not None and (self.flatten or xid != self.lastXid):
                self.lastXid = xid
                break
            
            self.lastXid = xid
            xtup.xid = self.lastXid
            if self.nolineage:
                a = xtuple.Alternative(self.lastAlt[:self.num_datavalues])
            else:
                a = xtuple.Alternative(self.lastAlt[:self.num_datavalues], 
                                   self.getLin(self.lastAlt[self.num_datavalues : 
                                          self.num_datavalues+self.num_joins]))
            if not self.noconf and not self.gotUniform and self.lastAlt[self.num_datavalues+(2*self.num_joins)+1] is not None:
                a.confidence = float(self.lastAlt[self.num_datavalues+(2*self.num_joins)+1])
            xtup.alternatives.append(a)
            xtup.enc = int(self.lastAlt[self.num_datavalues+(2*self.num_joins)])
            alts = int(self.lastAlt[self.num_datavalues+(2*self.num_joins)+2])
            questionMark = questionMark and (alts > 1 or xtup.enc > alts)
        
        if ((self.distinct or self.groupAlts > -1) and questionMark) or xtup.enc > len(xtup.alternatives):
            xtup.enc = len(xtup.alternatives) + 1
        else:
            xtup.enc = len(xtup.alternatives)
                    
        if self.gotUniform:
            for alt in xtup.alternatives:
                if xtup.enc > len(xtup.alternatives):
                    alt.confidence = 1.0 / (len(xtup.alternatives)+1)
                else:
                    alt.confidence = 1.0 / len(xtup.alternatives)
                    
        if self.merged:
            self.horizontal_merging(xtup)
        
        if self.nomaybe:
            xtup.enc = len(xtup.alternatives)
        if self.nomaybe:
            for alt in xtup.alternatives:
                alt.confidence = None
                
        return xtup
        
    def getLin(self, trio_aids):
        lin = xtuple.Lineage()
        if self.distinct:
            lin.append(self.distinctTable, int(trio_aids[0]))
        else:
            for i in range(len(trio_aids)):
                if isinstance(self.root_node.children[2].children[i], As):
                    lin.append(self.root_node.children[2].children[i].children[0].value, int(trio_aids[i]));
                else:
                    lin.append(self.root_node.children[2].children[i].value, int(trio_aids[i]));
        return lin

    def getXid(self, trio_xids):
        return "_".join("%s" % xid for xid in trio_xids)

    def horizontal_merging(self, xtuple):
        disj = 0
        isUnique = []
        merged_alternatives = []
        for i in range(len(xtuple.alternatives)):
            isUnique = isUnique + [0];
        for i in range(len(xtuple.alternatives)):
            if isUnique[i] == 0:
                j = i + 1
                while j < len(xtuple):
                    if common.compare_values(xtuple.alternatives[i].data, xtuple.alternatives[j].data, self.num_datavalues) == 0:
                        # Track disjunctive lineage
                        disj += 1
                        for lin in xtuple.alternatives[j].getLineage().items():
                            print lin
                            xtuple.alternatives[i].lineage.append(lin[0], lin[1], disj)
                        if xtuple.alternatives[i].getConfidence() is not None and \
                           xtuple.alternatives[j].getConfidence() is not None:
                            xtuple.alternatives[i].setConfidence(xtuple.alternatives[i].getConfidence() + xtuple.alternatives[j].getConfidence())    
                        isUnique[i] += 1
                        isUnique[j] += 1
                    j += 1
                merged_alternatives.append(xtuple.alternatives[i])
                
        # Normalize encoding and update xtuple alternatives
        if xtuple.enc > len(xtuple.alternatives):
            xtuple.enc = len(merged_alternatives) + 1
        else:
            xtuple.enc = len(merged_alternatives)
        xtuple.alternatives = merged_alternatives

