#!/usr/bin/env python

__id__ = "$Id: lexer.py 80 2006-12-03 02:39:11Z michi2 $"

import string
from ply import *

keywords = (
    'all',
    'and',
    'any',
    'as',
    'avg',
    'by',
    'cascade',
    'compute',
    'conf',
    'confidences',
    'count',
    'create',
    'distinct',
    'drop',
    'eavg',
    'ecount',
    'emax',
    'emin',
    'esum',
    'except',
    'exists',
    'explain',
    'flatten',
    'float',
    'float4',
    'float8',
    'from',
    'group',
    'groupalts',
    'having',
    'in',
    'index',
    'insert',
    'int',
    'int4',
    'int8',
    'intersect',
    'into',
    'is',
    'like',
    'lineage',
    #'lavg',
    #'lcount',
    #'lmax',
    #'lmin',
    #'lsum',
    'max',
    'maybe',
    'min',
    'noconf',
    'nolineage',
    'nomaybe',
    'not',
    'null',
    'on',
    'or',
    'order',
    'scaled',
    'scaledbyexp',
    'select',
    'sum',
    'table',
    'temporary',
    'trio',
    'trio_aid',
    'trio_xid',
    'uncertain',
    'uniform',
    'union',
    'merged',
    'values',
    'varchar',
    'view',
    'where',
    'with',
)

tokens = keywords + (
    'id',
    'lineage_arrow',
    'comparison',
    'string_literal',
    'numeric_literal',
    'command',
) 
t_command = r'\\\w'
literals= r'\+-*/.,;()?|[]:'
t_comparison = r'=|!=|<=|>=|<>|<|>|[Ii]?[Ll][Ii][Kk][Ee]'

def t_comment(t):
    r'--.*'
    pass

def t_string_literal(t):
    r"'(\\'|[^'])*'"
    t.value = t.value[1:-1]
    return t

def t_numeric_literal(t):
    r'\d*\.?\d+([eE][-+]?\d+)?'
    t.value = eval(t.value)
    return t

def t_id(t):
    r'\w+'
    t.value = t.value.lower()
    if not t.value.startswith("trio_aid") and not t.value.startswith("trio_xid"):
        if t.value.startswith("trio_"):
            raise Exception("'trio_' prefix is reserved by Trio system.")
    if t.value in keywords:
        if t.value == 'null':
            t.type = 'null'
            t.value = None
        else:
            t.type = t.value.lower()
    return t

t_ignore = ' \t'

def t_lineage_arrow(t):
    '==>'
    return t
    
def t_newline(t):
    r'\n'
    t.lexer.lineno += 1

def t_error(t):
    print "Illegal character", t.value[0]
    t.lexer.skip(1)

lexer=lex.lex(debug=0)
