#!/usr/bin/env python

__id__ = "$Id: trioplus.py 581 2007-12-05 19:22:52Z theobald $"

from getpass import getpass
import optparse
import signal
import sys
import triodb
import common

trio_options = {"aids" : ["on", "off"]}

def lin_command(comm, trio_cursor):
    try:
        c = trio_cursor.tdb.pdb.cursor()
        comm = comm.split()
        if len(comm) < 3 or len(comm) > 4 or comm[1] != "lineage":
            raise Exception("Invalid explain command. Use: explain lineage <tablename> [aid]")
        if len(comm) == 3:
            query = """view table %s explain lineage""" % (comm[2])
        else:
            query = """view table %s where trio_aid = %s 
                       explain lineage""" % (comm[2], comm[3])
        trio_cursor.execute(query)
        if trio_cursor.rowcount == 0:
            raise Exception("Invalid aid for table '%s'" % comm[2])
        print_result(trio_cursor)
    except Exception, e:
        print e

def conf_command(comm, trio_cursor):
    try:
        comm = comm.split()
        if len(comm) < 3 or len(comm) > 4 or comm[1] != "confidences":
            raise Exception("Invalid compute confidences command. Use: compute confidences <tablename> [aid]")
        if len(comm) == 3:
            query = """view table %s compute confidences""" % (comm[2])
        else:
            query = """view table %s where trio_aid = %s 
                       compute confidences""" % (comm[2], comm[3])
        trio_cursor.execute(query)
        if trio_cursor.rowcount == 0:
            raise Exception("Invalid aid for table '%s'" % comm[2])
        print_result(trio_cursor)
    except Exception, e:
        print e

def set_command(comm):
    try:
        (attr, value) = comm.split()[1:]
        if not (attr in trio_options and value in trio_options[attr]):
            raise Exception()
        common.cp.set("global", attr, value)
        print """"%s" set to "%s" """ % (attr, value)
    except:
        print("Invalid set command.")

def main_loop(cnx):
    c = cnx.cursor()
    commands = ""
    while True:
        try:
            line = raw_input('> ')
        except EOFError:
            break
        commands += "\n" + line
        parts = commands.split(";")
        commands = parts[-1:][0]
        comm_list = parts[:-1]
        for comm in comm_list:
            try:
                if comm.strip().lower() == "commit":
                    cnx.commit()
                elif comm.strip().lower() == "rollback":
                    cnx.rollback()
                elif comm.strip().lower().startswith("set"):
                    set_command(comm.strip().lower())
                elif comm.strip().lower().startswith("compute"):
                    print "COMP"
                    conf_command(comm.strip().lower(), c)
                elif comm.strip().lower().startswith("explain"):
                    lin_command(comm.strip().lower(), c)
                else:
                    c.execute(comm)
                    if c.isFetchable:
                        print_result(c)
            except Exception, e:
                common.print_exception()
                cnx.rollback();
                continue

def print_result(cursor):
    attrnames = []
    colwidth = 15   
    if cursor.isXCursor and common.cp.get("global", "aids") == "on" and cursor.isViewTable:
        cursor.description = [["aid"]] + cursor.description
    for d in cursor.description:
       attrnames.append( d[ 0 ].upper() )
    numdashes = ( colwidth + 2 ) * len( cursor.description )
    sys.stdout.write( "\n+" )
    for i in range( numdashes ):
        sys.stdout.write( "-" )
    print "+"
   
    sys.stdout.write( "| " )
    formatstring = "%%%ds |" % colwidth

    for a in attrnames:
       sys.stdout.write( formatstring % a )
    print
    sys.stdout.write( "+%s+\n" % ( "=" * numdashes ) )
    if cursor.isXCursor:
        print_xtuples(cursor)
    else:
        print_tuples(cursor)

def print_tuples(cursor):
    colwidth = 15   
    formatstring = "%%%ds |" % colwidth
    numdashes = ( colwidth + 2 ) * len( cursor.description )
    while True:
        tuple = cursor.fetchone()
        if tuple is None:
            break
        for i in range(len(tuple)):
            sys.stdout.write( "| " )
            sys.stdout.write(formatstring % str(tuple[i]))
            print
            sys.stdout.write( "+%s+\n" % ( "-" * numdashes ) )

def print_xtuples(cursor):
    colwidth = 15   
    formatstring = "%%%ds |" % colwidth
    numdashes = ( colwidth + 2 ) * len( cursor.description )
    while True:
        xtuple = cursor.fetchone()
        if xtuple is None:
            break
        for i in range(len(xtuple)):
            if common.cp.get("global", "aids") == "on" and cursor.isViewTable:
                xtuple[i].data = [xtuple[i].lineage.srcaid[0]] + xtuple[i].data
            sys.stdout.write( "| " )
            for a in xtuple[ i ]:
                if a is not None:
                    sys.stdout.write( formatstring % str( a ) )
                else:
                    sys.stdout.write( formatstring % "NULL" )
            conf = xtuple[i].getConfidence()
            if conf is not None and conf >= 0:
                sys.stdout.write(str(conf))
            elif xtuple.getQuestionMark():
                sys.stdout.write("?")
            if cursor.hasExplainLineage:
                sys.stdout.write(" " + print_lineage(cursor, xtuple[i].traceLineage(cursor.tdb)))
            print
        sys.stdout.write( "+%s+\n" % ( "-" * numdashes ) )

def print_lineage(cursor, lineage):
        tnc = cursor.tdb.ncursor()
        s = ""
        for i in range(len(lineage)):
            query = """select srctable, srclin, disjlin from trio_lin_%s where aid = %s order by disjlin asc""" % (lineage[i][0], lineage[i][1])
            tnc.execute(query)
            nested_lineage = tnc.fetchall()
            if i == 0 and lineage[-1][2] != lineage[0][2]:
                    s+= "["
            if len(nested_lineage) > 0:
                s += "[" + print_lineage(cursor, nested_lineage) + "]"
            else:
                s += lineage[i][0] + ":" + str(lineage[i][1])
            if i == len(lineage) -1 and lineage[-1][2] != lineage[0][2]:
                s += "]"
            if len(lineage) > 1 and i < len(lineage) - 1 and lineage[i+1][2] - lineage[i][2] == 0:
               s += " AND "
            elif len(lineage) > 1 and i < len(lineage) - 1 and lineage[i+1][2] - lineage[i][2] == 1:
               s += "] OR ["
            elif len(lineage) > 1 and i < len(lineage) - 1 and lineage[i+1][2] - lineage[i][2] == 2:
               s += "] AND NOT ["
            elif i < len(lineage) - 1:
               s += " AND "
        return s
                
def main(argv=None):
    if argv is None:
        argv = sys.argv
    p = optparse.OptionParser()
    p.add_option("-s", "--server", 
                 dest="dsn", help="The database server")
    p.add_option("-u", "--user", 
                 dest="user", help="The database user account")
    p.add_option("-p", "--password", 
                 dest="password", action="store_true", help="Prompt password")
    p.add_option("-P", "--Password", 
                 dest="password", help="Get password")
    p.add_option("-o", "--host", dest="host", 
                 help="The hostname of the client")
    p.add_option("-d", "--database", 
                 dest="database", help="The name of the database")
    (options, args) = p.parse_args()
    options = options.__dict__
    
    if options["password"] == True:
        try:
            options["password"] = getpass('Enter password: ')
        except EOFError:
            return 0
            
    try:
        cnx = triodb.connect(**options)
        print
        print "Welcome to the TrioPlus command-line terminal. Press <CTRL><Z><ENTER> to exit."
        print
        main_loop(cnx)
        cnx.commit()
        cnx.close()
    except:
        common.print_exception()
        return 1

if __name__ == "__main__":
    signal.signal(signal.SIGINT, signal.SIG_IGN) # ignores ctl-c
    sys.exit(main())
