__id__ = "$Id: printer.py 560 2007-08-30 22:56:57Z theobald $"

from node import *
from visitor import Visitor

class Printer(Visitor):
    def __init__(self, trio_cursor=None):
        self.cursor = trio_cursor
        
    def visit(self, node):
        nodetype = node.__class__.__name__
        visitfn = getattr(self, 'visit'+nodetype, self.visitDefault)
        return visitfn(node)

    def visitDefault(self, node):
        return str(node.value)

    def visitSetOperator(self, node):
        return "(" + self.visit(node.children[0]) + ") " + str(node.value) + " (" + self.visit(node.children[1]) + ")"
        
    def visitLiteral(self, node):
        if isinstance(node.value, str):
            return "'%s'" % node.value
        elif node.value is None:
            return "null"
        else:
            return str(node.value)

    def visitBinaryExpression(self, node): 
        return self.visit(node.children[0]) + \
               " " + str(node.value) + " " + \
               self.visit(node.children[1])
          
    def visitUnaryExpression(self, node): 
        return str(node.value) + " " + self.visit(node.children[0])

    def visitStar(self, node): 
        return "*"

    def visitAs(self, node):
        return self.visit(node.children[0]) + " as " + \
               self.visit(node.children[1])

    def visitDot(self, node):
        return self.visit(node.children[0]) + "." + \
               self.visit(node.children[1])

    def visitEmpty(self, node): 
        return ""

    def visitConfAgg(self, node):
        tables = "array[%s]" % \
            ", ".join("'%s'" % name for name in node.names)
        aids = "array[%s]" % \
            ", ".join("%s.trio_aid" % alias for alias in node.aliases)
        disj = "array[%s]" % \
            ", ".join("1" for alias in node.aliases)
        confs = "array[%s]" % \
            ", ".join("%s.trio_conf" % alias for alias in node.aliases)
        return "trio_get_conf(%s, %s, %s, %s, 1, 0)" % \
            (tables, aids, disj, confs)
            
    def visitConfUniform(self, node):
        conf = node.children[0]
        tables = "array[%s]" % \
            ", ".join("'%s'" % name for name in conf.names)
        aids = "array[%s]" % \
            ", ".join("%s.trio_aid" % alias for alias in conf.aliases)
        disj = "array[%s]" % \
            ", ".join("1" for alias in conf.aliases)
        confs = "array[%s]" % \
            ", ".join("%s.trio_conf" % alias for alias in conf.aliases)
        return "trio_get_conf(%s, %s, %s, %s, 1, 0)" % \
            (tables, aids, disj, confs)
            
    def visitConf(self, node):
        tables = "array[%s]" % \
            ", ".join("'%s'" % name for name in node.names)
        aids = "array[%s]" % \
            ", ".join("%s.trio_aid" % alias for alias in node.aliases)
        disj = "array[%s]" % \
            ", ".join("1" for alias in node.aliases)
        confs = "array[%s]" % \
            ", ".join("%s.trio_conf" % alias for alias in node.aliases)
        if self.cursor is not None and self.cursor.distinct:
            return "trio_get_conf(%s, %s, %s, %s, 0, 0)" % \
                (tables, aids, disj, confs)
        else:
            return "trio_get_conf(%s, %s, %s, %s, 0, 1)" % \
                (tables, aids, disj, confs)
            
    def visitLineage(self, node):
        if len(node.lineagePredicates[self.i_lin].lineagePaths) == 0:
            sql = "(false)"
        else:
            if node.hasOrPredicate or node.hasNotPredicate:
                exists = []
                for i in range(len(node.lineagePredicates[self.i_lin].lineagePaths)):
                    path = node.lineagePredicates[self.i_lin].lineagePaths[i]
                    existSQL = "exists (select null from " 
                    tableList = ["trio_lin_%s" % x for x in path[0:-1]]
                    existSQL += ", ".join( tableList ) + " where "
                    existSQL += " and ".join(node.lineagePredicates[self.i_lin].lineageWhereSQL[i]) + ")"
                    exists.append("(" + existSQL + ")")
                sql = "(" + " or ".join(exists) + ")"
            else:
                sql = "(%s.trio_aid = trio_lin_union_%d.lin_from and %s.trio_aid = trio_lin_union_%d.lin_to)" \
                    % (node.lineagePredicates[self.i_lin].fromAlias, self.i_lin, 
                       node.lineagePredicates[self.i_lin].toAlias, self.i_lin)
        self.i_lin += 1
        return sql

    def visitCreateIndex(self, node): pass

    def visitDropIndex(self, node): pass

    def visitMaybe(self, node): pass

    def visitDropTable(self, node):
        return "drop table " + self.visit(node.children[0]) \
                             + self.visit(node.children[1])

    def visitDropTableList(self, node): 
        return ", ".join([self.visit(x) for x in node.children])

    def visitCascade(self, node): 
        return " cascade"

    def visitInsert(self, node): pass
    
    def visitBrackets(self, node):
        return " ( " + self.visit(node.children[0]) + " ) "

    def visitCreateTableAs(self, node):
        return "create table " + self.visit(node.children[0]) + \
            " as " + self.visit(node.children[1]) 

    def visitSelect(self, node):
        self.i_lin = 0
        try:
            len(node.lineagePredicates)
        except:
            node.lineagePredicates = None
        
        if node.lineagePredicates is None or len(node.lineagePredicates) == 0 or node.hasOrPredicate or node.hasNotPredicate:
            return self.visit(node.children[0]) + \
               self.visit(node.children[1]) + \
               self.visit(node.children[2]) + \
               self.visit(node.children[3]) + \
               self.visit(node.children[4]) + \
               self.visit(node.children[5]) + \
               self.visit(node.children[6]) + \
               self.visit(node.children[7])
               
        else:
            
            sql = self.visit(node.children[0]) + \
                  self.visit(node.children[1]) + \
                  self.visit(node.children[2])
            l = 0
            for linpred in node.lineagePredicates:
                if len(linpred.lineagePaths) > 0:
                    sql += ", ("
                    i = 0
                    for path in linpred.lineagePaths:
                        sql += "(select %s.trio_aid as lin_from, %s.trio_aid as lin_to" % (linpred.fromAlias, linpred.toAlias)
                        sql += " from %s as %s, %s as %s" % (linpred.fromTable, linpred.fromAlias, linpred.toTable, linpred.toAlias)
                        j = 0
                        for lin_table in path:
                            if j < len(path) - 1:
                                sql += ", "
                            sql += "trio_lin_" + lin_table
                            if j == len(path) - 2:
                                break
                            j += 1
                        sql += " where " + " and ".join(linpred.lineageWhereSQL[i]) + ")"
                        if i < len(linpred.lineagePaths) - 1:
                            sql += " union "
                        i += 1
                    sql += ") trio_lin_union_%d" % l
                l += 1
            sql += self.visit(node.children[3]) + \
                   self.visit(node.children[4]) + \
                   self.visit(node.children[5]) + \
                   self.visit(node.children[6]) + \
                   self.visit(node.children[7])
            return sql

    def visitSelectClause(self, node):
        return "select " + ", ".join([self.visit(x) for x in node.children])

    def visitFromClause(self, node):
        return " from " + ", ".join([self.visit(x) for x in node.children])

    def visitWhereClause(self, node): 
        return " where " + self.visit(node.children[0])

    def visitAggregate(self, node):
        return node.value + "(" + self.visit(node.children[0]) + ")"
        
    def visitHorizontalSelect(self, node): 
        sql = "(" + self.visit(node.children[0]) + " from "
        sql += ", ".join([node.names[i] + " " +
                          node.aliases[i]
                          for i in range(len(node.names))])
        if isinstance(node.children[2], Empty):
            sql += " where "
        else:
            sql += self.visit(node.children[2]) + " and "
        sql += " and ".join([node.parent_aliases[i] + ".trio_xid = " + 
                             node.aliases[i] + ".trio_xid" 
                             for i in range(len(node.aliases))]) #+ ")"
        if isinstance(node.children[3], GroupbyClause):
            sql += self.visit(node.children[3])
        if isinstance(node.children[4], HavingClause):
            sql += self.visit(node.children[4])
        sql += ")"
        return sql
        
    def visitGroupbyClause(self, node):
        return " group by " + ",".join([self.visit(x) for x in node.children])
        
    def visitOrderBy(self, node):
        return " order by " + ", ".join([self.visit(x) for x in node.children])
        
    def visitHavingClause(self, node):
        return " having " + self.visit(node.children[0])

    def visitEsum(self, node):
        colProbability = self.visitConfAgg(node.value)
        return "sum(" + self.visit(node.children[0]) + " * " + colProbability + ")";

    def visitEcountStar(self, node):  
        sumStr = "sum(" + self.visitConfAgg(node.value) + ")"
	return "(case when %s is NULL then 0 else %s end)" % (sumStr, sumStr)

    def visitEcount(self, node):  
        colExists = "(case when ((" + self.visit(node.children[0]) + ") is not null) then 1 else 0 end)"
        colProbability = self.visitConfAgg(node.value)
        sumStr = "sum(" + colExists + "*" + colProbability + ")"
	return "(case when %s is NULL then 0 else %s end)" % (sumStr, sumStr)

    def visitEmin(self, node):
        return "min(" + self.visit(node.children[0]) + ")"

    def visitEmax(self, node):
        return "max(" + self.visit(node.children[0]) + ")"

    def visitEavg(self, node):
        return self.visitEsum(node) + "/" + self.visitEcount(node)

    def visitAggregate(self, node):
        return self.visit(node.children[0]) # visit the actual aggregate

    def visitGroupByKey(self, node):
        if (node.children):
            return " group by " + ",".join([self.visit(x) for x in node.children])
        else:
            return ""

    def visitSum(self, node):
        return "sum(" + self.visit(node.children[0]) + ")"

    def visitCount(self, node):
        if len(node.children) == 1:
            return "count(" + self.visit(node.children[0]) + ")"
        else:
            return "count(distinct " + self.visit(node.children[1]) + ")"

    def visitMin(self, node): 
        return "min(" + self.visit(node.children[0]) + ")"

    def visitMax(self, node):
        return "max(" + self.visit(node.children[0]) + ")"

    def visitAvg(self, node):
        return "avg(" + self.visit(node.children[0]) + ")"

    def visitCreateTable(self, node):
        return "create trio table " + self.visit(node.children[0]) + \
            "(" + self.visit(node.children[1]) + ")"

    def visitColumn(self, node): 
        return self.visit(node.children[0]) + " " + \
            self.visit(node.children[1])

    def visitMaybe(self, node):
        return "(%s.trio_enc > %s.trio_alts)" % ( node.value, node.value )

    def visitColumnList(self, node): 
        return ", ".join([self.visit(x) for x in node.children])

    def visitIdentifierList(self, node): pass

    def visitQuestion(self, node): pass

    def visitInsertList(self, node): pass

    def visitInsertSpec(self, node): pass

    def visitAlternative(self, node):
        return ", ".join([self.visit(x) for x in node.children])

    def visitDataType(self, node):
        if len(node.children) == 0 or isinstance(node.children[0], Empty):
            return node.value
        else:
            return node.value + "(" + self.visit(node.children[0]) + ")"
    
