__id__ = "$Id: rewriter.py 573 2007-10-29 17:11:54Z theobald $"

import common
from node import *
from visitor import Visitor
from triql_printer import TriQLPrinter

def lineageWhereSQLPath(source, target):
    sourceLineage = "trio_lin_%s" % source
    targetLineage = "trio_lin_%s" % target
    return ["%s.srclin = %s.aid" % (sourceLineage, targetLineage),
            "%s.srctable = '%s'" % (sourceLineage, target)]
                
def lineageWhereSQLStart(node):
        sourceLineage = "trio_lin_%s" % node.fromTable
        return ["%s.aid = %s.trio_aid" % (sourceLineage, node.fromAlias)]
        
def lineageWhereSQLEnd(source, node):
        sourceLineage = "trio_lin_%s" % source
        return ["%s.srclin = %s.trio_aid" % (sourceLineage, node.toAlias),
                "%s.srctable = '%s'" % (sourceLineage, node.toTable)]

def lineageWhereSQL(node):
    pathsSQL = []
    for path in node.lineagePaths:
        whereSQL = lineageWhereSQLStart(node)
        if len(path) > 1:
            for i in range(len(path)-2): 
                whereSQL += lineageWhereSQLPath(path[i] ,path[i+1])
            whereSQL += lineageWhereSQLEnd(path[-2], node)
        pathsSQL.append(whereSQL)
    return pathsSQL

class LineagePredicate(Visitor):
    def __init__(self, fromTable, fromAlias, toTable, toAlias, lineagePaths, lineageWhereSQL):
        self.fromTable = fromTable
        self.fromAlias = fromAlias
        self.toTable = toTable
        self.toAlias = toAlias
        self.lineagePaths = lineagePaths
        self.lineageWhereSQL = lineageWhereSQL

# Create mappings alias -> tablename for a select node
def checkAliases(selectNode, horizontal=False):
    if horizontal:
        fromIndex = 1
    else:
        fromIndex = 2
    selectNode.aliasdict = {}
    selectNode.names = []
    selectNode.aliases = []
    for child in selectNode.children[fromIndex].children:
        name = ""
        alias = "" 
        if isinstance(child, As):
            name = child.children[0].value
            if horizontal:
                alias = child.children[1].value + "_inner"
            else:
                alias = child.children[1].value
        else:
            name = child.value
            if horizontal:
                alias = child.value + "_inner"
            else:
                alias = child.value
        selectNode.names.append(name)
        selectNode.aliases.append(alias)
        if selectNode.aliasdict.has_key(alias):
            raise Exception('Table name "%s" specified more than once without aliasing' % alias)
        selectNode.aliasdict[alias] = name
            
class Rewriter(Visitor):
    def __init__(self, trio_cursor):
        self.trio_cursor = trio_cursor
        self.cursor = trio_cursor.tdb.pdb.cursor()
        
        # a stack of contexts for (potentially) nested select statments
        self.context = []
        
    def visit(self, node):
        nodetype = node.__class__.__name__
        visitfn = getattr(self, 'visit'+nodetype, self.visitDefault)
        new_node = visitfn(node)
        
        return new_node

    def visitDefault(self, node):
        new_children = []
        for child in node.children:
            new_children += [self.visit(child)]
        node.children = new_children
        return node
        
    def visitDot(self, node):
        if isinstance(self.context[-1], HorizontalSelect) and \
           self.context[-1].aliasdict.has_key(node.children[0].value + "_inner"):
                node.children[0].value += "_inner"
        return node
        
    def visitSetOperator(self, node):
        if node.value != "union all":
            hasDistinct = False
            for child in node.children[0].children[0].children:
                if child.value == "distinct":
                    hasDistinct = True
                    break
            if not hasDistinct:
                node.children[0].children[0].children.append(Node("distinct"))
            hasDistinct = False
            for child in node.children[1].children[0].children:
                if child.value == "distinct":
                    hasDistinct = True
                    break
            if not hasDistinct:
                node.children[1].children[0].children.append(Node("distinct"))
        else:
            if isinstance(node.children[2], ComputeConfidences):
                # left subquery
                gotAsConf = isinstance(node.children[0].children[4], ComputeConfidences)
                if not gotAsConf:
                    for child in node.children[0].children[1].children:
                        if isinstance(child, As) and child.children[1].value == "conf":
                            gotAsConf = True
                            break
                if not gotAsConf:
                    node.children[0].children[4] = ComputeConfidences()
                # right subquery
                gotAsConf = isinstance(node.children[1].children[4], ComputeConfidences)
                if not gotAsConf:
                    for child in node.children[1].children[1].children:
                        if isinstance(child, As) and child.children[1].value == "conf":
                            gotAsConf = True
                            break
                if not gotAsConf:
                    node.children[1].children[4] = ComputeConfidences()
        
        p = TriQLPrinter()
        node.leftquery = p.visit(node.children[0])
        node.rightquery = p.visit(node.children[1])
        node.children = [self.visit(node.children[0]), self.visit(node.children[1]), node.children[2]]
        
        if node.children[1].isConfBase and not node.children[0].isConfBase \
           and not node.children[0].gotAsConf \
           and node.children[0].gotUniformT != 1 \
           and common.tableType(node.children[0].names, self.trio_cursor.tdb.pdb.cursor()) > 1:
            node.children[0].children[1].children[-2].children[0] = Conf(children=[Star()])
            node.children[0].gotConfStar = node.children[0].children[1].children[-2].children[0]
            node.leftquery += " compute confidences"
            
        if node.children[0].isConfBase and not node.children[1].isConfBase \
           and not node.children[1].gotAsConf \
           and node.children[1].gotUniformT != 1 \
           and common.tableType(node.children[1].names, self.trio_cursor.tdb.pdb.cursor()) > 1:
            node.children[1].children[1].children[-2].children[0] = Conf(children=[Star()])
            node.children[1].gotConfStar = node.children[1].children[1].children[-2].children[0]
            node.rightquery +=  " compute confidences"
            
        return node
        
    def visitSelect(self, node):
        node.OrderByAliases = []
        node.OrderByColumns = []
        node.aliasStr = []
        node.gotAgg = 0
        node.gotUniformT = 0
        node.notAs = None
        
        node.lineageNodes = []
        node.lineagePredicates = []
        node.hasOrPredicate = False
        node.hasNotPredicate = False
        node.gotConfStar = None
        node.isConfBase = False
        node.gotAsConf = False
        node.gotUniform = False
        node.distinct = False
        node.merged = False
        node.flatten = False
        node.groupAlts = False
        
        # Create FROM table alias -> FROM table name dictionary.
        checkAliases(node)
        
        # Create a new list for result column inferencing
        node.inferenceList = []
        
        self.context.append(node)
        self.visit(node.children[0]) # select options
        self.visit(node.children[2]) # from clause
        self.visit(node.children[1]) # select clause
        
        if node.gotConfStar is not None:
            if node.distinct:
                raise Exception("Conf(*) not allowed in the SELECT clause in conjunction with DISTINCT")
            elif node.merged:
                raise Exception("Conf(*) in the SELECT clause or 'Compute Confidences' are not allowed in conjunction with MERGED")
                
        self.visit(node.children[3]) # where clause
        self.visit(node.children[5]) # group by keys
        
        # Check if lineage is only conjunctive
        for ln in node.lineageNodes:
            ln.hasOrPredicate = self.context[-1].hasOrPredicate
            ln.hasNotPredicate = self.context[-1].hasNotPredicate
        
        # Check for GroupAlts modifier in select-options and rewrite
        for i in range(len(node.children[0].children)):
            if isinstance(node.children[0].children[i], GroupAlts):
                self.context.append(node.children[0].children[i])
                self.visit(node.children[0].children[i])
                self.context.pop()
                node.groupAlts = True
        self.context.pop()
        
        # Read uncertain columns per table from metadata
        node.u_columns = []
        columnmap = []
        hasUncertain = False
        for name in node.names:
            columnmap.append([name, []])
            sql = "SELECT attrname FROM trio_attribs WHERE tablename = '%s' and iscertain=FALSE ORDER BY colnum ASC" % name
            self.cursor.execute(sql)
            cols = self.cursor.fetchall()
            for c in cols:
                columnmap[-1][1].append(c[0])
                hasUncertain = True
        
        # Map uncertain columns to their indexes in the SELECT
        i = 0
        for columnItem in node.inferenceList:
            for item in columnItem:
                if len(item) == 1:
                    for u in columnmap:
                        if item[0] in u[1]:
                            node.u_columns.append(i)
                            break
                elif len(item) == 2:
                    if item[0] in node.aliases:
                        table = node.names[node.aliases.index(item[0])]
                    elif item[0] in node.names:
                        table = node.names[node.names.index(item[0])]
                    else:
                        raise Exception("Unresolvable table reference '" + item[0] + "'")
                    for u in columnmap:
                        if u[0] == table and item[1] in u[1]:
                            node.u_columns.append(i)
                            break
            i += 1
        
        # Check if trio_cols have to be added to ORDER BY
        if len(node.OrderByColumns) > 0: 
            if isinstance(node.children[6], Empty):
                node.children[6] = OrderBy(children=node.OrderByColumns)
            else:
                node.children[6].children = node.OrderByColumns + node.children[6].children
                
        # ORDER BY certain columns, then by xid's
        if not node.distinct and not node.groupAlts:
            certainCols = []
            for o in range(len(node.OrderByAliases)):
                if not o in node.u_columns:
                    certainCols.append(node.OrderByAliases[o])
            node.children[6].children = certainCols + node.children[6].children
            
        return node
        
    def visitAggregate(self, node):
        if isinstance(node.children[0], Sum) or isinstance(node.children[0], Count) or \
           isinstance(node.children[0], Min) or isinstance(node.children[0], Max) or \
           isinstance(node.children[0], Avg):
            if not isinstance(self.context[-1], HorizontalSelect):
                raise Exception(node.children[0].__class__.__name__ + " only supported in horizontal subquery");
        else: # remove this later
            raise Exception("Aggregates only supported in horizontal subqueries");
        if not isinstance(self.context[-1], HorizontalSelect):
            if self.context[-1].notAs:
                if isinstance(node.children[0], EcountStar):
                    aliasStr = "ecount"
                else:
                    aliasStr = node.children[0].__class__.__name__
                self.context[-1].aliasStr.append(aliasStr.lower())
            self.context[-1].gotAgg = 1
            confNode = Conf(children=[Star()])
            confNode = self.visit(confNode)
            node.children[0].value = confNode
        else:
            for n in node.children:
                self.visit(n)
        return node
        
    def visitCount(self, node):
        for n in node.children:
            if not isinstance(n, Star) or not isinstance(self.context[-1], HorizontalSelect):
                self.visit(n)
        return node
        
    def visitSum(self, node):
        for n in node.children:
            if not isinstance(n, Star) or not isinstance(self.context[-1], HorizontalSelect):
                self.visit(n)
        return node
        
    def visitAvg(self, node):
        for n in node.children:
            if not isinstance(n, Star) or not isinstance(self.context[-1], HorizontalSelect):
                self.visit(n)
        return node
        
    def visitMax(self, node):
        for n in node.children:
            if not isinstance(n, Star) or not isinstance(self.context[-1], HorizontalSelect):
                self.visit(n)
        return node
        
    def visitMin(self, node):
        for n in node.children:
            if not isinstance(n, Star) or not isinstance(self.context[-1], HorizontalSelect):
                self.visit(n)
        return node
        
    def visitFromClause(self, node):
        if not isinstance(self.context[-1], HorizontalSelect):
            for child in node.children:
                if isinstance(child, As):
                    child = child.children[0]
                type = common.tableType([child.value], self.trio_cursor.tdb.pdb.cursor())
                if type != common.TYPE_UNCERTAIN and isinstance(child, UniformTable):
                    raise Exception("'uniform' can only be specified for uncertain tables without confidences")
                if type == common.TYPE_UNCERTAIN and isinstance(child, UniformTable):
                    self.context[-1].isConfBase = True
                    self.context[-1].gotUniformT = 1
                    self.context[-1].children[4] = Empty()
        else:
            return node
        
    def getConfNode(self, node):
        aliases = self.context[-1].aliases
        confNode = None
        
        for child in node.children:
            if isinstance(child, As) and child.children[1].value == "conf":
                self.context[-1].gotAsConf = True
                self.context[-1].isConfBase = True
                if confNode is not None:
                    raise Exception("Multiple 'as conf' columns specified!")
                if len(node.children) == 1:
                    raise Exception("No data columns selected!")
                if child.children[0].value == 'uniform':
                    confNode = As(children=[Literal(None), Node("trio_conf")])
                    self.context[-1].gotUniform = True
                    node.children.remove(child)
                else:
                    confNode = As(children=[self.visit(child.children[0]), Node("trio_conf")])
                    node.children.remove(child)
                    
        if confNode is None:
            if self.context[-1].gotUniformT == 1:
                confNode = As(children=[ConfUniform(children=[self.visit(Conf(children=[Star()]))]), Node("trio_conf")])
            elif self.context[-1].gotConfStar is not None and not self.context[-1].distinct:
                confNode = As(children=[self.context[-1].gotConfStar, Node("trio_conf")])
            elif isinstance(self.context[-1].children[4], ComputeConfidences) and not self.context[-1].distinct:
                confNode = As(children=[self.visit(Conf(children=[Star()])), Node("trio_conf")])
                self.context[-1].gotConfStar = confNode
                self.context[-1].gotAsConf = True
            elif len(aliases) == 1:
                try:
                    if self.context[-1].isViewTable:
                        confNode = As(children=[Node(aliases[0] + ".trio_conf"), Node("trio_conf")])
                except:
                    confNode = As(children=[Literal(None), Node("trio_conf")])
            else:
                confNode = As(children=[Literal(None), Node("trio_conf")])
                
        return confNode
        
    def appendTrioColumns(self, node):
        if self.context[-1].gotAgg == 1:
            node.children = node.children + \
            [As(children=[Node("nextval('aggs_trio_aid')"), Node("trio_aid")])] + \
            [As(children=[Node("nextval('aggs_trio_xid')"), Node("trio_xid")])] + \
            [As(children=[Node("1"), Node("trio_enc")])] + \
            [As(children=[Node("1"), Node('trio_conf')])] + \
            [As(children=[Node("1"), Node("trio_alts")])] 
            return node
        tables = self.context[-1].names
        aliases = self.context[-1].aliases
        pairs = map(None, tables, aliases)
        trio_columns = []
        # CAUTION: the order of system columns in the select clause is fixed!
        for pair in pairs:
            # Add aid
            trio_columns += [As(children=[Node("%s.trio_aid" % pair[1]), Node("trio_aid_%s_%s" \
                             % (pair[0], pair[1]))])]
        for pair in pairs:
            # Add xid
            trio_columns += [As(children=[Node("%s.trio_xid" % pair[1]), Node("trio_xid_%s_%s" \
                             % (pair[0], pair[1]))])]
            self.context[-1].OrderByColumns += [Node("trio_xid_%s_%s" % (pair[0], pair[1]))]
        # Add enc
        enc = "*".join("%s.trio_enc" % table for table in aliases)
        trio_columns += [As(children=[Node(enc), Node("trio_enc")])]
        # Add conf
        trio_columns += [self.getConfNode(node)]
        # Add alts necessary for GroupAlts modifier, CAUTION: the alts-node MUST directly succeed the conf-node!
        alts = "*".join("%s.trio_alts" % table for table in aliases)
        trio_columns += [As(children=[Node(alts), Node("trio_alts")])]
        node.children = node.children + trio_columns
    
    def visitSelectClause(self, node):
        self.context[-1].notAs = 0
        self.context[-1].OrderByAliases = []
        new_children = []
        
        for child in node.children:
            if isinstance(child, As) and child.children[1].value == "conf":
                new_child = child
            else:
                # Add an alias for aggregates if no alias is provided in the query
                # Make sure that this is not an expression itself. If it is, then don't add an alias
                if len(child.children) == 1:
                    self.context[-1].notAs = 1
                    self.context[-1].aliasStr = []
                new_child = self.visit(child)
                # Add an alias if a single aggregate was found in this projected column
                if len(self.context[-1].aliasStr) == 1:
                    aliasStr = self.context[-1].aliasStr[0]
                    new_child = As(children=[new_child, Node(aliasStr)])
                if isinstance(child, As):
                    self.context[-1].OrderByAliases += [new_child.children[1]]
                else:
                    if isinstance(new_child, list):
                        self.context[-1].OrderByAliases += new_child
                    else:
                        self.context[-1].OrderByAliases += [new_child]
            if isinstance(new_child, list):
                new_children += new_child
            else:
                new_children += [new_child]
        node.children = new_children
        
        if not isinstance(self.context[-1], HorizontalSelect) and \
           not isinstance(self.context[-1], GroupAlts):
            # Check for certain and uncertain columns
            for child in node.children:
                if not isinstance(child, As) or not child.children[1].value == "conf":
                    self.context[-1].inferenceList.append([])
                    self.appendInferenceList(child)
                    
            # Append Trio columns to outer query only
            self.appendTrioColumns(node)
            if not self.context[-1].isConfBase:
                self.context[-1].isConfBase = self.checkHorizontalSelect(node)
                
            # Check for 'DISTINCT'
            if self.context[-1].distinct:
                self.context[-1].OrderByColumns = self.context[-1].OrderByAliases + self.context[-1].OrderByColumns
                
        return node
    
    def appendInferenceList(self, node):
        if len(node.children) == 0:
            self.context[-1].inferenceList[-1].append([node.value])
        elif isinstance(node, Dot):
            self.context[-1].inferenceList[-1].append([node.children[0].value, node.children[1].value])
        elif isinstance(node, HorizontalSelect):
            # Check for a nested select list
            for child in node.children[0].children:
                self.appendInferenceList(child)
        elif isinstance(node, Aggregate):
            # Currently still considered certain
            pass        
        elif isinstance(node, As):
            self.appendInferenceList(node.children[0])
        else:
            for child in node.children:
                self.appendInferenceList(child)
        
    def visitConf(self, node):
        if isinstance(node.children[0], Star):
            if self.context[-1].gotUniformT != 1:
                self.context[-1].gotConfStar = node
            node.children = []
            for alias in self.context[-1].aliases:
                node.children.append(Node(alias))
        
        node.names = []
        node.aliases = []
        for child in node.children:
            try:
                if isinstance(self.context[-1], HorizontalSelect) and \
                   self.context[-1].aliasdict.has_key(child.value + "_inner"):
                       node.names += [self.context[-1].aliasdict[child.value + "_inner"]]
                       node.aliases += [child.value + "_inner"]
                else:
                    node.names += [self.context[-1].aliasdict[child.value]]
                    node.aliases += [child.value]
            except:
                try:
                    node.names += [self.context[-2].aliasdict[child.value]]
                    node.aliases += [child.value]
                except:
                    raise Exception("Unresolvable table alias '" + child.value + "' in conf")
        
        # Check if conf() called correctly on a non-orange set of tables
        if common.tableType(node.names, self.cursor) != common.TYPE_PROBABILISTIC and \
           common.tableType(node.names, self.cursor) != common.TYPE_CONFIDENCE and \
           not self.context[-1].gotAgg == 1 and not self.context[-1].gotUniformT == 1:
            raise Exception("Conf() called on a table with no confidences")
            
        return node
        
    def getLineagePaths(self, source, target):
        current = source[-1]
        if current == target:
            return [list(source)]
        result = []
        self.trio_cursor.execute("\l %s" % current)
        for nextTable in self.trio_cursor.fetchall():
            source.append(nextTable[0])
            result += self.getLineagePaths(source, target)
            del source[-1]
        return result

    def visitLineage(self, node):
        node.fromAlias = node.children[0].value
        node.toAlias = node.children[1].value
        node.fromTable = self.context[-1].aliasdict[node.fromAlias]
        node.toTable = self.context[-1].aliasdict[node.toAlias]
        node.lineagePaths = self.getLineagePaths([node.fromTable], node.toTable)
        self.context[-1].lineagePredicates.append(LineagePredicate(node.fromTable, \
                                                  node.fromAlias, \
                                                  node.toTable, \
                                                  node.toAlias, \
                                                  node.lineagePaths, \
                                                  lineageWhereSQL(node)))
        node.lineagePredicates = self.context[-1].lineagePredicates
        self.context[-1].lineageNodes.append(node)
        return node

    def visitGroupAlts(self, node):
        self.context[-1].isConfBase = True
        self.context[-1].OrderByColumns = [Node("trio_groupalts")]
        return node

    def visitStar(self, node):
        # Expand * according to the current context, excluding Trio system columns
        tables = self.context[-1].names
        aliases = self.context[-1].aliases
        pairs = map(None, tables, aliases)
        attribs = self.expandStar(pairs)
        for attrib in attribs:
            self.visit(attrib)
        return attribs
        
    def visitConfScaled(self, node):
        if isinstance(node.children[0], Empty):
            new_node = BinaryExpression("/", children=[Conf(children=[Star()]), \
                               HorizontalSelect(children=[SelectClause(children=[Aggregate(children=[Sum(children=[Conf(children=[Star()])])])]), Empty(), Empty(), Empty(), Empty()])])
        else:
            temp_node = node.children[0]
            if isinstance(node.children[0], Conf):
                temp_node_2 = Conf("", node.children[0].children)
            else:
                temp_node_2 = node.children[0]
            new_node = BinaryExpression("/", children=[BinaryExpression("*", children=[temp_node, Literal(1.0)]), \
                               HorizontalSelect(children=[SelectClause(children=[Aggregate(children=[Sum(children=[temp_node_2])])]), Empty(), Empty(), Empty(), Empty()])])
        self.context[-1].gotAsConf = True
        return self.visit(new_node)
        
    def checkHorizontalSubset(self, selectClause):
        parent_names = list(self.context[-2].names)
        parent_aliases = list(self.context[-2].aliases)
        selectClause.parent_aliases = []
        for i in range(len(selectClause.names)):
            if selectClause.names[i] in parent_names:
                j = parent_names.index(selectClause.names[i])
            elif selectClause.names[i] in parent_aliases:
                j = parent_aliases.index(selectClause.names[i])
                # We have an alias reference to an outer table
                # => swap it with the outer table name & keep the inner alias
                selectClause.names[i] = parent_names[j]
            else:
                raise Exception("Table reference '%s' not found in outer query!" % selectClause.names[i])
            selectClause.parent_aliases.append(parent_aliases[j])
            del parent_names[j]
            del parent_aliases[j]
            
    def visitHorizontalSelect(self, node):
        node.OrderByColumns = []
        node.aliasStr = []
        node.gotAgg = 0
        node.gotUniformT = 0
        node.notAs = None
        
        self.context.append(node)
        # abbreviated
        if isinstance(node.children[1], Empty):
            node.names = self.context[-2].names
            node.aliases = [x + "_inner" for x in self.context[-2].aliases]
            node.aliasdict = dict(map(None, node.aliases, node.names))
            node.parent_names = self.context[-2].names
            node.parent_aliases = self.context[-2].aliases
            node.children[1] = self.context[-2].children[2]
        # non-abbreviated
        else:
            checkAliases(node, True)
            self.checkHorizontalSubset(node)
            
        # Visit children
        new_children = []
        for child in node.children:
            new_children += [self.visit(child)]
        node.children = new_children
        
        # Clear subquery context
        self.context.pop()
        
        return node

    def expandStar(self, tables):
        sql = "SELECT attrname FROM trio_attribs WHERE tablename='%s' ORDER BY colnum ASC"
        attribs = []
        for table in tables:
            q = sql % table[0]
            self.cursor.execute(q)
            for row in self.cursor.fetchall():
                attribs.append(Dot(children=[Node(table[1]), Node(row[0])]))
        return attribs
        
    def visitBinaryExpression(self, node):
        if node.value == "or":
            self.context[-1].hasOrPredicate = True
        return self.visitDefault(node)
    
    def visitUnaryExpression(self, node):
        if node.value == "not":
            self.context[-1].hasNotPredicate = True
        return self.visitDefault(node)
        
    def checkHorizontalSelect(self, node):
        if isinstance(node, HorizontalSelect):
            return True
        elif len(node.children) > 0:
            for n in node.children:
                return self.checkHorizontalSelect(n)
        else:
            return False
            
    def visitSelectOptions(self, node):
        for child in node.children:
            if child.value == 'nolineage':
                self.context[-1].isConfBase = True
            elif child.value == 'flatten':
                self.context[-1].isConfBase = True
            elif child.value == 'distinct':
               self.context[-1].distinct = True
            elif child.value == 'merged':
               self.context[-1].merged = True
            self.visit(child)
        return node

