__id__ = "$Id: printer.py 314 2007-03-01 04:39:17Z ragho $"

from node import *
from visitor import Visitor

# This class prints TriQL parse trees back to TriQL query strings
# (needed for manipulating subqueries etc. at the TriQL level)
class TriQLPrinter(Visitor):
    
    def visit(self, node):
        nodetype = node.__class__.__name__
        visitfn = getattr(self, 'visit'+nodetype, self.visitDefault)
        return visitfn(node)

    def visitDefault(self, node):
        return str(node.value)

    def visitCreateTable(self, node): pass
    
    def visitCreateTableAs(self, node): pass
    
    def visitInsert(self, node): pass

    def visitCreateTempTableAs(self, node): pass
    
    def visitCreateIndex(self, node): pass

    def visitDropIndex(self, node): pass

    def visitMaybe(self, node): pass

    def visitIdentifierList(self, node): pass

    def visitQuestion(self, node): pass

    def visitInsertList(self, node): pass

    def visitInsertSpec(self, node): pass
    
    def visitViewTable(self, node): pass

    def visitDropTable(self, node): pass

    def visitDropTableList(self, node): pass
    
    def visitLiteral(self, node):
        if isinstance(node.value, str):
            return "'%s'" % node.value
        elif node.value is None:
            return "null"
        else:
            return str(node.value)

    def visitBinaryExpression(self, node): 
        return self.visit(node.children[0]) + \
               " " + str(node.value) + " " + \
               self.visit(node.children[1])
          
    def visitUnaryExpression(self, node): 
        return str(node.value) + self.visit(node.children[0])

    def visitStar(self, node): 
        return "*"

    def visitAs(self, node):
        return self.visit(node.children[0]) + " as " + \
               self.visit(node.children[1])

    def visitDot(self, node):
        return self.visit(node.children[0]) + "." + \
               self.visit(node.children[1])

    def visitEmpty(self, node): 
        return ""

    def visitConfAgg(self, node):
        return ""

    def visitConfUniform(self, node):
        return "uniform as conf"

    def visitConf(self, node):
        return "conf(" + self.visit(node.children[0]) + ")"

    def visitComputeConfidences(self, node):
        return " compute confidences"
        
    def visitLineage(self, node):
        return self.visit(node.children[0]) + " ==> " + self.visit(node.children[1])

    def visitCascade(self, node): 
        return " cascade"

    def visitBrackets(self, node):
        return " ( " + self.visit(node.children[0]) + " ) "

    def visitCreateTableAs(self, node):
        return "create table " + self.visit(node.children[0]) + \
            " as " + self.visit(node.children[1]) 

    def visitSelect(self, node):
        sql = "select " + self.visit(node.children[0]) + " " +\
              self.visit(node.children[1]) + \
              self.visit(node.children[2]) + \
              self.visit(node.children[3]) + \
              self.visit(node.children[4]) + \
              self.visit(node.children[5]) + \
              self.visit(node.children[6]) + \
              self.visit(node.children[7])
        return sql

    def visitSelectClause(self, node):
        return ", ".join([self.visit(x) for x in node.children])
        
    def visitSelectOptions(self, node):
        return " ".join([self.visit(x) for x in node.children])

    def visitFromClause(self, node):
        return " from " + ", ".join([self.visit(x) for x in node.children])

    def visitWhereClause(self, node): 
        return " where " + self.visit(node.children[0])
        
    def visitGroupAlts(self, node):
        return "groupalts(" + self.visit(node.children[0]) + ")"

    def visitAggregate(self, node):
        return node.value + "(" + self.visit(node.children[0]) + ")"
        
    def visitHorizontalSelect(self, node): 
        return " [ " + self.visit(node.children[0]) + \
            self.visit(node.children[1]) + \
            self.visit(node.children[2]) + \
            self.visit(node.children[3]) + \
            self.visit(node.children[4])+ " ] "
        
    def visitGroupbyClause(self, node):
        return " group by " + ",".join([self.visit(x) for x in node.children])
        
    def visitOrderBy(self, node):
        return " order by " + ", ".join([self.visit(x) for x in node.children])
        
    def visitHavingClause(self, node):
        return " having " + self.visit(node.children[0])

    def visitEavg(self, node):
        return "eavg(" + self.visit(node.children[0]) + ")"

    def visitEsum(self, node):
        return "esum(" + self.visit(node.children[0]) + ")";

    def visitEcountStar(self, node):  
        return "ecount(*)"
        
    def visitEcount(self, node):  
        return "ecount(" + self.visit(node.children[0]) + ")";

    def visitEmin(self, node):
        return "min(" + self.visit(node.children[0]) + ")"

    def visitEmax(self, node):
        return "max(" + self.visit(node.children[0]) + ")"

    def visitAggregate(self, node):
        return self.visit(node.children[0]) # visit the actual aggregate

    def visitGroupByKey(self, node):
        return " group by " + ",".join([self.visit(x) for x in node.children])

    def visitSum(self, node):
        return "sum(" + self.visit(node.children[0]) + ")"

    def visitCount(self, node):
        return "count(" + self.visit(node.children[0]) + ")"

    def visitMin(self, node): 
        return "min(" + self.visit(node.children[0]) + ")"

    def visitMax(self, node):
        return "max(" + self.visit(node.children[0]) + ")"

    def visitAvg(self, node):
        return "avg(" + self.visit(node.children[0]) + ")"

    def visitColumn(self, node): 
        return self.visit(node.children[0]) + " " + self.visit(node.children[1])

    def visitMaybe(self, node):
        return "(%s.trio_enc > %s.trio_alts)" % ( node.value, node.value )

    def visitColumnList(self, node): 
        return ", ".join([self.visit(x) for x in node.children])

    def visitAlternative(self, node):
        return ", ".join([self.visit(x) for x in node.children])

    def visitDataType(self, node):
        if len(node.children) == 0 or isinstance(node.children[0], Empty):
            return node.value
        else:
            return node.value + "(" + self.visit(node.children[0]) + ")"
            
    def visitUniformTable(self, node):
        return "uniform " + node.value
    
